/*
Directory listing CGI program
By Steven Lawrance
(c) Steven Lawrance, 2000, 2002, 2004
(c) Bryant College, 1998-2000

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "dirlist.h"
#include "database.h"
#include "util.h"
#include "io.h"
#include "globals.h"

LPDIRLIST_CLASSVALUE GetClassValue(LPDIRLIST_CLASS lpClass, DWORD idValue)
{
  /* Get the class data corresponding to the given class ID number */
  DWORD i = 0;
  BOOL bFound = FALSE;
  if (!lpClass)
    return NULL;
  while ((i < lpClass->nValues) && !bFound) {
    if (lpClass->lpValues[i].idValue == idValue)
      bFound = TRUE;
    else
      i++;
  }
  if (bFound)
    return &(lpClass->lpValues[i]);
  else
    return NULL;
}

LPDIRLIST_CLASS GetClass(LPDIRLIST_CLASS lpClassList, DWORD idClass)
{
  /* Get the class given the class ID number */
  BOOL bFound = FALSE;
  while ((lpClassList != NULL) && !bFound) {
    if (lpClassList->idClass == idClass)
      bFound = TRUE;
    else
      lpClassList = (LPDIRLIST_CLASS) lpClassList->next;
  }
  return lpClassList;
}

LPDIRLIST_CLASSDATA GetClassData(LPDIRLIST_ENTRY lpDirEntry, DWORD idClass)
{
  /* Get the class data corresponding to the given class ID number */
  DWORD i = 0;
  BOOL bFound = FALSE;
  while ((i < lpDirEntry->nClassDataValues) && !bFound) {
    if (lpDirEntry->lpClassData[i].idClass == idClass)
      bFound = TRUE;
    else
      i++;
  }
  if (bFound)
    return &(lpDirEntry->lpClassData[i]);
  else
    return NULL;
}

/* Compare two dirlist entries */
int CompareEntries(LPDIRLIST_ENTRY lpEntry1, LPDIRLIST_ENTRY lpEntry2, LPDIRLIST_CLASS lpClassRoot, LPDWORD dwOrder, DWORD nLayers, BOOL *bExt, BOOL *bDesc)
{
  /* Compare two entries using the requested sort order */
  LPDIRLIST_CLASSDATA lpClassData1;
  LPDIRLIST_CLASSDATA lpClassData2;
  LPDIRLIST_CLASS lpClass;
  LPDIRLIST_CLASSVALUE lpDefaultValue;
  DWORD i = 0, a;
  BOOL iCmp = 0;

  /* Loop through while we think two entries are equal and while more sort layers exist */
  while (!iCmp && (i < nLayers))
  {
    /* Sort the name after deciding how to sort it: */
    if (!bExt[i])
    {
      /* Username */
      if (dwOrder[i] & DISPLAY_USERNAME)
        iCmp = CompareStrings(lpEntry1->lpszUserName, lpEntry2->lpszUserName);

      /* Firstname, then Lastname */
      if ((dwOrder[i] & DISPLAY_LASTNAME) && (dwOrder[i] & DISPLAY_FIRSTNAME) && ((dwOrder[i] & DISPLAY_LASTNAMEFIRST) != DISPLAY_LASTNAMEFIRST) && !iCmp)
        iCmp = CompareNames(lpEntry1->lpszFirstName, lpEntry1->lpszLastName, lpEntry2->lpszFirstName, lpEntry2->lpszLastName, FALSE);

      /* Lastname, then Firstname */
      if ((dwOrder[i] & DISPLAY_LASTNAMEFIRST) && !iCmp)
        iCmp = CompareNames(lpEntry1->lpszFirstName, lpEntry1->lpszLastName, lpEntry2->lpszFirstName, lpEntry2->lpszLastName,TRUE);

      /* Firstname */
      if (!(dwOrder[i] & DISPLAY_LASTNAME) && (dwOrder[i] & DISPLAY_FIRSTNAME) && !iCmp)
        iCmp = CompareStrings(lpEntry1->lpszFirstName, lpEntry2->lpszFirstName);

      /* Lastname */
      if ((dwOrder[i] & DISPLAY_LASTNAME) && !(dwOrder[i] & DISPLAY_FIRSTNAME) && !iCmp)
        iCmp = CompareStrings(lpEntry1->lpszLastName, lpEntry2->lpszLastName);

      /* Go through each attribute and process them, if this isn't extended */
      a = DISPLAY_MIN;
      while ((a <= DISPLAY_MAX) && !iCmp) {
        if (dwOrder[i] & a)
        {
          /* Breakout for sorting by this attribute */
          switch (a) {

          case DISPLAY_FLAGS :
            if (lpEntry1->dwFlags > lpEntry2->dwFlags)
              iCmp = 1;
            else {
              if (lpEntry1->dwFlags < lpEntry2->dwFlags)
                iCmp = -1;
              else
                iCmp = 0;
            }
            break;

          case DISPLAY_WEBSITE :
            iCmp = CompareStrings(lpEntry1->lpszWebSite, lpEntry2->lpszWebSite);
            break;

          case DISPLAY_EMAIL :
            iCmp = CompareStrings(lpEntry1->lpszEMail, lpEntry2->lpszEMail);
            break;

          case DISPLAY_TELEPHONE :
            iCmp = CompareStrings(lpEntry1->lpszTelephone, lpEntry2->lpszTelephone);
            break;
          }
        }

        /* Next attribute */
        a *= 2;
      }
    } else
    {
      /* Compare extended attributes, if any */
      a = 1;
      while ((a <= MAX_CLASS) && !iCmp)
      {
        /* If this attribute is requested for display, then get the attribute */
        if (dwOrder[i] & a)
        {
          /* Get the class data for both entries if they exist */
          lpClassData1 = GetClassData(lpEntry1, a);
          lpClassData2 = GetClassData(lpEntry2, a);

          /* If any do not exist, then we must get the default value if one exists */
          if (!lpClassData1 || !lpClassData2)
          {
            /* If both are NULL, they equal */
            if (!(lpClassData1 && lpClassData2))
            {
              /* Get the default value */
              lpClass = GetClass(lpClassRoot, a);
              if (lpClass) {
                if (lpClass->dwFlags & CLASS_DEFAULTVALUE) {
                  lpDefaultValue = GetClassValue(lpClass, lpClass->idDefaultValue);
                  if (lpDefaultValue)
                  {
                    /* Now that we have the default value, compare the NULL to the other */

                    /* Process this if the first class data is missing */
                    if (!lpClassData1) {
                      if ((lpClassData2->dwFlags & CLASSDATA_OTHER) || (bExt[i] == TRUE))
                        iCmp = CompareStrings(lpDefaultValue->lpszValue, lpClassData2->lpszValue);
                      else {
                        if (lpDefaultValue->idValue > lpClassData2->dwValue)
                          iCmp = 1;
                        else {
                          if (lpDefaultValue->idValue < lpClassData2->dwValue)
                            iCmp = -1;
                        }
                      }
                    }

                    /* Process this if the second class data is missing */
                    if (!lpClassData2) {
                      if ((lpClassData1->dwFlags & CLASSDATA_OTHER) || (bExt[i] == TRUE))
                        iCmp = CompareStrings(lpClassData1->lpszValue, lpDefaultValue->lpszValue);
                      else {
                        if (lpClassData1->dwValue > lpDefaultValue->idValue)
                          iCmp = 1;
                        else {
                          if (lpClassData2->dwValue < lpDefaultValue->idValue)
                            iCmp = -1;
                        }
                      }
                    }
                  }


                  /* If one does not exist and the other does, the one without is less */
                  if (!lpClassData1 && lpClassData2)
                    iCmp = -1;
                  if (lpClassData1 && !lpClassData2)
                    iCmp = 1;
                }
              }
            }
          } else
          {
            /* If any of the two have their own string values or if string */
            /* mode was requested, then use string mode                    */
            if ((lpClassData1->dwFlags & CLASSDATA_OTHER) ||
                (lpClassData2->dwFlags & CLASSDATA_OTHER) ||
                (bExt[i] == TRUE))
            {
              /* If a string value does not exist in either of them, which */
              /* should never happen as long as DirList2 edits the file,   */
              /* assume they are equal                                     */
              if ((lpClassData1->dwFlags & CLASSDATA_STRING) &&
                  (lpClassData2->dwFlags & CLASSDATA_STRING))
              {
                /* Both strings exist, so let's compare them :-) */
                iCmp = CompareStrings(lpClassData1->lpszValue, lpClassData2->lpszValue);
              }
            } else
            {
              /* Sort using ordinal values */
              if (lpClassData1->dwValue > lpClassData2->dwValue)
                iCmp = 1;
              else {
                if (lpClassData1->dwValue < lpClassData2->dwValue)
                  iCmp = -1;
              }
            }
          }
        }

        /* Increment safely */
        if (a == MAX_CLASS)
          a++;
        else
          a *= 2;
      }
    }

    /* Next sort layer */
    i++;
  }

  /* If the comparision is not equal, then adjust the answer if descending */
  if (iCmp) {
    if (bDesc[i - 1])
      iCmp = -iCmp;
  }

  /* Return the final result */
  return iCmp;
}

/* Add a directory entry to a sort table, given the sort method and order layers */
BOOL AddSortEntry(LPDIRLIST_SORTENTRY *lpRoot, LPDIRLIST_SORTENTRY *lpPrev, LPDIRLIST_SORTENTRY *lpCurrent, LPDIRLIST_ENTRY lpEntry, LPDIRLIST_CLASS lpClass, LPDWORD dwOrder, DWORD nLayers, BOOL *bExt, BOOL *bDesc)
{
  /* Add an entry to the linked list lpRoot and return the new lpRoot */
  LPDIRLIST_SORTENTRY lpNew;
  BOOL found, cmpResult;

  /* Check parameters */
  if (!*lpCurrent) {
    *lpCurrent = *lpRoot;
    *lpPrev = NULL;
  } else
  {
    if (*lpRoot)
    {
      /* If the name is less than the current, then start from the beginning */
      if (CompareEntries(lpEntry, (*lpCurrent)->lpDirEntry, lpClass, dwOrder, nLayers, bExt, bDesc) < 0) {
        *lpCurrent = *lpRoot;
        *lpPrev = NULL;
      }
    }
  }

  /* Prepare to find the correct placement */
  found = FALSE;

  while (!found && (*lpCurrent != NULL)) {
    if ((cmpResult = CompareEntries(lpEntry, (*lpCurrent)->lpDirEntry, lpClass, dwOrder, nLayers, bExt, bDesc)) < 0)
      found = TRUE;
    else {
      if (cmpResult == 0)
        found = TRUE;
      else {
        *lpPrev = *lpCurrent;
        *lpCurrent = (LPDIRLIST_SORTENTRY) (*lpCurrent)->next;
      }
    }
  }

  /* Create a new sort entry and fill out its structure */
  lpNew = (LPDIRLIST_SORTENTRY) malloc(sizeof(DIRLIST_SORTENTRY));
  lpNew->lpDirEntry = lpEntry;
  lpNew->dwFilePos = 0;

  /* Add this to the list */
  if (!*lpPrev) {
    lpNew->next = (LPBYTE) *lpRoot;
    *lpRoot = lpNew;
    *lpCurrent = lpNew;
  } else {
    lpNew->next = (LPBYTE) *lpCurrent;
    (*lpPrev)->next = (LPBYTE) lpNew;
    *lpCurrent = lpNew;
  }

  /* Entry created; return TRUE */
  return TRUE;
}

BOOL CreateSortTable(LPDIRLIST_SORTTABLE lpTable, LPDIRLIST_ENTRY lpRoot, LPDIRLIST_CLASS lpClass)
{
  /* Create a sort table using the given order */
  LPDIRLIST_ENTRY lpCurEntry;
  LPDIRLIST_SORTENTRY lpPrev = NULL;
  LPDIRLIST_SORTENTRY lpCurrent = NULL;
  LPTSTR lpszOrder = malloc(strlen(lpTable->lpszOrder) + 1);
  LPDWORD dwOrder = (LPDWORD) malloc(1);
  BOOL *bExt = (BOOL *) malloc(1);
  BOOL *bDesc = (BOOL *) malloc(1);
  LPTSTR lpszCurrent;
  DWORD nLayers = 0;

  /* Parse the lpszOrder string into order arrays */
  strcpy(lpszOrder, lpTable->lpszOrder);
  lpszCurrent = strtok(lpszOrder, ",");
  while (lpszCurrent)
  {
    /* Get the sort type */
    dwOrder = realloc(dwOrder, sizeof(DWORD) * (++nLayers));
    dwOrder[nLayers - 1] = strtoul(lpszCurrent, &lpszCurrent, 10);

    /* Get the status on whether or not this is for string extended data (number ends in "s") */
    bExt = realloc(bExt, sizeof(BOOL) * (nLayers));
    if ((lpszCurrent[0] == 's') || (lpszCurrent[0] == 'd')) {
      if ((lpszCurrent[0] == 's') || (lpszCurrent[1] == 's'))
        bExt[nLayers - 1] = TRUE;
      else
        bExt[nLayers - 1] = FALSE;
      if (lpszCurrent[0] == 's')
        lpszCurrent = &(lpszCurrent[1]);
    } else
      bExt[nLayers - 1] = FALSE;

    /* Get the status on whether or not this is for ordinal extended data (number ends in "o") */
    if ((lpszCurrent[0] == 'o') || (lpszCurrent[0] == 'd')) {
      if ((lpszCurrent[0] == 'o') || (lpszCurrent[1] == 'o'))
        bExt[nLayers - 1] = TRUE + 1;
      if (lpszCurrent[0] == 'o')
        lpszCurrent = &(lpszCurrent[1]);
    }

    /* Get the status on whether or not to sort descending (number ends in "d") */
    bDesc = realloc(bDesc, sizeof(BOOL) * (nLayers));
    if (lpszCurrent[0] == 'd')
      bDesc[nLayers - 1] = TRUE;
    else
      bDesc[nLayers - 1] = FALSE;

    /* Next token */
    lpszCurrent = strtok(NULL, ",");
  }
  free(lpszOrder);

  /* If we have no layers, return FALSE (using this sort number uses the default sort */
  /* order, which is by ascending username)                                           */
  if (!nLayers)
    return FALSE;

  /* Fill out the sort table structure */
  lpTable->nEntries = 0;
  lpTable->lpSortEntries = NULL;

  /* Copy the entries into the sort table, sorting at the time of addition */
  lpCurEntry = lpRoot;
  while (lpCurEntry) {
    AddSortEntry(&(lpTable->lpSortEntries), &lpPrev, &lpCurrent, lpCurEntry, lpClass, dwOrder, nLayers, bExt, bDesc);
    lpTable->nEntries++;
    lpCurEntry = (LPDIRLIST_ENTRY) lpCurEntry->next;
  }

  /* Free the memory */
  free(dwOrder);
  free(bExt);
  free(bDesc);

  /* We did it :-) */
  return TRUE;
}

/* Search the given string with the access plan and array position */
BOOL SearchString(LPCTSTR lpszValue, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwCurrent)
{
  /* If this is an exact match, then simply call CompareStrings */
  BOOL bMatch;
  if (!lpszValue)
    return FALSE;

  /* Dispatch the search */
  switch (lpPlan->fSearchExact[dwCurrent] & ~SEARCH_ESCAPEMASK)
  {

  case SEARCH_EXACT :
    {
      /* If this is an exact search, then simply call CompareStrings */
      bMatch = CompareStrings(lpszValue, lpPlan->lpszSearch[dwCurrent]) == 0;
      return (lpPlan->bSearchNot[dwCurrent])? !bMatch: bMatch;
    }

  case SEARCH_STARTSWITH :
    {
      /* If this is a "starts with" search, then simply call SearchStartingStrings */
      bMatch = SearchStartingStrings(lpszValue, lpPlan->lpszSearch[dwCurrent]);
      return (lpPlan->bSearchNot[dwCurrent])? !bMatch: bMatch;
    }

  case SEARCH_ENDSWITH :
    {
      /* If this is an "ends with" search, then simply call SearchEndingStrings */
      bMatch = SearchEndingStrings(lpszValue, lpPlan->lpszSearch[dwCurrent]);
      return (lpPlan->bSearchNot[dwCurrent])? !bMatch: bMatch;
    }

  case SEARCH_LIKE :
    {
      /* If this is a LIKE search, then simply call SearchLIKEStrings */
      bMatch = SearchLIKEStrings(lpszValue, lpPlan->lpszSearch[dwCurrent], (lpPlan->fSearchExact[dwCurrent] & SEARCH_ESCAPEMASK) / 256);
      return (lpPlan->bSearchNot[dwCurrent])? !bMatch: bMatch;
    }

  case SEARCH_STRICTSTARTSWITH :
    {
      /* If this is a strict "starts with" search, then simply call SearchStrictStartingStrings */
      bMatch = SearchStrictStartingStrings(lpszValue, lpPlan->lpszSearch[dwCurrent]);
      return (lpPlan->bSearchNot[dwCurrent])? !bMatch: bMatch;
    }
  }

  /* This is a substring search if the low and high ranges equal (same string allocation) */
  if (lpPlan->lpszRangeLow[dwCurrent] == lpPlan->lpszRangeHigh[dwCurrent]) {
    bMatch = FindSubstring(lpszValue, lpPlan->lpszSearch[dwCurrent]);
    return (lpPlan->bSearchNot[dwCurrent])? !bMatch: bMatch;
  } else
  {
    /* Range search -- determine how well this value fits the criteria */
    int iCmpLow, iCmpHigh;
    if (lpPlan->lpszRangeLow[dwCurrent][0] == 0)
      iCmpLow = 1;
    else
      iCmpLow = CompareStrings(lpszValue, lpPlan->lpszRangeLow[dwCurrent]);
    if (lpPlan->lpszRangeHigh[dwCurrent][0] == 0)
      iCmpHigh = -1;
    else
      iCmpHigh = CompareStrings(lpszValue, lpPlan->lpszRangeHigh[dwCurrent]);

    /* Adjust for inclusive searches */
    if (lpPlan->bRangeLowInclusive[dwCurrent] && !iCmpLow)
      iCmpLow = 1;
    if (lpPlan->bRangeHighInclusive[dwCurrent] && !iCmpHigh)
      iCmpHigh = -1;

    /* If lpszValue is higher than the low and is lower than the high, then return true :-) */
    if ((iCmpLow > 0) && (iCmpHigh < 0))
      return (lpPlan->bSearchNot[dwCurrent])? FALSE: TRUE;
    else
      return (lpPlan->bSearchNot[dwCurrent])? TRUE: FALSE;
  }
}

/* Search the given DWORD with the access plan and array position */
BOOL SearchOrdinal(DWORD dwValue, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwCurrent)
{
  /* If this is any type of match, then simply compare them */
  if (lpPlan->lpszRangeLow[dwCurrent] == lpPlan->lpszRangeHigh[dwCurrent]) {
    DWORD bMatch = dwValue == strtoul(lpPlan->lpszSearch[dwCurrent], NULL, 10);
    return (lpPlan->bSearchNot[dwCurrent])? !bMatch: bMatch;
  } else
  {
    /* Range search -- determine how well this value fits the criteria */
    DWORD dwLow = strtoul(lpPlan->lpszRangeLow[dwCurrent], NULL, 10);
    DWORD dwHigh = strtoul(lpPlan->lpszRangeHigh[dwCurrent], NULL, 10);
    int iCmpLow, iCmpHigh;
    if (dwValue < dwLow)
      iCmpLow = -1;
    else {
      if (dwValue == dwLow)
        iCmpLow = 0;
      else
        iCmpLow = 1;
    }
    if (dwValue < dwHigh)
      iCmpHigh = -1;
    else {
      if (dwValue == dwHigh)
        iCmpHigh = 0;
      else
        iCmpHigh = 1;
    }

    /* If the high value is zero, then assume infinity */
    if (dwHigh == 0)
      iCmpHigh = -1;

    /* Adjust for inclusive searches */
    if (lpPlan->bRangeLowInclusive[dwCurrent] && !iCmpLow)
      iCmpLow = 1;
    if (lpPlan->bRangeHighInclusive[dwCurrent] && !iCmpHigh)
      iCmpHigh = -1;

    /* If dwValue is higher than the low and is lower than the high, then return true :-) */
    if ((iCmpLow > 0) && (iCmpHigh < 0))
      return (lpPlan->bSearchNot[dwCurrent])? FALSE: TRUE;
    else
      return (lpPlan->bSearchNot[dwCurrent])? TRUE: FALSE;
  }
}

/* Compare a condition to an entry */
BOOL SearchEntry(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_CLASS lpClass, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwCurrent)
{
  /* Check the username and personal names, paying attention to order :-) */
  BOOL bFound = FALSE;
  DWORD i;
  if (lpPlan->dwSearchFields[dwCurrent] & DISPLAY_USERNAME)
    bFound = SearchString(lpDirEntry->lpszUserName, lpPlan, dwCurrent);
  if (!bFound && (lpPlan->dwSearchFields[dwCurrent] & DISPLAY_FIRSTNAME))
    bFound = SearchString(lpDirEntry->lpszFirstName, lpPlan, dwCurrent);
  if (!bFound && (lpPlan->dwSearchFields[dwCurrent] & DISPLAY_LASTNAME))
    bFound = SearchString(lpDirEntry->lpszLastName, lpPlan, dwCurrent);

  /* Check the other attributes */
  i = DISPLAY_MIN;
  while (!bFound && (i <= DISPLAY_MAX)) {
    if (i & lpPlan->dwSearchFields[dwCurrent])
    {
      /* Compare the appropriate substring */
      switch (i) {

      case DISPLAY_FLAGS :

        /* Search the flags taking everything into account, including admin mode */
        if (lpPlan->bAdminMode || (!lpPlan->bAdminMode && !(lpDirEntry->dwFlags & FLAGS_EXCLUDE))) {
          if (lpPlan->fSearchExact[dwCurrent]) {
            if (lpPlan->bSearchNot[dwCurrent])
              bFound = lpDirEntry->dwFlags != strtoul(lpPlan->lpszSearch[dwCurrent], NULL, 10);
            else
              bFound = lpDirEntry->dwFlags == strtoul(lpPlan->lpszSearch[dwCurrent], NULL, 10);
          } else {
            if (lpPlan->bSearchNot[dwCurrent])
              bFound = !(lpDirEntry->dwFlags & strtoul(lpPlan->lpszSearch[dwCurrent], NULL, 10));
            else
              bFound = lpDirEntry->dwFlags & strtoul(lpPlan->lpszSearch[dwCurrent], NULL, 10);
          }
        }
        break;

      case DISPLAY_WEBSITE :
        bFound = SearchString(lpDirEntry->lpszWebSite, lpPlan, dwCurrent);
        break;

      case DISPLAY_EMAIL :
        bFound = SearchString(lpDirEntry->lpszEMail, lpPlan, dwCurrent);
        break;

      case DISPLAY_TELEPHONE :
        bFound = SearchString(lpDirEntry->lpszTelephone, lpPlan, dwCurrent);
        break;

      }
    }

    /* Next attribute */
    i *= 2;
  }

  /* Check the extended attributes */
  i = 1;
  while (!bFound && (i <= MAX_CLASS))
  {
    /* If this attribute is in the search, then fetch it */
    if (lpPlan->dwExtSearchFields[dwCurrent] & i) {
      LPDIRLIST_CLASSDATA lpClassData = GetClassData(lpDirEntry, i);

      /* Search the ordinal or the string, depending on the request */
      if (lpClassData) {
        if (lpPlan->dwExtSearchOrdinals[dwCurrent] & i)
          bFound = SearchOrdinal(lpClassData->dwValue, lpPlan, dwCurrent);
          /*            if (lpClassData->dwValue == strtoul(lpPlan->lpszSearch[dwCurrent], NULL, 10))
          bFound = (lpPlan->bSearchNot[dwCurrent])? FALSE: TRUE;
          else
        bFound = (lpPlan->bSearchNot[dwCurrent])? TRUE: FALSE;*/
        else
          bFound = SearchString(lpClassData->lpszValue, lpPlan, dwCurrent);
      } else
      {
        /* Search the default value if available */
        LPDIRLIST_CLASS lpCurrentClass = GetClass(lpClass, i);
        if (lpCurrentClass)
        {
          /* See if it has a default value */
          if (lpCurrentClass->dwFlags & CLASS_DEFAULTVALUE)
          {
            if (lpPlan->dwExtSearchOrdinals[dwCurrent] & i)
              bFound = SearchOrdinal(lpCurrentClass->idDefaultValue, lpPlan, dwCurrent);
            else
            {
              /* Get the default value name and compare it to the search */
              LPDIRLIST_CLASSVALUE lpClassValue = GetClassValue(lpClass, lpCurrentClass->idDefaultValue);
              if (lpClassValue)
                bFound = SearchString(lpClassValue->lpszValue, lpPlan, dwCurrent);
            }
          }
        }
      }
    }

    /* Increment safely */
    if (i == MAX_CLASS)
      i++;
    else
      i *= 2;
  }

  /* Return our findings */
  return bFound;
}

/* Search an entry with the given criteria */
BOOL SearchFields(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_CLASS lpClass, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwLevel, LPDWORD lpdwCurrent)
{
  /* Process each AND in sequence and then loop through the ORs while recursing on parenthesis */
  BOOL bOrFound = FALSE, bAndFound, bAndEnd, bLevelEnd;
  while (*lpdwCurrent < lpPlan->nSearches)
  {
    /* If this is an AND condition, process the sequence as one */
    if (lpPlan->bSearchAnd[*lpdwCurrent]) {
      bAndFound = TRUE;
      bAndEnd = FALSE;
      bLevelEnd = FALSE;
      while ((*lpdwCurrent < lpPlan->nSearches)? (lpPlan->bSearchAnd[*lpdwCurrent] || bAndEnd): FALSE)
      {
        /* If this condition is on a new level, then recurse */
        if (lpPlan->dwSearchLevel[*lpdwCurrent] > dwLevel) {
          bAndFound = bAndFound && SearchFields(lpDirEntry, lpClass, lpPlan, dwLevel + 1, lpdwCurrent);
          bLevelEnd = TRUE;
        } else if (lpPlan->dwSearchLevel[*lpdwCurrent] < dwLevel) {
          (*lpdwCurrent)--;
          return bOrFound || bAndFound;
        } else if (bAndFound)
          bAndFound = SearchEntry(lpDirEntry, lpClass, lpPlan, *lpdwCurrent);

        /* If this is a proper level ending, then return our result after decrementing the level */
        if (lpPlan->dwSearchLevelEnd[*lpdwCurrent]) {
          lpPlan->dwSearchLevelEnd[*lpdwCurrent]--;
          return bOrFound || bAndFound;
        }

        /* Next condition */
        (*lpdwCurrent)++;

        /* If this is not AND, then we're at the end of the AND string of conditions; process it */
        if (!bAndEnd && !bLevelEnd) {
          if (*lpdwCurrent < lpPlan->nSearches) {
            if (!lpPlan->bSearchAnd[*lpdwCurrent])
              bAndEnd = TRUE;
          }
        } else
          bAndEnd = FALSE;
      }

      /* Integrate our results into bOrFound */
      bOrFound = bOrFound || bAndFound;
    }

    /* If we are not at the end, then process this */
    if ((*lpdwCurrent < lpPlan->nSearches) && !lpPlan->bSearchAnd[*lpdwCurrent])
    {
      /* If this condition is on a new level, then recurse */
      if (lpPlan->dwSearchLevel[*lpdwCurrent] > dwLevel)
        bOrFound = bOrFound || SearchFields(lpDirEntry, lpClass, lpPlan, dwLevel + 1, lpdwCurrent);
      else if (lpPlan->dwSearchLevel[*lpdwCurrent] < dwLevel) {
        (*lpdwCurrent)--;
        return bOrFound;
      } else if (!bOrFound)
        bOrFound = SearchEntry(lpDirEntry, lpClass, lpPlan, *lpdwCurrent);

      /* If this is a proper level ending, then return our result after decrementing the level */
      if (lpPlan->dwSearchLevelEnd[*lpdwCurrent]) {
        lpPlan->dwSearchLevelEnd[*lpdwCurrent]--;
        return bOrFound;
      }

      /* Next condition */
      (*lpdwCurrent)++;
    }
  }

  /* Decrease the current condition and return our findings */
  (*lpdwCurrent)--;
  return bOrFound;
}

/* Search an entry with the given criteria */
BOOL SearchFieldsOld(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_CLASS lpClass, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwLevel, LPDWORD lpdwCurrent)
{
  /* Loop while searches exist */
  BOOL bLastFound = FALSE;
  BOOL bFound = FALSE;
  DWORD i;
  while ((*lpdwCurrent < lpPlan->nSearches) && !bFound)
  {
    /* If the search level is higher than this function call's current, then call us again */
    if (lpPlan->dwSearchLevel[*lpdwCurrent] > dwLevel) {
      bFound = SearchFields(lpDirEntry, lpClass, lpPlan, dwLevel + 1, lpdwCurrent);

      /* Ensure that we're still within the bounds of the array */
      if (*lpdwCurrent >= lpPlan->nSearches)
        return bFound;

      /* If we still have search level ends, then exit this level */
      if (lpPlan->dwSearchLevelEnd[*lpdwCurrent]) {
        lpPlan->dwSearchLevelEnd[*lpdwCurrent]--;
        return bFound;
      }

      /* If the next entry is a lower level, then return bFound now */
      if (*lpdwCurrent + 1 < lpPlan->nSearches) {
        if (lpPlan->dwSearchLevel[*lpdwCurrent + 1] < dwLevel)
          return bFound;
      }
    } else
    {
      /* If the new level is less, then decrease lpdwCurrent and return bLastFound */
      if (lpPlan->dwSearchLevel[*lpdwCurrent] < dwLevel) {
        (*lpdwCurrent)--;
        return bLastFound;
      }

      /* Check the username and personal names, paying attention to order :-) */
      if (lpPlan->dwSearchFields[*lpdwCurrent] & DISPLAY_USERNAME)
        bFound = SearchString(lpDirEntry->lpszUserName, lpPlan, *lpdwCurrent);
      if (!bFound && ((lpPlan->dwSearchFields[*lpdwCurrent] & DISPLAY_LASTNAMEFIRST) == DISPLAY_LASTNAMEFIRST)) {
        if (!(bFound = SearchString(lpDirEntry->lpszLastName, lpPlan, *lpdwCurrent)))
          bFound = SearchString(lpDirEntry->lpszFirstName, lpPlan, *lpdwCurrent);
      } else {
        if (!bFound && (lpPlan->dwSearchFields[*lpdwCurrent] & DISPLAY_FIRSTNAME))
          bFound = SearchString(lpDirEntry->lpszFirstName, lpPlan, *lpdwCurrent);
        if (!bFound && (lpPlan->dwSearchFields[*lpdwCurrent] & DISPLAY_LASTNAME))
          bFound = SearchString(lpDirEntry->lpszLastName, lpPlan, *lpdwCurrent);
      }

      /* Check the other attributes */
      i = DISPLAY_MIN;
      while (!bFound && (i <= DISPLAY_MAX)) {
        if (i & lpPlan->dwSearchFields[*lpdwCurrent])
        {
          /* Compare the appropriate substring */
          switch (i) {

          case DISPLAY_FLAGS :

            /* Search the flags taking everything into account, including admin mode */
            if (lpPlan->bAdminMode || (!lpPlan->bAdminMode && !(lpDirEntry->dwFlags & FLAGS_EXCLUDE))) {
              if (lpPlan->fSearchExact[*lpdwCurrent]) {
                if (lpPlan->bSearchNot[*lpdwCurrent])
                  bFound = lpDirEntry->dwFlags != strtoul(lpPlan->lpszSearch[*lpdwCurrent], NULL, 10);
                else
                  bFound = lpDirEntry->dwFlags == strtoul(lpPlan->lpszSearch[*lpdwCurrent], NULL, 10);
              } else {
                if (lpPlan->bSearchNot[*lpdwCurrent])
                  bFound = !(lpDirEntry->dwFlags & strtoul(lpPlan->lpszSearch[*lpdwCurrent], NULL, 10));
                else
                  bFound = lpDirEntry->dwFlags & strtoul(lpPlan->lpszSearch[*lpdwCurrent], NULL, 10);
              }
            }
            break;

          case DISPLAY_WEBSITE :
            bFound = SearchString(lpDirEntry->lpszWebSite, lpPlan, *lpdwCurrent);
            break;

          case DISPLAY_EMAIL :
            bFound = SearchString(lpDirEntry->lpszEMail, lpPlan, *lpdwCurrent);
            break;

          case DISPLAY_TELEPHONE :
            bFound = SearchString(lpDirEntry->lpszTelephone, lpPlan, *lpdwCurrent);
            break;

          }
        }

        /* Next attribute */
        i *= 2;
      }

      /* Check the extended attributes */
      i = 1;
      while (!bFound && (i <= MAX_CLASS))
      {
        /* If this attribute is in the search, then fetch it */
        if (lpPlan->dwExtSearchFields[*lpdwCurrent] & i) {
          LPDIRLIST_CLASSDATA lpClassData = GetClassData(lpDirEntry, i);

          /* Search the ordinal or the string, depending on the request */
          if (lpClassData) {
            if (lpPlan->dwExtSearchOrdinals[*lpdwCurrent] & i)
              bFound = SearchOrdinal(lpClassData->dwValue, lpPlan, *lpdwCurrent);
              /*            if (lpClassData->dwValue == strtoul(lpPlan->lpszSearch[*lpdwCurrent], NULL, 10))
              bFound = (lpPlan->bSearchNot[*lpdwCurrent])? FALSE: TRUE;
              else
            bFound = (lpPlan->bSearchNot[*lpdwCurrent])? TRUE: FALSE;*/
            else
              bFound = SearchString(lpClassData->lpszValue, lpPlan, *lpdwCurrent);
          } else
          {
            /* Search the default value if available */
            LPDIRLIST_CLASS lpCurrentClass = GetClass(lpClass, i);
            if (lpCurrentClass)
            {
              /* See if it has a default value */
              if (lpCurrentClass->dwFlags & CLASS_DEFAULTVALUE)
              {
                if (lpPlan->dwExtSearchOrdinals[*lpdwCurrent] & i)
                  bFound = SearchOrdinal(lpCurrentClass->idDefaultValue, lpPlan, *lpdwCurrent);
                else
                {
                  /* Get the default value name and compare it to the search */
                  LPDIRLIST_CLASSVALUE lpClassValue = GetClassValue(lpClass, lpCurrentClass->idDefaultValue);
                  if (lpClassValue)
                    bFound = SearchString(lpClassValue->lpszValue, lpPlan, *lpdwCurrent);
                }
              }
            }
          }
        }

        /* Increment safely */
        if (i == MAX_CLASS)
          i++;
        else
          i *= 2;
      }
    }

    /* If this is the end of a level, then return bFound now */
    if (lpPlan->dwSearchLevelEnd[*lpdwCurrent]) {
      lpPlan->dwSearchLevelEnd[*lpdwCurrent]--;
      return bFound;
    }

    /* If this iteration didn't find anything, then set bLastFound to FALSE */
    if (!bFound)
      bLastFound = FALSE;
    else
      bLastFound = TRUE;

    /* Processing for AND */
    if (lpPlan->bSearchAnd[*lpdwCurrent])
    {
      /* If we found this one, then set ourselves up for the next one; otherwise, exit */
      if (bFound) {
        bLastFound = TRUE;
        bFound = FALSE;
      } else
      {
        /* Find the search level's ending and then return FALSE */
        i = *lpdwCurrent;  /* removed "+ 1" */
        while (i < lpPlan->nSearches) {
          if (dwLevel > lpPlan->dwSearchLevel[i]) {
            *lpdwCurrent = i - 1;
            return FALSE;
          }
          if (lpPlan->dwSearchLevelEnd[i]) {
            *lpdwCurrent = i;
            return FALSE;
          }
          i++;
        }

        /* Added as a temporary bugfix */
        return FALSE;
      }
    }

    /* Next search array entry */
    (*lpdwCurrent)++;
  }

  /* Seek to the end of the current level */
  i = *lpdwCurrent; /* removed "+ 1" */
  while (i < lpPlan->nSearches) {
    if (dwLevel != lpPlan->dwSearchLevel[i]) {
      *lpdwCurrent = i - 1;
      return bFound;
    }
    if (lpPlan->dwSearchLevelEnd[i]) {
      lpPlan->dwSearchLevelEnd[i]--;
      *lpdwCurrent = i;
      return bFound;
    }
    i++;
  }

  /* Return whether or not this entry meets the search criteria */
  return bFound;
}

/* See if this user falls within the requested criteria */
BOOL CheckRelevance(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_CLASS lpClass, LPDIRLIST_ACCESSPLAN lpPlan)
{
  LPDIRLIST_CLASSDATA lpClassData;
  BOOL noDisplay = FALSE;
  DWORD i;

  /* Check the permissions on this entry before attempting to write it */
  if (!lpPlan->bAdminMode)
  {
    /* This user MUST be hidden from all views, regardless of display mode */
    if (lpDirEntry->dwFlags & FLAGS_EXCLUDE)
      return FALSE;
  }

  /* Go through the attributes and see if this user has been locked out */
  if ((lpDirEntry->lpszUserName[0] == 0) && (lpPlan->dwXFields & DISPLAY_USERNAME))
    noDisplay = TRUE;
  if ((lpDirEntry->lpszFirstName[0] == 0) && (lpPlan->dwXFields & DISPLAY_FIRSTNAME))
    noDisplay = TRUE;
  if ((lpDirEntry->lpszLastName[0] == 0) && (lpPlan->dwXFields & DISPLAY_LASTNAME))
    noDisplay = TRUE;
  i = DISPLAY_MIN;
  while ((i <= DISPLAY_MAX) && !noDisplay)
  {
    /* Breakout for processing this attribute */
    if (lpPlan->dwXFields & i)
    {
      switch (i) {

      case DISPLAY_FLAGS :
        if (!lpDirEntry->dwFlags)
          noDisplay = TRUE;
        break;

      case DISPLAY_WEBSITE :
        if (lpDirEntry->dwFlags & FLAGS_NOWEBSITE)
          noDisplay = TRUE;
        break;

      case DISPLAY_EMAIL :
        if (lpDirEntry->dwFlags & FLAGS_NOEMAIL)
          noDisplay = TRUE;
        break;

      case DISPLAY_TELEPHONE :
        if (lpDirEntry->dwFlags & FLAGS_NOTELEPHONE)
          noDisplay = TRUE;
        break;
      }
    }

    /* Next attribute */
    i *= 2;
  }
  if (noDisplay)
    return FALSE;

  /* Go through the extended attributes, if any, to see if we should lock out */
  i = 1;
  while ((i <= MAX_CLASS) && !noDisplay)
  {
    if (lpPlan->dwXExtFields & i)
    {
      /* Get the class data for this entry */
      lpClassData = GetClassData(lpDirEntry, i);

      /* Process the display of the value */
      if (!lpClassData)
      {
        /* If this attribute is set to exclude the user when not present, then */
        /* signal the noDisplay flag */
        noDisplay = TRUE;
      }
    }

    /* Increment safely */
    if (i == MAX_CLASS)
      i++;
    else
      i *= 2;
  }

  /* If we have a search variable, then compare them to their field(s) */
  if (!noDisplay && lpPlan->nSearches) {
    DWORD dwCurrent = 0;
    noDisplay = !SearchFields(lpDirEntry, lpClass, lpPlan, 0, &dwCurrent);
  }

  /* Return whether or not we should display this user */
  return !noDisplay;
}

BOOL FetchEntry(LPDIRLIST_ENTRY *lpRoot, LPDIRLIST_ENTRY *lpPrev, LPDIRLIST_ENTRY *lpCurrent, LPTSTR lpszUserName, BOOL bCreate)
{
  /* Add an entry to the linked list lpRoot and return the new lpRoot */
  BOOL found, cmpResult;

  /* Check parameters */
  if (!*lpCurrent) {
    *lpCurrent = *lpRoot;
    *lpPrev = NULL;
  } else
  {
    if (*lpRoot)
    {
      /* If the name is less than the current, then start from the beginning */
      if (CompareStrings(lpszUserName, (*lpCurrent)->lpszUserName) < 0) {
        *lpCurrent = *lpRoot;
        *lpPrev = NULL;
      }
    }
  }

  /* Prepare to find the correct placement */
  found = FALSE;

  while (!found && (*lpCurrent != NULL)) {
    if ((cmpResult = CompareStrings(lpszUserName, (*lpCurrent)->lpszUserName)) < 0)
      found = TRUE + 1;
    else {
      if (cmpResult == 0)
        found = TRUE;
      else {
        *lpPrev = *lpCurrent;
        *lpCurrent = (LPDIRLIST_ENTRY) (*lpCurrent)->next;
      }
    }
  }

  /* New entry conveniently added in ascending alphabetical order according to username */
  if ((found != TRUE) && bCreate)
  {
    /* Allocate memory and set the variables */
    LPDIRLIST_ENTRY lpNew = (LPDIRLIST_ENTRY) malloc(sizeof(DIRLIST_ENTRY));
    memset(lpNew, (char) NULL, sizeof(DIRLIST_ENTRY));
    lpNew->lpszUserName = (LPTSTR) malloc(strlen(lpszUserName) + 1);
    strcpy(lpNew->lpszUserName, lpszUserName);

    /* Create zero-length strings */
    lpNew->lpszFirstName = malloc(1);
    lpNew->lpszLastName = malloc(1);
    lpNew->lpszWebSite = malloc(1);
    lpNew->lpszEMail = malloc(1);
    lpNew->lpszTelephone = malloc(1);
    lpNew->lpszFirstName[0] = 0;
    lpNew->lpszLastName[0] = 0;
    lpNew->lpszWebSite[0] = 0;
    lpNew->lpszEMail[0] = 0;
    lpNew->lpszTelephone[0] = 0;

    /* Add this to the list */
    if (!*lpPrev) {
      lpNew->next = (LPBYTE) *lpRoot;
      *lpRoot = lpNew;
      *lpCurrent = lpNew;
    } else {
      lpNew->next = (LPBYTE) *lpCurrent;
      (*lpPrev)->next = (LPBYTE) lpNew;
      *lpCurrent = lpNew;
    }

    /* Entry created; return TRUE + 1*/
    return TRUE + 1;
  }

  /* If the entry was found, return TRUE; otherwise, return FALSE */
  if (found == 1)
    return TRUE;
  else
    return FALSE;
}

/* Loads a directory into list while sorting it (by username) and returns its length */
DWORD LoadDirectory(LPCTSTR dir, LPDIRLIST_ENTRY *lpDirEntry, LPCTSTR lpszEMailDomain)
{
  /* List the directory */
  LPDIRLIST_ENTRY lpCurrent = NULL, lpPrev = NULL;
  FILEFIND_DATA ffData;
  BOOL more;
  BOOL bNewEntry;
  LPTSTR lpszRealName;
  DWORD num = 0;
  DWORD lEMailDomain = strlen(lpszEMailDomain);
  more = FindFirst(dir, NULL, &ffData, FALSE);
  lpCurrent = *lpDirEntry;

  /* If this directory is not empty, then process it */
  if (more) {
    do {
      if ((strcmp(ffData.ff_name, ".") != 0) && (strcmp(ffData.ff_name, "..") != 0))
      {
        /* Increment num and fetch/create the entry for this user */
        num++;
        bNewEntry = (FetchEntry(lpDirEntry, &lpPrev, &lpCurrent, ffData.ff_name, TRUE) == TRUE + 1)? TRUE: FALSE;

        /* This directory exists, so unset the FLAGS_NOHOMEDIR if set */
        lpCurrent->dwFlags &= ~FLAGS_NOHOMEDIR;

        /* Get the real name, if we're supposed to */
        if (!(lpCurrent->dwFlags & FLAGS_NONAMECHECK)) {
          LPTSTR lpszLastName;

/*          pw = getpwnam(ffData.ff_name);
          if (pw != NULL) {
            if ((lpszRealName = strtok(pw->pw_gecos, ",")) == NULL)
              lpszRealName = pw->pw_gecos;
          } else */

          if (!GetPasswdEntry(ffData.ff_name, &lpszRealName))
          {
            /* This directory entry does not have an associated user! */
            lpszRealName = DEFAULT_USERNAME;
            lpCurrent->dwFlags |= FLAGS_EXCLUDE | FLAGS_NOWEBSITE | FLAGS_NOEMAIL | FLAGS_NOTELEPHONE | FLAGS_NONAMECHECK | FLAGS_NOWEBCHECK | FLAGS_NOWEBSITE;
          }

          /* Copy the real name (free any existing name) */

          /* Allocate our own lpszRealName (use lpszLastName as temp variable) */
          lpszLastName = lpszRealName;
          lpszRealName = malloc(strlen(lpszLastName) + 1);
          strcpy(lpszRealName, lpszLastName);

          /* Free the current names if they have values */
          if (lpCurrent->lpszFirstName)
            free(lpCurrent->lpszFirstName);
          if (lpCurrent->lpszLastName)
            free(lpCurrent->lpszLastName);

          /* Split the real name into the first and last names */
          lpszLastName = GetLastName(lpszRealName);
          if (lpszLastName) {
            lpCurrent->lpszLastName = (LPTSTR) malloc(strlen(lpszLastName) + 1);
            strcpy(lpCurrent->lpszLastName, lpszLastName);
            lpCurrent->lpszFirstName = lpszRealName;
            lpCurrent->lpszFirstName = (LPTSTR) realloc(lpCurrent->lpszFirstName, lpszLastName - lpszRealName);
            lpCurrent->lpszFirstName[lpszLastName - lpszRealName - 1] = 0;
          } else {
            lpCurrent->lpszLastName = lpszRealName;
            lpCurrent->lpszFirstName = (LPTSTR) malloc(1);
            lpCurrent->lpszFirstName[0] = 0;
          }

          /* Fill in the default values for the other strings */
          if (bNewEntry) {
            lpCurrent->lpszWebSite = malloc(strlen(lpCurrent->lpszUserName) + 3);
            sprintf(lpCurrent->lpszWebSite, "/~%s", lpCurrent->lpszUserName);
            lpCurrent->lpszEMail = malloc(strlen(lpCurrent->lpszUserName) + 2 + lEMailDomain);
            sprintf(lpCurrent->lpszEMail, "%s@%s", lpCurrent->lpszUserName, lpszEMailDomain);
            lpCurrent->lpszTelephone = malloc(1);
            lpCurrent->lpszTelephone[0] = 0;
            lpCurrent->dwFlags |= FLAGS_NOTELEPHONE;
          }
        }
      }
      more = FindNext(&ffData);
    } while (more);
    CloseFind(&ffData);
  }

  /* Return the length of the list */
  return num;
}

void FreeSortTables(LPDIRLIST_SORTTABLE lpTable)
{
  /* Free all the sort tables in the given linked list */
  LPDIRLIST_SORTTABLE lpCurTable = lpTable, lpPrevTable;
  LPDIRLIST_SORTENTRY lpCurEntry, lpPrevEntry;
  while (lpCurTable)
  {
    /* Free any entries that this sort table might have */
    lpCurEntry = lpTable->lpSortEntries;
    while (lpCurEntry) {
      lpPrevEntry = lpCurEntry;
      lpCurEntry = (LPDIRLIST_SORTENTRY) lpCurEntry->next;
      free(lpPrevEntry);
    }

    /* Free the sort order string, if it exists */
    if (lpCurTable->lpszOrder)
      free(lpCurTable->lpszOrder);

    /* Free the current table */
    lpPrevTable = lpCurTable;
    lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
    free(lpPrevTable);
  }
}

void FreeEntry(LPDIRLIST_ENTRY lpDirEntry)
{
  /* Free one entry's data areas (but not lpDirEntry itself) */
  if (lpDirEntry->lpszUserName)
    free(lpDirEntry->lpszUserName);
  if (lpDirEntry->lpszFirstName)
    free(lpDirEntry->lpszFirstName);
  if (lpDirEntry->lpszLastName)
    free(lpDirEntry->lpszLastName);
  if (lpDirEntry->lpszWebSite)
    free(lpDirEntry->lpszWebSite);
  if (lpDirEntry->lpszEMail)
    free(lpDirEntry->lpszEMail);
  if (lpDirEntry->lpszTelephone)
    free(lpDirEntry->lpszTelephone);

  /* Free the class data, if any */
  if ((lpDirEntry->dwFlags & FLAGS_CLASSDATA) && lpDirEntry->lpClassData) {
    DWORD i;
    for (i = 0; i < lpDirEntry->nClassDataValues; i++) {
      if (lpDirEntry->lpClassData[i].lpszValue && (lpDirEntry->lpClassData[i].dwFlags & CLASSDATA_OTHER))
        free(lpDirEntry->lpClassData[i].lpszValue);
    }
    free(lpDirEntry->lpClassData);
  }
}

void FreeHeader(LPDIRLIST_HEADER lpHeader)
{
  /* Free the two strings in the header if they exist */
  if (lpHeader) {
    if (lpHeader->lpszWebDocsDir)
      free(lpHeader->lpszWebDocsDir);
    if (lpHeader->lpszBaseDir)
      free(lpHeader->lpszBaseDir);
    if (lpHeader->lpszAuditEMail)
      free(lpHeader->lpszAuditEMail);
  }
}

void FreeUsers(LPDIRLIST_USER lpUser)
{
  /* Free each user */
  LPDIRLIST_USER lpCurrent = lpUser, lpPrev;
  while (lpCurrent)
  {
    /* Free the strings!  Free the strings...!  A mistake is made..  A story must be told... ;-) */
    if (lpCurrent->lpszName)
      free(lpCurrent->lpszName);
    if (lpCurrent->lpszPassword)
      free(lpCurrent->lpszPassword);
    if (lpCurrent->lpszDescription)
      free(lpCurrent->lpszDescription);

    /* Next entry */
    lpPrev = lpCurrent;
    lpCurrent = (LPDIRLIST_USER) lpCurrent->next;

    /* Free the entry we were just working with */
    free(lpPrev);
  }
}

void FreeAccessPlan(LPDIRLIST_ACCESSPLAN lpPlan)
{
  /* Free the access plan's allocations */
  DWORD i;
  if (lpPlan->lpszBinarySearch)
    free(lpPlan->lpszBinarySearch);
  if (lpPlan->lpszAuditMsg)
    free(lpPlan->lpszAuditMsg);

  /* Free the strings in the search/range parameters */
  for (i = 0; i < lpPlan->nSearches; i++) {
    if (lpPlan->lpszSearch[i])
      free(lpPlan->lpszSearch[i]);
    if ((lpPlan->lpszRangeLow[i] || lpPlan->lpszRangeHigh[i]) && (lpPlan->lpszRangeLow[i] != lpPlan->lpszSearch[i])) {
      if (lpPlan->lpszRangeLow[i] == lpPlan->lpszRangeHigh[i])
        free(lpPlan->lpszRangeLow[i]);
      else {
        if (lpPlan->lpszRangeLow[i])
          free(lpPlan->lpszRangeLow[i]);
        if (lpPlan->lpszRangeHigh[i])
          free(lpPlan->lpszRangeHigh[i]);
      }
    }
  }

  /* Free the search/range parameters */
  if (lpPlan->nSearches) {
    free(lpPlan->dwSearchFields);
    free(lpPlan->dwExtSearchFields);
    free(lpPlan->dwExtSearchOrdinals);
    free(lpPlan->bSearchNot);
    free(lpPlan->bSearchAnd);
    free(lpPlan->dwSearchLevel);
    free(lpPlan->dwSearchLevelEnd);
    free(lpPlan->fSearchExact);
    free(lpPlan->lpszSearch);
    free(lpPlan->lpszRangeLow);
    free(lpPlan->lpszRangeHigh);
    free(lpPlan->bRangeLowInclusive);
    free(lpPlan->bRangeHighInclusive);
  }

  /* Free the lpUser structure */
  if (lpPlan->lpUser)
    FreeUsers(lpPlan->lpUser);
}

void FreeEntries(LPDIRLIST_ENTRY lpDirEntry)
{
  /* Go through and delete each entry's memory */
  LPDIRLIST_ENTRY lpCurrent = lpDirEntry, lpPrev;
  while (lpCurrent)
  {
    /* Free the current entry's data */
    FreeEntry(lpCurrent);

    /* Next entry */
    lpPrev = lpCurrent;
    lpCurrent = (LPDIRLIST_ENTRY) lpCurrent->next;

    /* Free the entry we were just working with */
    free(lpPrev);
  }
}

void FreeClasses(LPDIRLIST_CLASS lpClass)
{
  /* Go through and delete each class's memory */
  LPDIRLIST_CLASS lpCurrent = lpClass, lpPrev;
  DWORD i;
  while (lpCurrent) {
    if (lpCurrent->lpszName)
      free(lpCurrent->lpszName);

    /* Free the class values, if any */
    for (i = 0; i < lpCurrent->nValues; i++) {
      if (lpCurrent->lpValues[i].lpszValue)
        free(lpCurrent->lpValues[i].lpszValue);
    }
    if (lpCurrent->lpValues)
      free(lpCurrent->lpValues);

    /* Next entry */
    lpPrev = lpCurrent;
    lpCurrent = (LPDIRLIST_CLASS) lpCurrent->next;

    /* Free the entry we were just working with */
    free(lpPrev);
  }
}

void FreeIndex(LPDIRLIST_INDEX lpIndex)
{
  /* Free all the linked lists within the index */
  if (lpIndex->lpHeader) {
    FreeHeader(lpIndex->lpHeader);
    free(lpIndex->lpHeader);
    lpIndex->lpHeader = NULL;
  }
  if (lpIndex->lpSortTable) {
    FreeSortTables(lpIndex->lpSortTable);
    lpIndex->lpSortTable = NULL;
  }
  if (lpIndex->lpClass) {
    FreeClasses(lpIndex->lpClass);
    lpIndex->lpClass = NULL;
  }
  if (lpIndex->lpDirEntry) {
    FreeEntries(lpIndex->lpDirEntry);
    lpIndex->lpDirEntry = NULL;
  }
  if (lpIndex->lpUser) {
    FreeUsers(lpIndex->lpUser);
    lpIndex->lpUser = NULL;
  }
  if (lpIndex->lpLock) {
    free(lpIndex->lpLock);
    lpIndex->lpLock = NULL;
  }
}

/* Login the user */
LPDIRLIST_USER LoginUser(LPDIRLIST_USER *lpUser, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwOperations, BOOL bEmitStatus)
{
  /* Find a username-password pair in the lpUser linked list that matches the one in lpPlan */
  LPDIRLIST_USER lpCurrent = *lpUser;
  BOOL bFound = FALSE;
  BOOL bAdminFound = FALSE;
  if (!lpPlan->lpUser) {
    if (bEmitStatus)
      infprintf("#Could not login: Username and/or password not given!\n");
    return NULL;
  }
  while (lpCurrent && !bFound)
  {
    /* Compare the username */
    if (CompareStrings(lpCurrent->lpszName, "Administrator") == 0)
      bAdminFound = TRUE;
    if (CompareStrings(lpCurrent->lpszName, lpPlan->lpUser->lpszName) == 0)
    {
      /* Username ok; check password */
      bFound = TRUE;
      if (strcmp(lpCurrent->lpszPassword, lpPlan->lpUser->lpszPassword) == 0)
      {
      /* If the user is Administrator, then automatically allow all operations (just in case if the Administrator
       revoked his/her own permissions) */
    if (bAdminFound) {
      lpCurrent->dwOperations = 0xFFFFFFFF;
      lpCurrent->bAdminMode = TRUE;
    }

        /* Logged in; check the requested operation with the allowed operations (ignore OPERATION_NOOP) */
        dwOperations &= ~OPERATION_NOOP;
        if (dwOperations & ~lpCurrent->dwOperations) {
          if (bEmitStatus)
            infprintf("#Access denied: You are not allowed to perform one or more requested operation(s).\n");
          return NULL;
        }

        /* Check the search levels if a search is a part of the requested operations */
        if ((dwOperations == OPERATION_WEBQUERY) || (dwOperations & (OPERATION_GETENTRIES | OPERATION_MODIFYENTRIES | OPERATION_DELETEENTRIES)))
        {
          /* If the user cannot search at all (but has OPERATION_GETENTRIES access??) then return an error */
          if (((dwOperations == OPERATION_WEBQUERY) && !lpCurrent->dwSearchLevel) || (!lpCurrent->dwSearchLevel && ((lpPlan->nSearches > 0) || lpPlan->dwXFields || lpPlan->dwXExtFields))) {
            if (bEmitStatus)
              infprintf("#Access denied: You do not have any search rights.\n");
            return NULL;
          }
          if (lpCurrent->dwSearchLevel & SEARCHLEVEL_SOME)
          {
            /* If searches exist, then check the search code */
            DWORD i;
            BOOL bFound = FALSE;
            BOOL bAtException = FALSE;
            LPTSTR lpszAt;
            if (lpPlan->nSearches)
            {
              /* Search level is "some"; ensure that at least one string has 2 or more chars in it */
              for (i = 0; (i < lpPlan->nSearches) && !bFound; i++)
              {
                /* Return FALSE if AND is not set */
                if (!lpPlan->bSearchAnd[i] && (i < lpPlan->nSearches - 1)) {
                  if (bEmitStatus)
                    infprintf("#Access denied: Your search level is restricted to SEARCHLEVEL_SOME and must have AND (instead of OR) between every term in the conditional expression.\n");
                  return NULL;
                }

                /* Give an error if any search ranges exist */
                if (lpPlan->lpszRangeLow[i] != lpPlan->lpszRangeHigh[i]) {
                  if (bEmitStatus)
                    infprintf("#Access denied: Your search level is restricted to SEARCHLEVEL_SOME and must not have any ranges present (i.e., when requesting records with values between two given parameters).\n");
                  return NULL;
                }

                /* Perform special execution for "@domain.edu" strings */
                if ((lpPlan->dwSearchFields[i] & DISPLAY_EMAIL) && !lpPlan->fSearchExact[i])
                {
                  lpszAt = strchr(lpPlan->lpszSearch[i], '@');
                  if (lpszAt) {
                    if ((P_TYPE) lpszAt < (P_TYPE) lpPlan->lpszSearch[i] + 2)
                      bAtException = TRUE;
                  } else
                  {
                    /* Add a @ at the end to be cruel ;-) */
                    DWORD dwSearchLen = strlen(lpPlan->lpszSearch[i]) + 1;
                    LPTSTR lpszNewSearch = malloc(dwSearchLen + 1);
                    memcpy(lpszNewSearch, lpPlan->lpszSearch[i], dwSearchLen - 1);
                    lpszNewSearch[dwSearchLen - 1] = '@';
                    lpszNewSearch[dwSearchLen] = 0;
                    free(lpPlan->lpszSearch[i]);
                    lpPlan->lpszSearch[i] = lpszNewSearch;

                    /* Check the length, taking our @ addition into account */
                    if (dwSearchLen < 3)
                      bAtException = TRUE;
                  }
                }

                /* If DISPLAY_FLAGS or DISPLAY_TELEPHONE exist, then give an error */
                if (lpPlan->dwSearchFields[i] & (DISPLAY_FLAGS | DISPLAY_TELEPHONE)) {
                  if (bEmitStatus)
                    infprintf("#Access denied: Your search level is restricted to SEARCHLEVEL_SOME and will not allow you to see the flags.\n");
                  return NULL;
                }

                /* Perform the grand IF conditional to determine whether or not this search is okay */
                if (((lpPlan->dwSearchFields[i] & DISPLAY_FIRSTNAME) || (lpPlan->dwSearchFields[i] & DISPLAY_LASTNAME) ||
                  (lpPlan->dwSearchFields[i] & DISPLAY_WEBSITE) || (lpPlan->dwSearchFields[i] & DISPLAY_EMAIL))
                  && (((lpPlan->fSearchExact[i] == SEARCH_EXACT) || (lpPlan->fSearchExact[i] == SEARCH_STRICTSTARTSWITH))
                  || (strlen(lpPlan->lpszSearch[i]) >= 2)) && !bAtException && (!lpPlan->bSearchNot[i]) &&
                  (lpPlan->lpszRangeHigh[i] == lpPlan->lpszRangeLow[i]))
                  bFound = TRUE;
                bAtException = FALSE;
              }
            }

            /* If a web-site list is being generated, then ensure e-mail and telephone are not included */
            if (lpPlan->dwXFields & DISPLAY_WEBSITE) {
              if (lpPlan->dwFields & (DISPLAY_EMAIL | DISPLAY_TELEPHONE)) {
                if (bEmitStatus)
                  infprintf("#Access denied: Your search level is restricted to SEARCHLEVEL_SOME and it restricts you from viewing e-mail addresses and telephone numbers when generating a list of people with web sites.\n");
                return NULL;
              }

              /* If we're just doing the web sites, then users can search for almost anything :-) */
              bFound = TRUE;
            }

            /* If a valid search was not found, then return FALSE */
            if (!bFound) {
              if (bEmitStatus)
                infprintf("#Access denied: Your search level is restricted to SEARCHLEVEL_SOME and your search parameters did not fall within the required criteria.\n");
              return NULL;
            }
          }
        }

        /* Ensure the requested administrative mode state is allowed */
        if (lpPlan->bAdminMode && !lpCurrent->bAdminMode) {
          if (bEmitStatus)
            infprintf("#Access denied: You do not have the required permissions to use the Administrative Mode.\n");
          return NULL;
        }

        /* Set the audit mode */
        lpPlan->bAudit = lpCurrent->bAudit;
      } else {
        if (bEmitStatus)
          infprintf("#Could not login: Username and/or password not correct!\n");
        return NULL;
      }
    }

    /* Next user */
    if (!bFound)
      lpCurrent = (LPDIRLIST_USER) lpCurrent->next;
  }

  /* User not found! */
  if (!bFound)
  {
    /* If the username was "admin" then that means that the administrator does not exist; we will grant access */
    if (!bAdminFound) {
      if ((CompareStrings(lpPlan->lpUser->lpszName, "Administrator") == 0) && (strcmp(lpPlan->lpUser->lpszPassword, "") == 0))
      {
        /* Create the administrative user */
        lpCurrent = (LPDIRLIST_USER) malloc(sizeof(DIRLIST_USER));
        memset(lpCurrent, 0, sizeof(DIRLIST_USER));
        lpCurrent->idUser = 0;
        lpCurrent->lpszName = strdup("Administrator");
        lpCurrent->lpszPassword = strdup("");
        lpCurrent->lpszDescription = strdup("Administrator of the DirList2 database");
        lpCurrent->dwOperations = 0xFFFFFFFF;
        lpCurrent->bAdminMode = TRUE;
        lpCurrent->dwSearchLevel = SEARCHLEVEL_ALL;
        lpCurrent->next = (LPBYTE) *lpUser;
        *lpUser = lpCurrent;
      } else {
        if (bEmitStatus)
          infprintf("#Could not login: Username and/or password not correct!\n");
        return NULL;
      }
    } else {
      if (bEmitStatus)
        infprintf("#Could not login: Username and/or password not correct!\n");
      return NULL;
    }
  }

  /* Return the user */
  return lpCurrent;
}

/* Returns whether or not a change should be made */
DWORD GetUpdateMode(LPTSTR lpszClient, LPTSTR *lpszValue, BOOL bEmitStatus)
{
  /* Switch based on the first character */
  DWORD dwUpdateMode;
  switch (lpszClient[0])
  {
  case '%' :
    {
      /* Set and remove mode (first field has fields to set, then next has fields to remove from flags) */
      dwUpdateMode = FLAG_SETREMOVE;
      break;
    }

  case '+' :
    {
      /* Set mode */
      dwUpdateMode = FLAG_SET;
      break;
    }

  case '-' :
    {
      /* Remove mode */
      dwUpdateMode = FLAG_REMOVE;
      break;
    }

  case '|' :
    {
      /* Replace mode */
      dwUpdateMode = FLAG_REPLACE;
      break;
    }

  case '<' :
    {
      /* Ignore mode */
      dwUpdateMode = FLAG_IGNORE;
      break;
    }

  default :
    {
      /* Unknown mode */
      if (bEmitStatus)
        infprintf("#Unknown line operation '%c'\n", lpszClient[0]); /* This is safe; it's not called from web browsers */
      return FLAG_UNKNOWN;
    }
  }

  /* Set the value equal to the first character */
  *lpszValue = &(lpszClient[1]);
  return dwUpdateMode;
}

/* Get the flag for FLAG_SETREMOVE mode */
BOOL GetSetRemoveFlag(LPTSTR lpszStr, BOOL bFirst, DWORD *dwFlag)
{
  /* Find the comma */
  LPTSTR lpszComma = strchr(lpszStr, ',');
  if (!lpszComma)
    return FALSE;
  lpszComma[0] = 0;
  if (bFirst)
    *dwFlag = strtoul(lpszStr, NULL, 10);
  else
    *dwFlag = strtoul(&(lpszComma[1]), NULL, 10);
  lpszComma[0] = ',';
  return TRUE;
}

/* Modify the header of the index file */
BOOL ModifyIndexInfo(LPDIRLIST_INDEX lpIndex, FILE *pInput, LPDWORD lpdwMaxRead, UCHAR cDelimiter, BOOL bEmitStatus)
{
  /* Initialize variables */
  LPDIRLIST_SORTTABLE lpCurTable;
  LPTSTR lpszLine = (LPTSTR) malloc(32);
  LPTSTR lpszCurrent;
  LPTSTR lpszValue;
  DWORD dwLineLen = 0, dwLineAlloc = 32, dwUpdateMode;

  /* Tell the user that we're about to import the file */
  if (bEmitStatus) {
    infprintf("#Merging input into memory...\n");
    fflush(stdout);
  }

  /* Read the only line of input */
  if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead))
  {
    if (bEmitStatus)
      infprintf("#Could not read the input line (one must exist)!\n");
    free(lpszLine);
    return FALSE;
  }

  /* Get the flags */
  lpszCurrent = strtok2(lpszLine, cDelimiter);
  if (!lpszCurrent) {
    if (bEmitStatus)
      infprintf("#Could not read the Flags field!\n");
    free(lpszLine);
    return FALSE;
  }
  if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN)
    return FALSE;
  if (dwUpdateMode == FLAG_SET)
    lpIndex->lpHeader->dwFlags |= strtoul(lpszValue, NULL, 10);
  if (dwUpdateMode == FLAG_REMOVE)
    lpIndex->lpHeader->dwFlags &= ~strtoul(lpszValue, NULL, 10);
  if (dwUpdateMode == FLAG_REPLACE)
    lpIndex->lpHeader->dwFlags = strtoul(lpszValue, NULL, 10);
  if (dwUpdateMode == FLAG_SETREMOVE) {
    DWORD dwFlag;
    if (!GetSetRemoveFlag(lpszValue, TRUE, &dwFlag)) {
      if (bEmitStatus)
        infprintf("#Could not read the second part of the Flags field!\n");
      free(lpszLine);
      return FALSE;
    }
    lpIndex->lpHeader->dwFlags |= dwFlag;
    GetSetRemoveFlag(lpszValue, FALSE, &dwFlag);
    lpIndex->lpHeader->dwFlags &= ~dwFlag;
  }

  /* Get the web docs directory */
  lpszCurrent = strtok2(NULL, cDelimiter);
  if (!lpszCurrent) {
    if (bEmitStatus)
      infprintf("#Could not read the WebDocsDir field!\n");
    free(lpszLine);
    return FALSE;
  }
  if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
    free(lpszLine);
    return FALSE;
  }
  if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
  {
    if (lpIndex->lpHeader->lpszWebDocsDir)
      free(lpIndex->lpHeader->lpszWebDocsDir);
    lpIndex->lpHeader->lpszWebDocsDir = strdup(lpszValue);
  } else
    if (!lpIndex->lpHeader->lpszWebDocsDir)
      lpIndex->lpHeader->lpszWebDocsDir = strdup("");

  /* Get the base directory */
  lpszCurrent = strtok2(NULL, cDelimiter);
  if (!lpszCurrent) {
    if (bEmitStatus)
      infprintf("#Could not read the BaseDir field!\n");
    free(lpszLine);
    return FALSE;
  }
  if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
    free(lpszLine);
    return FALSE;
  }
  if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
  {
    if (lpIndex->lpHeader->lpszBaseDir)
      free(lpIndex->lpHeader->lpszBaseDir);
    lpIndex->lpHeader->lpszBaseDir = strdup(lpszValue);
  } else {
    if (!lpIndex->lpHeader->lpszBaseDir)
      lpIndex->lpHeader->lpszBaseDir = strdup("");
  }

  /* Get the audit e-mail address */
  lpszCurrent = strtok2(NULL, cDelimiter);
  if (!lpszCurrent) {
    if (bEmitStatus)
      infprintf("#Could not read the AuditEMail field!\n");
    free(lpszLine);
    return FALSE;
  }
  if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
    free(lpszLine);
    return FALSE;
  }
  if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
  {
    if (lpIndex->lpHeader->lpszAuditEMail)
      free(lpIndex->lpHeader->lpszAuditEMail);
    lpIndex->lpHeader->lpszAuditEMail = strdup(lpszValue);
    SafeEMailAddress(lpIndex->lpHeader->lpszAuditEMail);
  } else {
    if (!lpIndex->lpHeader->lpszAuditEMail)
      lpIndex->lpHeader->lpszAuditEMail = strdup("");
  }

  /* Get the e-mail domain, if given */
  lpszCurrent = strtok2(NULL, cDelimiter);
  if (lpszCurrent) {
    if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
      free(lpszLine);
      return FALSE;
    }
    if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
    {
      if (lpIndex->lpHeader->lpszEMailDomain)
        free(lpIndex->lpHeader->lpszEMailDomain);
      lpIndex->lpHeader->lpszEMailDomain = strdup(lpszValue);
      SafeDomainName(lpIndex->lpHeader->lpszEMailDomain);
    } else {
      if (!lpIndex->lpHeader->lpszEMailDomain)
        lpIndex->lpHeader->lpszEMailDomain = strdup("");
    }

    /* Get the default URL, if given */
    lpszCurrent = strtok2(NULL, cDelimiter);
    if (lpszCurrent) {
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    }
      if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      {
        if (lpIndex->lpHeader->lpszDefaultURL)
          free(lpIndex->lpHeader->lpszDefaultURL);
        lpIndex->lpHeader->lpszDefaultURL = strdup(lpszValue);
      } else {
        if (!lpIndex->lpHeader->lpszDefaultURL)
          lpIndex->lpHeader->lpszDefaultURL = strdup("");
      }

      /* If more columns exist, then assert a formatting error */
      if (strtok2(NULL, cDelimiter) != NULL) {
        if (bEmitStatus) {
          infprintf("#Formatting error in input!  Too many columns exist.\n");
          fflush(stdout);
        }
        free(lpszLine);
        return FALSE;
      }
    } else {
      if (!lpIndex->lpHeader->lpszDefaultURL)
        lpIndex->lpHeader->lpszDefaultURL = strdup("");
    }
  } else {
    if (!lpIndex->lpHeader->lpszEMailDomain)
      lpIndex->lpHeader->lpszEMailDomain = strdup("");
    if (!lpIndex->lpHeader->lpszDefaultURL)
      lpIndex->lpHeader->lpszDefaultURL = strdup("");
  }

  /* Tell the user what we did */
  infprintf("*Modified index information\n");

  /* Everything was modified correctly :-)!  Rebuild the sort tables */
  if (bEmitStatus) {
    infprintf("#Refreshing sort tables...\n");
    fflush(stdout);
  }
  lpCurTable = lpIndex->lpSortTable;
  while (lpCurTable) {
    CreateSortTable(lpCurTable, lpIndex->lpDirEntry, lpIndex->lpClass);
    lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
  }

  /* We did it :-) */
  free(lpszLine);
  return TRUE;
}

/* Return a user's record by looking up the user's name */
LPDIRLIST_USER GetUserByName(LPDIRLIST_INDEX lpIndex, LPCTSTR lpszName)
{
  /* Search for the given user name in the given index */
  LPDIRLIST_USER lpCurrent = lpIndex->lpUser;
  while (lpCurrent)
  {
    /* If the name matches, then return this user */
    if (lpCurrent->lpszName && !strcasecmp(lpszName, lpCurrent->lpszName))
      return lpCurrent;

    /* Next user */
    lpCurrent = (LPDIRLIST_USER) lpCurrent->next;
  }

  /* We didn't find this user */
  return NULL;
}

/* Modify the given users of an index file */
BOOL ModifyUsers(LPDIRLIST_INDEX lpIndex, LPDIRLIST_USER lpActiveUser, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, UCHAR cDelimiter, BOOL bEmitStatus)
{
  /* Initialize variables */
  LPDIRLIST_USER lpCurrent, lpCurUser;
  LPDIRLIST_SORTTABLE lpCurTable;
  LPTSTR lpszLine = (LPTSTR) malloc(32);
  LPTSTR lpszCurrent;
  LPTSTR lpszValue;
  DWORD dwLineLen = 0, dwLineAlloc = 32, idUser, dwUpdateMode;
  BOOL bFormatError = FALSE, bFound;
  DWORD origSafeStringsLen = safeStringsLen;

  /* Tell the user that we're about to import the file */
  if (bEmitStatus) {
    infprintf("#Merging input into memory...\n");
    fflush(stdout);
  }

  /* Go through each line in the input (each line MUST have a delimiter) */
  while (ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead) && !bFormatError)
  {
    /* Break the string up into parts and process each one */
    lpszCurrent = strtok2(lpszLine, cDelimiter);
    if (!lpszCurrent) {
      bFormatError = TRUE;
      break;
    }

    /* Convert the current string into a user ID and search for it */
    idUser = strtoul(lpszCurrent, NULL, 10);
    bFound = FALSE;
    lpCurrent = lpIndex->lpUser;
    while (!bFound && lpCurrent) {
      if (lpCurrent->idUser == idUser)
        bFound = TRUE;
      else
        lpCurrent = (LPDIRLIST_USER) lpCurrent->next;
    }

    /* Make sure the active user's dwUserLevel is lower than the user we are modifying (except for Administrator and the same user) */
    if (bFound && lpActiveUser) {
      if (lpActiveUser->dwUserLevel >= lpCurrent->dwUserLevel && lpActiveUser->idUser != lpCurrent->idUser && CompareStrings(lpActiveUser->lpszName, "Administrator")) {
        if (bEmitStatus)
          infprintf("#Cannot %s user with ID %u; that user's level equals %u, which is %s yours (%u).\n", (bDelete)? "delete": "modify", lpCurrent->idUser, lpCurrent->dwUserLevel, (lpActiveUser->dwUserLevel == lpCurrent->dwUserLevel)? "the same as": "less than", lpActiveUser->dwUserLevel);
        free(lpszLine);
        return FALSE;
      }
    }

    /* Process destruction and creation differently */
    if (bDelete)
    {
      /* If the user was not found, then give an error */
      if (!lpCurrent)
      {
        /* The user already does not exist (this code should never execute) */
        if (bEmitStatus)
          infprintf("#User %u does not exist or already has been deleted.  Entire operation aborted.\n", idUser);
        free(lpszLine);
        return FALSE;
      }

      /* Tell the user what we are about to do */
      infprintf("*Deleted user \"%s\" (%u)\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszName, STR_NUL_TERM), lpCurrent->idUser);
      safeStringsLen = origSafeStringsLen;

      /* If the user ID is zero, then we must not delete it (it's the administrator) */
      if (!idUser) {
        if (bEmitStatus)
          infprintf("#Attempt to delete \"admin\" failed; this account cannot be deleted in DirList2.\n");
        free(lpszLine);
        return FALSE;
      }

      /* Remove this user from the linked list */
      if (lpCurrent != lpIndex->lpUser) {
        lpCurUser = (LPDIRLIST_USER) lpIndex->lpUser;
        while (lpCurUser->next != (LPBYTE) lpCurrent) {
          lpCurUser = (LPDIRLIST_USER) lpCurUser->next;
          if (!lpCurUser)
            break;
        }
        if (!lpCurUser)
        {
          /* The user already does not exist (this code should never execute) */
          if (bEmitStatus)
            infprintf("#User %u does not exist or already has been deleted.  This is the second-level warning, which should never appear.  Entire operation aborted.\n", idUser);
          free(lpszLine);
          return FALSE;
        }

        /* Pop this user out of the linked list */
        lpCurUser->next = (LPBYTE) lpCurrent->next;
        lpCurrent->next = NULL;
        FreeUsers(lpCurrent);
      } else
      {
        /* Easy removal; it is at the beginning of the linked list */
        lpIndex->lpUser = (LPDIRLIST_USER) lpCurrent->next;
        lpCurrent->next = NULL;
        FreeUsers(lpCurrent);
      }
    } else
    {
      /* If we did not find it, then create an allocation and add it to the list */
      if (!bFound) {
        lpCurrent = (LPDIRLIST_USER) malloc(sizeof(DIRLIST_USER));
        memset(lpCurrent, 0, sizeof(DIRLIST_USER));
        lpCurrent->idUser = idUser;
        lpCurrent->next = (LPBYTE) lpIndex->lpUser;
        lpIndex->lpUser = lpCurrent;
      }

      /* Get the name */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the Name field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN)
        return FALSE;
      if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      {
        /* If the active user equals the user being modified and the user does not
           have admin mode, then prevent this change */
        if (lpActiveUser->idUser == lpCurrent->idUser && !lpActiveUser->bAdminMode && ((lpCurrent->lpszName)? CompareStrings(lpszValue, lpCurrent->lpszName): TRUE)) {
          if (bEmitStatus)
            infprintf("#You cannot change your own name unless if you have the Admin Mode permission.\n");
          free(lpszLine);
          return FALSE;
        }

        /* Ensure that the given username doesn't exist */
        LPDIRLIST_USER lpExistingUser = GetUserByName(lpIndex, lpszValue);
        if (lpExistingUser) {
          if (bEmitStatus) {
            infprintf("#User %u already has the name \"%s\", preventing this operation from succeeding.\n", lpExistingUser->idUser,
                      SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpszValue, STR_NUL_TERM));
            safeStringsLen = origSafeStringsLen;
          }
          free(lpszLine);
          return FALSE;
        }
        if (lpCurrent->lpszName)
          free(lpCurrent->lpszName);
        lpCurrent->lpszName = strdup(lpszValue);
      } else
        if (!lpCurrent->lpszName)
          lpCurrent->lpszName = strdup("");

      /* Get the password */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the Password field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    }
      if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      {
        /* Change the password */
        if (lpCurrent->lpszPassword)
          free(lpCurrent->lpszPassword);
        lpCurrent->lpszPassword = strdup(lpszValue);
      } else
        if (!lpCurrent->lpszPassword)
          lpCurrent->lpszPassword = strdup("");

      /* Get the description */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the Description field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    }
    if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      {
        if (lpCurrent->lpszDescription)
          free(lpCurrent->lpszDescription);
        lpCurrent->lpszDescription = strdup(lpszValue);
      } else
        if (!lpCurrent->lpszDescription)
          lpCurrent->lpszDescription = strdup("");

      /* Get the user level */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the UserLevel field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    }
      if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      {
        /* If the active user equals the user being modified and the user does not
           have admin mode, then prevent this change */
        DWORD newValue = strtoul(lpszValue, NULL, 10);
        if (lpActiveUser->idUser == lpCurrent->idUser && !lpActiveUser->bAdminMode && newValue != lpCurrent->dwUserLevel) {
          if (bEmitStatus)
            infprintf("#You cannot change your own user level unless if you have the Admin Mode permission.\n");
          free(lpszLine);
          return FALSE;
        }

        /* Verify the user level; it must be greater than the active user's level,
           though Administrator can reset it to anything */
        if (newValue <= lpActiveUser->dwUserLevel) {
          if (CompareStrings(lpActiveUser->lpszName, "Administrator")) /* meaning that it's not the Administrator */
          {
            /* Reject if this was a change or is a new user */
            if (lpCurrent->dwUserLevel != newValue || !bFound) {
              if (bEmitStatus)
                infprintf("#Only a user with a lower user level than you or Administrator can set the user level on user %u to %u. The lowest user level that you can set is %u.\n", lpCurrent->idUser, newValue, lpActiveUser->dwUserLevel + 1);
              free(lpszLine);
              return FALSE;
            }
          }
        }

        /* Set the new value */
        lpCurrent->dwUserLevel = newValue;
      }

      /* Get the operations */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the Operations field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    } else
      {
        /* Determine what the new operations will be */
        DWORD dwOperations = lpCurrent->dwOperations;
        if (dwUpdateMode == FLAG_REPLACE)
          dwOperations = strtoul(lpszValue, NULL, 10);
        if (dwUpdateMode == FLAG_SET)
          dwOperations |= strtoul(lpszValue, NULL, 10);
        if (dwUpdateMode == FLAG_REMOVE)
          dwOperations &= ~strtoul(lpszValue, NULL, 10);
        if (dwUpdateMode == FLAG_SETREMOVE) {
          DWORD dwFlag;
          if (!GetSetRemoveFlag(lpszValue, TRUE, &dwFlag)) {
            if (bEmitStatus)
              infprintf("#Could not read the second part of the Operations field!\n");
            free(lpszLine);
            return FALSE;
          }
          dwOperations |= dwFlag;
          GetSetRemoveFlag(lpszValue, FALSE, &dwFlag);
          dwOperations &= ~dwFlag;
        }

        /* If the active user equals the user being modified, then prevent this change if not the Administrator */
        if (lpActiveUser->idUser == lpCurrent->idUser && dwOperations != lpCurrent->dwOperations && CompareStrings(lpActiveUser->lpszName, "Administrator")) {
          if (bEmitStatus)
            infprintf("#Only the Administrator can change their own permitted operations.\n");
          free(lpszLine);
          return FALSE;
        }

        /* Set the new operations */
        lpCurrent->dwOperations = dwOperations;
      }

      /* Get the admin mode */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the AdminMode field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    }
      if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      {
        /* Prevent non-admins from changing bAdminMode (unless if they are changing it to
           the same value). Ensure that this doesn't apply to Administrator */
        BOOL newValue = (BOOL) strtoul(lpszValue, NULL, 10);
        if (!lpActiveUser->bAdminMode && (newValue ^ lpCurrent->bAdminMode) && CompareStrings(lpActiveUser->lpszName, "Administrator")) {
          if (bEmitStatus)
            infprintf("#Non-administrators cannot change AdminMode\n");
          free(lpszLine);
          return FALSE;
        } else
          lpCurrent->bAdminMode = newValue;
      }

      /* Get the search level */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the SearchLevel field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    }
      if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      {
        /* If the active user equals the user being modified and the user does not
           have admin mode, then prevent this change */
        DWORD newValue = strtoul(lpszValue, NULL, 10);
        if (lpActiveUser->idUser == lpCurrent->idUser && !lpActiveUser->bAdminMode && newValue != lpCurrent->dwSearchLevel) {
          if (bEmitStatus)
            infprintf("#You cannot change your own search level unless if you have the Admin Mode permission.\n");
          free(lpszLine);
          return FALSE;
        }
        lpCurrent->dwSearchLevel = newValue;
      }

      /* Get the audit flag */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the Audit field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    }
      if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      {
        /* If the active user equals the user being modified, then prevent this change if not the Administrator */
        BOOL bAudit = (BOOL) strtoul(lpszValue, NULL, 10);
        if (lpActiveUser->idUser == lpCurrent->idUser && bAudit != lpCurrent->bAudit && CompareStrings(lpActiveUser->lpszName, "Administrator")) {
          if (bEmitStatus)
            infprintf("#Only the Administrator can change their own audit on/off flag.\n");
          free(lpszLine);
          return FALSE;
        }
        lpCurrent->bAudit = bAudit;
      }

      /* Get the audited operations */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent) {
        if (bEmitStatus)
          infprintf("#Could not read the AuditOperations field!\n");
        free(lpszLine);
        return FALSE;
      }
      if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
        free(lpszLine);
        return FALSE;
    } else
      {
        /* Determine the new audit operations */
        DWORD dwAuditOperations = lpCurrent->dwAuditOperations;
        if (dwUpdateMode == FLAG_REPLACE)
          dwAuditOperations = strtoul(lpszValue, NULL, 10);
        if (dwUpdateMode == FLAG_SET)
          dwAuditOperations |= strtoul(lpszValue, NULL, 10);
        if (dwUpdateMode == FLAG_REMOVE)
          dwAuditOperations &= ~strtoul(lpszValue, NULL, 10);
        if (dwUpdateMode == FLAG_SETREMOVE) {
          DWORD dwFlag;
          if (!GetSetRemoveFlag(lpszValue, TRUE, &dwFlag)) {
            if (bEmitStatus)
              infprintf("#Could not read the second part of the AuditOperations field!\n");
            free(lpszLine);
            return FALSE;
          }
          dwAuditOperations |= dwFlag;
          GetSetRemoveFlag(lpszValue, FALSE, &dwFlag);
          dwAuditOperations &= ~dwFlag;
        }

        /* If the active user equals the user being modified, then prevent this change if not the Administrator */
        if (lpActiveUser->idUser == lpCurrent->idUser && dwAuditOperations != lpCurrent->dwAuditOperations && CompareStrings(lpActiveUser->lpszName, "Administrator")) {
          if (bEmitStatus)
            infprintf("#Only the Administrator can change their own audited operations.\n");
          free(lpszLine);
          return FALSE;
        }

        /* Set the audited operations */
        lpCurrent->dwAuditOperations = dwAuditOperations;
      }

      /* If the above generated a formatting error, then break */
      if (bFormatError)
        break;

      /* Tell the user what we did */
      infprintf("*Modified user \"%s\" (%u)\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszName, STR_NUL_TERM), lpCurrent->idUser);
      safeStringsLen = origSafeStringsLen;
    }

    /* If more columns exist, then assert a formatting error */
    if (strtok2(NULL, cDelimiter) != NULL) {
      if (bEmitStatus) {
        infprintf("#Formatting error in input!  Too many columns exist.\n");
        fflush(stdout);
      }
      free(lpszLine);
      return FALSE;
    }
  }

  /* If we had a formatting error, then tell the user */
  if (bFormatError) {
    if (bEmitStatus) {
      infprintf("#Formatting error in input!  Not enough columns exist.\n");
      fflush(stdout);
    }
    free(lpszLine);
    return FALSE;
  }

  /* Everything was modified/added correctly :-)!  Rebuild the sort tables */
  if (bEmitStatus) {
    infprintf("#Refreshing sort tables...\n");
    fflush(stdout);
  }
  lpCurTable = lpIndex->lpSortTable;
  while (lpCurTable) {
    CreateSortTable(lpCurTable, lpIndex->lpDirEntry, lpIndex->lpClass);
    lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
  }

  /* We did it :-) */
  free(lpszLine);
  return TRUE;
}

/* Modify the given sort tables of an index file */
BOOL ModifySortTables(LPDIRLIST_INDEX lpIndex, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, UCHAR cDelimiter, BOOL bEmitStatus)
{
  /* Initialize variables */
  LPDIRLIST_SORTTABLE lpCurrent, lpCurTable;
  LPTSTR lpszLine = (LPTSTR) malloc(32);
  LPTSTR lpszCurrent;
  LPTSTR lpszValue;
  DWORD dwLineLen = 0, dwLineAlloc = 32, dwUpdateMode;
  BOOL bFormatError = FALSE, bFound;
  DWORD idSort;

  /* Tell the user that we're about to import the file */
  if (bEmitStatus) {
    infprintf("#Merging input into memory...\n");
    fflush(stdout);
  }

  /* Go through each line in the input (each line MUST have a delimiter) */
  while (ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead) && !bFormatError)
  {
    /* Break the string up into parts and process each one */
    lpszCurrent = strtok2(lpszLine, cDelimiter);
    if (!lpszCurrent) {
      bFormatError = TRUE;
      break;
    }

    /* Convert the current string into a sort ID and search for it */
    idSort = strtoul(lpszCurrent, NULL, 10);
    bFound = FALSE;
    lpCurrent = lpIndex->lpSortTable;
    while (!bFound && lpCurrent) {
      if (lpCurrent->idSort == idSort)
        bFound = TRUE;
      else
        lpCurrent = (LPDIRLIST_SORTTABLE) lpCurrent->next;
    }

    /* Process destruction and creation differently */
    if (bDelete)
    {
      /* If the sort table was not found, then give an error */
      if (!lpCurrent)
      {
        /* The class already does not exist (this code should never execute) */
        if (bEmitStatus)
          infprintf("#Sort table %u does not exist or already has been deleted.  Entire operation aborted.\n", idSort);
        free(lpszLine);
        return FALSE;
      }

      /* Tell the user what we are about to do */
      infprintf("*Deleted sort table %u\n", lpCurrent->idSort);

      /* Remove this sort table from the linked list */
      if (lpCurrent != lpIndex->lpSortTable) {
        lpCurTable = (LPDIRLIST_SORTTABLE) lpIndex->lpSortTable;
        while (lpCurTable->next != (LPBYTE) lpCurrent) {
          lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
          if (!lpCurTable)
            break;
        }
        if (!lpCurTable)
        {
          /* The class already does not exist (this code should never execute) */
          if (bEmitStatus)
            infprintf("#Sort table %u does not exist or already has been deleted.  This is the second-level warning, which should never appear.  Entire operation aborted.\n", idSort);
          free(lpszLine);
          return FALSE;
        }

        /* Pop this class out of the linked list */
        lpCurTable->next = (LPBYTE) lpCurrent->next;
        lpCurrent->next = NULL;
        FreeSortTables(lpCurrent);
      } else
      {
        /* Easy removal; it is at the beginning of the linked list */
        lpIndex->lpSortTable = (LPDIRLIST_SORTTABLE) lpCurrent->next;
        lpCurrent->next = NULL;
        FreeSortTables(lpCurrent);
      }
    } else
    {
      /* If we did not find it, then create an allocation and add it to the list */
      if (!bFound) {
        lpCurrent = (LPDIRLIST_SORTTABLE) malloc(sizeof(DIRLIST_SORTTABLE));
        memset(lpCurrent, 0, sizeof(DIRLIST_SORTTABLE));
        lpCurrent->idSort = idSort;
        lpCurrent->next = (LPBYTE) lpIndex->lpSortTable;
        lpIndex->lpSortTable = lpCurrent;
      }

      /* Get the sort order string */
      lpszCurrent = strtok2(NULL, cDelimiter);
      if (!lpszCurrent)
        bFormatError = TRUE;
      else {
        if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
          free(lpszLine);
          return FALSE;
      }
        if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
        {
          if (lpCurrent->lpszOrder)
            free(lpCurrent->lpszOrder);
          lpCurrent->lpszOrder = strdup(lpszValue);
        } else
          if (!lpCurrent->lpszOrder)
            lpCurrent->lpszOrder = strdup("0");
      }

      /* If the above generated a formatting error, then break */
      if (bFormatError)
        break;

      /* Tell the user what we did */
      infprintf("*Modified sort table %u\n", lpCurrent->idSort);
    }

    /* If more columns exist, then assert a formatting error */
    if (strtok2(NULL, cDelimiter) != NULL) {
      if (bEmitStatus) {
        infprintf("#Formatting error in input!  Too many columns exist.\n");
        fflush(stdout);
      }
      free(lpszLine);
      return FALSE;
    }
  }

  /* If we had a formatting error, then tell the user */
  if (bFormatError) {
    if (bEmitStatus) {
      infprintf("#Formatting error in input!  Not enough columns exist.\n");
      fflush(stdout);
    }
    free(lpszLine);
    return FALSE;
  }

  /* Everything was modified/added correctly :-)!  Rebuild the sort tables */
  if (bEmitStatus) {
    infprintf("#Refreshing sort tables...\n");
    fflush(stdout);
  }
  lpCurTable = lpIndex->lpSortTable;
  while (lpCurTable) {
    CreateSortTable(lpCurTable, lpIndex->lpDirEntry, lpIndex->lpClass);
    lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
  }

  /* We did it :-) */
  free(lpszLine);
  return TRUE;
}

/* Modify the given class of an index file (only does one-at-a-time for add/modify!) */
BOOL ModifyClasses(LPDIRLIST_INDEX lpIndex, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, DWORD idClass, UCHAR cDelimiter, BOOL bEmitStatus)
{
  /* Initialize variables */
  LPDIRLIST_SORTTABLE lpCurTable;
  LPDIRLIST_CLASS lpNew;
  LPTSTR lpszLine = (LPTSTR) malloc(32);
  LPTSTR lpszCurrent;
  LPTSTR lpszValue;
  DWORD dwLineLen, dwLineAlloc = 32, dwUpdateMode;
  BOOL bFormatError = FALSE;
  DWORD i;
  DWORD origSafeStringsLen = safeStringsLen;

  /* If the idClass parameter is not a power of 2, then return FALSE now */
  i = idClass;
  while (i > 1) {
    if (i % 2 != 0) {
      if (bEmitStatus) {
        infprintf("#Class ID %u is not valid; it must be 2 raised to any value between or including 0 to 31.\n", idClass);
        fflush(stdout);
      }
      free(lpszLine);
      return FALSE;
    } else
      i = i / 2;
  }

  /* If no class was given, then display an error */
  if (!idClass) {
    if (bEmitStatus) {
      infprintf("#You must specify a non-zero Class ID in the Extended-Fields (ef) parameter.\n");
      fflush(stdout);
    }
    free(lpszLine);
    return FALSE;
  }

  /* Tell the user that we're about to import the file */
  if (bEmitStatus) {
    infprintf("#Merging input into memory...\n");
    fflush(stdout);
  }
  /*anderson nichols was here!!!!!!*/

  /* If we are destructive, delete the class; otherwise, modify it */
  if (bDelete)
  {
    /* Read each line, which consists of a single number indicating what class to delete */
    LPDIRLIST_CLASS lpCurrent;
    while (ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead))
    {
      /* Fetch the current class, if any, with the same ID */
      lpNew = GetClass(lpIndex->lpClass, strtoul(lpszLine, NULL, 10));
      if (!lpNew)
      {
        /* The class already does not exist */
        if (bEmitStatus)
          infprintf("#Class %u does not exist or already has been deleted.  Entire operation aborted.\n", idClass);
        free(lpszLine);
        return FALSE;
      }

      /* Remove this class from the linked list */
      if (lpIndex->lpClass != lpNew) {
        lpCurrent = lpIndex->lpClass;
        while (lpCurrent->next != (LPBYTE) lpNew) {
          lpCurrent = (LPDIRLIST_CLASS) lpCurrent->next;
          if (!lpCurrent)
            break;
        }
        if (!lpCurrent)
        {
          /* The class already does not exist (this code should never execute) */
          if (bEmitStatus)
            infprintf("#Class %u does not exist or already has been deleted.  This is the second-level warning, which should never appear.  Entire operation aborted.\n", idClass);
          free(lpszLine);
          return FALSE;
        }

        /* Pop this class out of the linked list */
        infprintf("*Deleted class \"%s\" (%u)\n",
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszName, STR_NUL_TERM), lpCurrent->idClass);
        safeStringsLen = origSafeStringsLen;
        lpCurrent->next = (LPBYTE) lpNew->next;
        lpNew->next = NULL;
        FreeClasses(lpNew);
      } else
      {
        /* Easy removal; it is at the beginning of the linked list */
        infprintf("*Deleted class \"%s\" (%u)\n",
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpNew->lpszName, STR_NUL_TERM), lpNew->idClass);
        safeStringsLen = origSafeStringsLen;
        lpIndex->lpClass = (LPDIRLIST_CLASS) lpNew->next;
        lpNew->next = NULL;
        FreeClasses(lpNew);
      }
    }
  } else
  {
    /* Fetch the current class, if any, with the same ID */
    BOOL bEmitClassModify = TRUE;
    lpNew = GetClass(lpIndex->lpClass, idClass);
    if (!lpNew)
    {
      /* Allocate a new class and stick it at the end of the last class */
      LPDIRLIST_CLASS lpCurrent = lpIndex->lpClass;
      LPDIRLIST_CLASS lpPrev = NULL;
      lpNew = malloc(sizeof(DIRLIST_CLASS));
      memset(lpNew, 0, sizeof(DIRLIST_CLASS));
      lpNew->idClass = idClass;
      lpNew->next = NULL;
      while (lpCurrent) {
        lpPrev = lpCurrent;
        lpCurrent = (LPDIRLIST_CLASS) lpCurrent->next;
      }
      if (!lpPrev)
        lpIndex->lpClass = lpNew;
      else
        lpPrev->next = (LPBYTE) lpNew;
    }

    /* Get the name of the class */
    if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead)) {
      if (bEmitStatus)
        infprintf("#Could not read the Name field!\n");
      free(lpszLine);
      return FALSE;
    }
    if ((dwUpdateMode = GetUpdateMode(lpszLine, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
    free(lpszLine);
    return FALSE;
    }
    if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      lpNew->lpszName = strdup(lpszValue);
    else
      if (!lpNew->lpszName)
        lpNew->lpszName = strdup("");

    /* Get the flags */
    if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead)) {
      if (bEmitStatus)
        infprintf("#Could not read the Flags field!\n");
      free(lpszLine);
      return FALSE;
    }
    if ((dwUpdateMode = GetUpdateMode(lpszLine, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
      free(lpszLine);
      return FALSE;
    }
    if (dwUpdateMode == FLAG_REPLACE)
      lpNew->dwFlags = strtoul(lpszValue, NULL, 10);
    if (dwUpdateMode == FLAG_SET)
      lpNew->dwFlags |= strtoul(lpszValue, NULL, 10);
    if (dwUpdateMode == FLAG_REMOVE)
      lpNew->dwFlags &= ~strtoul(lpszValue, NULL, 10);
    if (dwUpdateMode == FLAG_SETREMOVE) {
      DWORD dwFlag;
      if (!GetSetRemoveFlag(lpszValue, TRUE, &dwFlag)) {
        if (bEmitStatus)
          infprintf("#Could not read the second part of the Flags field!\n");
        free(lpszLine);
        return FALSE;
      }
      lpNew->dwFlags |= dwFlag;
      GetSetRemoveFlag(lpszValue, FALSE, &dwFlag);
      lpNew->dwFlags &= ~dwFlag;
    }

    /* Read the default value string */
    if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead)) {
      if (bEmitStatus)
        infprintf("#Could not read the DefaultValue field!\n");
      free(lpszLine);
      return FALSE;
    }
    if ((dwUpdateMode = GetUpdateMode(lpszLine, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
      free(lpszLine);
      return FALSE;
    }
    if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
      lpNew->idDefaultValue = strtoul(lpszValue, NULL, 10);

    /* Get the number of values */
    if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead)) {
      if (bEmitStatus)
        infprintf("#Could not read the NumberOfValues field!\n");
      free(lpszLine);
      return FALSE;
    }
    if ((dwUpdateMode = GetUpdateMode(lpszLine, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
      free(lpszLine);
      return FALSE;
    }
    if (dwUpdateMode == FLAG_REPLACE)
    {
      /* Use replacement mode for the values */
      lpNew->nValues = strtoul(lpszValue, NULL, 10);

      /* Go through each line in the input (each line MUST have a delimiter) */
      i = 0;
      if (lpNew->lpValues)
        free(lpNew->lpValues);
      lpNew->lpValues = malloc(1);
      while (ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead) && (i < lpNew->nValues) && !bFormatError)
      {
        /* Break the string up into parts and process both */
        lpszCurrent = strtok2(lpszLine, cDelimiter);
        if (!lpszCurrent) {
          bFormatError = TRUE;
          break;
        }

        /* Add a new class value to the current allocation */
        lpNew->lpValues = realloc(lpNew->lpValues, (i + 1) * sizeof(DIRLIST_CLASSVALUE));
        lpNew->lpValues[i].idValue = strtoul(lpszCurrent, NULL, 10);

        /* Get the value string */
        lpszCurrent = strtok2(NULL, cDelimiter);
        if (!lpszCurrent) {
          bFormatError = TRUE;
          break;
        } else {
          if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
            free(lpszLine);
            return FALSE;
        }
          if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
            lpNew->lpValues[i].lpszValue = strdup(lpszValue);
          else
            lpNew->lpValues[i].lpszValue = strdup("");
        }
        i++;

        /* Update the user on our status */
        infprintf("*Modified value %u in class \"%s\" (%u)\n", lpNew->lpValues[i - 1].idValue,
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpNew->lpszName, STR_NUL_TERM), lpNew->idClass);
        safeStringsLen = origSafeStringsLen;
        bEmitClassModify = FALSE;
      }
    } else if (dwUpdateMode == FLAG_SET)
    {
      /* Add the given number of values */
      DWORD dwNumToAdd = strtoul(lpszValue, NULL, 10), dwNewID;
      LPDIRLIST_CLASSVALUE lpCurValue;

      /* Loop through the number of values given */
      i = 0;
      if (!lpNew->lpValues) {
        lpNew->lpValues = malloc(1);
        lpNew->nValues = 0;
      }
      while (ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead) && (i < dwNumToAdd) && !bFormatError)
      {
        /* Break the string up into parts and process both */
        lpszCurrent = strtok2(lpszLine, cDelimiter);
        if (!lpszCurrent) {
          bFormatError = TRUE;
          break;
        }

        /* Fetch the class value for this ID, or create one if one does not exist */
        if (!(lpCurValue = GetClassValue(lpNew, (dwNewID = strtoul(lpszCurrent, NULL, 10)))))
        {
          /* Allocate a new class value and set the value ID */
          (lpNew->nValues)++;
          lpNew->lpValues = realloc(lpNew->lpValues, lpNew->nValues * sizeof(DIRLIST_CLASSVALUE));
          lpCurValue = &(lpNew->lpValues[lpNew->nValues - 1]);
          lpCurValue->idValue = dwNewID;
          lpCurValue->lpszValue = NULL;
        }

        /* Get the value string */
        lpszCurrent = strtok2(NULL, cDelimiter);
        if (!lpszCurrent) {
          bFormatError = TRUE;
          break;
        } else {
          if ((dwUpdateMode = GetUpdateMode(lpszCurrent, &lpszValue, TRUE)) == FLAG_UNKNOWN) {
            free(lpszLine);
            return FALSE;
        }
          if ((dwUpdateMode == FLAG_SET) || (dwUpdateMode == FLAG_REPLACE))
            lpCurValue->lpszValue = strdup(lpszValue);
          else
            if (!lpCurValue->lpszValue)
              lpCurValue->lpszValue = strdup("");
        }
        i++;

        /* Update the user on our status */
        infprintf("*Modified value %u in class \"%s\" (%u)\n", dwNewID,
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpNew->lpszName, STR_NUL_TERM), lpNew->idClass);
        safeStringsLen = origSafeStringsLen;
        bEmitClassModify = FALSE;
      }
    } else if (dwUpdateMode == FLAG_REMOVE)
    {
      /* Remove the given number of values */
      DWORD dwNumToRemove = strtoul(lpszValue, NULL, 10), dwOldID, t;
      BOOL bFound;

      /* Loop through the number of values given */
      i = 0;
      if (!lpNew->lpValues || (lpNew->nValues < dwNumToRemove)) {
        infprintf("#Value does not exist or was already deleted (this class does not have the requested number of values to delete)\n");
        free(lpszLine);
        return FALSE;
      }
      while (ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead) && (i < dwNumToRemove) && !bFormatError)
      {
        /* Get the current value ID */
        dwOldID = strtoul(lpszLine, NULL, 10);

        /* Find the class and then delete it */
        bFound = FALSE;
        t = 0;
        while (!bFound && (t < lpNew->nValues)) {
          if (lpNew->lpValues[t].idValue == dwOldID)
            bFound = TRUE;
          else
            t++;
        }
        if (!bFound) {
          infprintf("#Class value with ID %u was not found in the list within \"%s\"\n", dwOldID,
                    SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpNew->lpszName, STR_NUL_TERM));
          safeStringsLen = origSafeStringsLen;
          free(lpszLine);
          return FALSE;
        }

        /* Update the user on our status */
        infprintf("*Deleted value %u in class \"%s\" (%u)\n", lpNew->lpValues[t].idValue,
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpNew->lpszName, STR_NUL_TERM), lpNew->idClass);
        safeStringsLen = origSafeStringsLen;
        bEmitClassModify = FALSE;

        /* Free the string for this value if one exists */
        if (lpNew->lpValues[t].lpszValue)
          free(lpNew->lpValues[t].lpszValue);

        /* If this value was not at the end, copy the last value to the current */
        if (i != lpNew->nValues - 1)
          memcpy(&(lpNew->lpValues[t]), &(lpNew->lpValues[lpNew->nValues - 1]), sizeof(DIRLIST_CLASSVALUE));

        /* Pop off the last value */
        (lpNew->nValues)--;

        /* Next value to delete */
        i++;
      }
    }

    /* Tell the user what happened */
    if (bEmitClassModify)
      infprintf("*Modified class \"%s\" (%u)\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpNew->lpszName, STR_NUL_TERM), lpNew->idClass);
    else
      infprintf("#Modified class \"%s\" (%u)\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpNew->lpszName, STR_NUL_TERM), lpNew->idClass);
    safeStringsLen = origSafeStringsLen;

    /* If we had a formatting error, then tell the user */
    if (bFormatError) {
      if (bEmitStatus) {
        infprintf("#Formatting error in input!  Not enough columns exist.\n");
        fflush(stdout);
      }
      free(lpszLine);
      return FALSE;
    }

    /* If more columns exist, then assert a formatting error */
    if (strtok2(NULL, cDelimiter) != NULL) {
      if (bEmitStatus) {
        infprintf("#Formatting error in input!  Too many columns exist.\n");
        fflush(stdout);
      }
      free(lpszLine);
      return FALSE;
    }
  }

  /* Everything was modified/added/deleted correctly :-)!  Rebuild the sort tables */
  if (bEmitStatus) {
    infprintf("#Refreshing sort tables...\n");
    fflush(stdout);
  }
  lpCurTable = lpIndex->lpSortTable;
  while (lpCurTable) {
    CreateSortTable(lpCurTable, lpIndex->lpDirEntry, lpIndex->lpClass);
    lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
  }

  /* We did it :-) */
  free(lpszLine);
  return TRUE;
}

/* Modify an index file to reflect data in the input text file */
BOOL ModifyEntries(LPDIRLIST_INDEX lpIndex, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_ACCESSPLAN lpPlan, UCHAR cDelimiter, BOOL bEmitStatus)
{
  /* Initialize variables */
  LPDIRLIST_SORTTABLE lpCurTable;
  LPDIRLIST_ENTRY lpCurrent = NULL, lpPrev = NULL;
  LPTSTR lpszLine = (LPTSTR) malloc(32);
  LPTSTR lpszCurrent;
  DWORD dwLineLen = 0, dwLineAlloc = 32;
  BOOL bFormatError = FALSE;
  DWORD origSafeStringsLen = safeStringsLen;

  /* Make sure DISPLAY_USERNAME is set; otherwise, this function cannot operate */
  if (!(lpPlan->dwFields & DISPLAY_USERNAME)) {
    if (bEmitStatus)
      infprintf("#The DISPLAY_USERNAME flag MUST be used and the user names MUST be in the input file to modify the index!\n");
    free(lpszLine);
    return FALSE;
  }

  /* If we have any searches, then this is the wrong function */
  if (lpPlan->nSearches) {
    if (bEmitStatus)
      infprintf("#ModifyEntries() cannot be used to modify multiple rows.  Use ModifyEntriesToMatchOne() instead.\n");
    free(lpszLine);
    return FALSE;
  }

  /* Tell the user that we're about to import the file */
  if (bEmitStatus) {
    infprintf("#Merging input into memory...\n");
    fflush(stdout);
  }

  /* Go through each line in the input (each line MUST have a delimiter) */
  while (ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead) && !bFormatError)
  {
    /* Break the string up into parts and process each one */
    lpszCurrent = strtok2(lpszLine, cDelimiter);
    if (!lpszCurrent) {
      bFormatError = TRUE;
      break;
    }

    /* Fetch the entry using the username (always the first parameter) */
    FetchEntry(&(lpIndex->lpDirEntry), &lpPrev, &lpCurrent, lpszCurrent, TRUE);

    /* If this entry has FLAGS_EXCLUDE set and the user is not in admin mode, then give an error */
    if ((lpCurrent->dwFlags & FLAGS_EXCLUDE) && (!lpPlan->bAdminMode)) {
      if (bEmitStatus) {
        infprintf("#Security violation adding/modifying \"%s\"\n",
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszUserName, STR_NUL_TERM));
        safeStringsLen = origSafeStringsLen;
      }
      free(lpszLine);
      return FALSE;
    }

    /* If we are being creative instead of destructive, then get the other parameters */
    if (!bDelete)
    {
      /* Check to see if the names are separate, if any */
      if (lpPlan->dwFields & DISPLAY_SEPARATENAMES)
      {
        /* Check to see if which name is first */
        if ((lpPlan->dwFields & DISPLAY_LASTNAMEFIRST) == DISPLAY_LASTNAMEFIRST)
        {
          /* Get the last name, then first name */
          lpszCurrent = strtok2(NULL, cDelimiter);
          if (!lpszCurrent)
            bFormatError = TRUE;
          else {
            if (lpCurrent->lpszLastName)
              free(lpCurrent->lpszLastName);
            lpCurrent->lpszLastName = strdup(lpszCurrent);
            lpszCurrent = strtok2(NULL, cDelimiter);
            if (!lpszCurrent)
              bFormatError = TRUE;
            else {
              if (lpCurrent->lpszFirstName)
                free(lpCurrent->lpszFirstName);
              lpCurrent->lpszFirstName = strdup(lpszCurrent);
            }
          }
        } else
        {
          /* Get the first name, then last name */
          lpszCurrent = strtok2(NULL, cDelimiter);
          if (!lpszCurrent)
            bFormatError = TRUE;
          else {
            if (lpCurrent->lpszFirstName)
              free(lpCurrent->lpszFirstName);
            lpCurrent->lpszFirstName = strdup(lpszCurrent);
            lpszCurrent = strtok2(NULL, cDelimiter);
            if (!lpszCurrent)
              bFormatError = TRUE;
            else {
              if (lpCurrent->lpszLastName)
                free(lpCurrent->lpszLastName);
              lpCurrent->lpszLastName = strdup(lpszCurrent);
            }
          }
        }
      } else
      {
        /* See if both the first and last names are given together */
        if ((lpPlan->dwFields & DISPLAY_FIRSTNAME) && (lpPlan->dwFields & DISPLAY_LASTNAME))
        {
          /* Free any existing allocations */
          if (lpCurrent->lpszFirstName)
            free(lpCurrent->lpszFirstName);
          if (lpCurrent->lpszLastName)
            free(lpCurrent->lpszLastName);

          /* Get the string, split it, and set it */
          lpszCurrent = strtok2(NULL, cDelimiter);
          if (!lpszCurrent)
            bFormatError = TRUE;
          else
          {
            /* Lastname, Firstname import mode */
            if ((lpPlan->dwFields & DISPLAY_LASTNAMEFIRST) == DISPLAY_LASTNAMEFIRST)
            {
              /* Last name is first -- the comma position + 2 is the first name */
              LPTSTR lpszFirstName = strrchr(lpszCurrent, ',');
              if (!lpszFirstName)
              {
                /* Use whatever we have as the last name, and set the first name to "" */
                lpCurrent->lpszLastName = strdup(lpszCurrent);
                lpCurrent->lpszFirstName = malloc(1);
                lpCurrent->lpszFirstName[0] = 0;
              } else
              {
                /* Set the first name (remove any leading spaces too) */
                lpszFirstName[0] = 0;
                do {
                  lpszFirstName++;
                } while (lpszFirstName[0] == ' ');
                lpCurrent->lpszFirstName = strdup(lpszFirstName);

                /* Set the last name */
                lpCurrent->lpszLastName = strdup(lpszCurrent);
              }
            } else

              /* Firstname Lastname import mode */
            {
              LPTSTR lpszLastName = GetLastName(lpszCurrent);
              if (lpszLastName) {
                lpCurrent->lpszLastName = strdup(lpszLastName);
                lpszCurrent[lpszLastName - lpszCurrent - 1] = 0;
                lpCurrent->lpszFirstName = strdup(lpszCurrent);
              } else {
                lpCurrent->lpszLastName = strdup(lpszCurrent);
                lpCurrent->lpszFirstName = malloc(1);
                lpCurrent->lpszFirstName[0] = 0;
              }
            }
          }
        }

        /* See if only the first name is to be set */
        if ((lpPlan->dwFields & DISPLAY_FIRSTNAME) && !(lpPlan->dwFields & DISPLAY_LASTNAME)) {
          if (lpCurrent->lpszFirstName)
            free(lpCurrent->lpszFirstName);
          lpszCurrent = strtok2(NULL, cDelimiter);
          if (!lpszCurrent)
            bFormatError = TRUE;
          else
            lpCurrent->lpszFirstName = strdup(lpszCurrent);
        }

        /* See if only the last name is to be set */
        if (!(lpPlan->dwFields & DISPLAY_FIRSTNAME) && (lpPlan->dwFields & DISPLAY_LASTNAME)) {
          if (lpCurrent->lpszLastName)
            free(lpCurrent->lpszLastName);
          lpszCurrent = strtok2(NULL, cDelimiter);
          if (!lpszCurrent)
            bFormatError = TRUE;
          else
            lpCurrent->lpszLastName = strdup(lpszCurrent);
        }
      }

      /* If the above generated a formatting error, then break */
      if (bFormatError)
        break;

      /* See if the flags need to be set */
      if (lpPlan->dwFields & DISPLAY_FLAGS)
      {
        /* Get each flag */
        DWORD i = (lpPlan->bAdminMode)? 2: 4;
        while ((i <= FLAGS_NOTELEPHONE) && !bFormatError) {
          lpszCurrent = strtok2(NULL, cDelimiter);
          if (!lpszCurrent)
            bFormatError = TRUE;
          else {
            switch (lpPlan->wFlagModifyType)
            {
            case FLAG_SET :
              {
                if (strtoul(lpszCurrent, NULL, 10))
                  lpCurrent->dwFlags |= i;
                break;
              }

            case FLAG_REMOVE :
              {
                if (strtoul(lpszCurrent, NULL, 10))
                  lpCurrent->dwFlags &= ~i;
                break;
              }

            case FLAG_SETREMOVE :
              {
                DWORD dwFlag;
                if (!GetSetRemoveFlag(lpszCurrent, TRUE, &dwFlag)) {
                  if (bEmitStatus)
                    infprintf("#Could not read the second part of the Flags field!\n");
                  free(lpszLine);
                  return FALSE;
                }
                if (dwFlag)
                  lpCurrent->dwFlags |= i;
                GetSetRemoveFlag(lpszCurrent, FALSE, &dwFlag);
                if (dwFlag)
                  lpCurrent->dwFlags &= ~i;
                break;
              }

            default :
              {
                if (strtoul(lpszCurrent, NULL, 10))
                  lpCurrent->dwFlags |= i;
                else
                  lpCurrent->dwFlags &= ~i;
                break;
              }
            }
          }

          /* Next flag */
          i *= 2;
        }
      }

      /* If the above generated a formatting error, then break */
      if (bFormatError)
        break;

      /* Read the website if it's in the column list */
      if (lpPlan->dwFields & DISPLAY_WEBSITE) {
        if (lpCurrent->lpszWebSite)
          free(lpCurrent->lpszWebSite);
        lpszCurrent = strtok2(NULL, cDelimiter);
        if (!lpszCurrent)
          bFormatError = TRUE;
        else
          lpCurrent->lpszWebSite = strdup(lpszCurrent);
      }
      if (bFormatError)
        break;

      /* Read the email value if it's in the column list */
      if (lpPlan->dwFields & DISPLAY_EMAIL) {
        if (lpCurrent->lpszEMail)
          free(lpCurrent->lpszEMail);
        lpszCurrent = strtok2(NULL, cDelimiter);
        if (!lpszCurrent)
          bFormatError = TRUE;
        else
          lpCurrent->lpszEMail = strdup(lpszCurrent);
      }
      if (bFormatError)
        break;

      /* Read the telephone number if it's in the column list */
      if (lpPlan->dwFields & DISPLAY_TELEPHONE) {
        if (lpCurrent->lpszTelephone)
          free(lpCurrent->lpszTelephone);
        lpszCurrent = strtok2(NULL, cDelimiter);
        if (!lpszCurrent)
          bFormatError = TRUE;
        else
          lpCurrent->lpszTelephone = strdup(lpszCurrent);
      }
      if (bFormatError)
        break;

      /* Read the class entries if requested */
      if (lpPlan->dwExtFields)
      {
        DWORD i = 1;
        while ((i <= MAX_CLASS) && !bFormatError)
        {
          /* If this attribute is requested for import, then get the attribute */
          if (lpPlan->dwExtFields & i) {
            LPDIRLIST_CLASSDATA lpClassData = GetClassData(lpCurrent, i);
            DWORD dwOrigFlags = (DWORD) NULL;

            /* If the value isn't already defined, then create one */
            if (!lpClassData) {
              if (lpCurrent->lpClassData)
                lpCurrent->lpClassData = (LPDIRLIST_CLASSDATA) realloc(lpCurrent->lpClassData, (++lpCurrent->nClassDataValues) * sizeof(DIRLIST_CLASSDATA));
              else
                lpCurrent->lpClassData = (LPDIRLIST_CLASSDATA) malloc((++lpCurrent->nClassDataValues) * sizeof(DIRLIST_CLASSDATA));
              lpClassData = &(lpCurrent->lpClassData[lpCurrent->nClassDataValues - 1]);
              memset(lpClassData, 0, sizeof(DIRLIST_CLASSDATA));
              lpClassData->idClass = i;
            } else
              dwOrigFlags = lpClassData->dwFlags;
            /* else
            {*/
              /* Free any previously allocated memory in the class data */
            /*  if (lpClassData->lpszValue && (lpClassData->dwFlags & CLASSDATA_OTHER))
                free(lpClassData->lpszValue);
            }*/

            /* Get the flags if requested */
            if (lpPlan->dwExtFlags) {
              lpszCurrent = strtok2(NULL, cDelimiter);
              if (!lpszCurrent)
                bFormatError = TRUE;
              else {
                if (!strtoul(lpszCurrent, NULL, 10))
                  lpClassData->idClass = 0;
                lpszCurrent = strtok2(NULL, cDelimiter);
                if (!lpszCurrent)
                  bFormatError = TRUE;
                else {
                  if (strtoul(lpszCurrent, NULL, 10))
                    lpClassData->dwFlags |= CLASSDATA_OTHER;
                  else
                    lpClassData->dwFlags &= ~CLASSDATA_OTHER;
                }
              }
            }

            /* If both values are given, then set them */
            if ((lpPlan->dwExtBoth & i) && !bFormatError) {
              lpszCurrent = strtok2(NULL, cDelimiter);
              if (!lpszCurrent)
                bFormatError = TRUE;
              else {
                lpClassData->dwValue = strtoul(lpszCurrent, NULL, 10);
                lpszCurrent = strtok2(NULL, cDelimiter);
                if (!lpszCurrent)
                  bFormatError = TRUE;
                else
                {
                  /* Free any previously allocated memory in the class data */
                  if (lpClassData->lpszValue && (dwOrigFlags & CLASSDATA_OTHER))
                    free(lpClassData->lpszValue);

                  /* If this value has the "other" flag set, then set the string; */
                  /* otherwise, ignore any string that exists in that column */
                  if (lpClassData->dwFlags & CLASSDATA_OTHER)
                    lpClassData->lpszValue = strdup(lpszCurrent);
                  else
                    lpClassData->lpszValue = NULL;
                }
              }
            }

            /* If the value numbers are to be imported, then do so */
            if ((lpPlan->dwExtValues & i) && !(lpPlan->dwExtBoth & i) && !bFormatError) {
              lpszCurrent = strtok2(NULL, cDelimiter);
              if (!lpszCurrent)
                bFormatError = TRUE;
              else
                lpClassData->dwValue = strtoul(lpszCurrent, NULL, 10);
            }

            /* If the string representation is imported instead, then do so */
            if (!(lpPlan->dwExtValues & i) && !(lpPlan->dwExtBoth & i) && !bFormatError) {
              lpszCurrent = strtok2(NULL, cDelimiter);
              if (!lpszCurrent)
                bFormatError = TRUE;
              else
              {
                /* Free any previously allocated memory in the class data */
                if (lpClassData->lpszValue && (dwOrigFlags & CLASSDATA_OTHER))
                  free(lpClassData->lpszValue);

                /* If this value has the "other" flag set, then set the string; */
                /* otherwise, ignore any string that exists in that column */
                if (lpClassData->dwFlags & CLASSDATA_OTHER)
                  lpClassData->lpszValue = strdup(lpszCurrent);
                else
                  lpClassData->lpszValue = NULL;
              }
            }
          }

          /* Increment safely */
          if (i == MAX_CLASS)
            i++;
          else
            i *= 2;
        }
      }
      if (bFormatError)
        break;

      /* Tell the user what we did */
      infprintf("*Modified user \"%s %s\" (%s)\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszFirstName, STR_NUL_TERM),
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszLastName, STR_NUL_TERM),
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszUserName, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
    } else
    {
      /* If a previous entry exists, then pop this entry off of the linked list */
      infprintf("*Deleted user \"%s %s\" (%s)\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszFirstName, STR_NUL_TERM),
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszLastName, STR_NUL_TERM),
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszUserName, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
      if (lpPrev) {
        lpPrev->next = lpCurrent->next;
        lpCurrent->next = NULL;
        FreeEntries(lpCurrent);
        lpCurrent = (LPDIRLIST_ENTRY) lpPrev->next;
      } else {
        lpIndex->lpDirEntry = (LPDIRLIST_ENTRY) lpCurrent->next;
        lpCurrent->next = NULL;
        FreeEntries(lpCurrent);
        lpCurrent = lpIndex->lpDirEntry;
      }
    }

    /* If more columns exist, then assert a formatting error */
    if (strtok2(NULL, cDelimiter) != NULL) {
      if (bEmitStatus) {
        infprintf("#Formatting error in input!  Too many columns exist.\n");
        fflush(stdout);
      }
      free(lpszLine);
      return FALSE;
    }
  }

  /* If we had a formatting error, then tell the user */
  if (bFormatError) {
    if (bEmitStatus) {
      infprintf("#Formatting error in input!  Not enough columns exist.\n");
      fflush(stdout);
    }
    free(lpszLine);
    return FALSE;
  }

  /* Everything was modified/added/deleted correctly :-)!  Rebuild the sort tables */
  if (bEmitStatus) {
    infprintf("#Refreshing sort tables...\n");
    fflush(stdout);
  }
  lpCurTable = lpIndex->lpSortTable;
  while (lpCurTable) {
    CreateSortTable(lpCurTable, lpIndex->lpDirEntry, lpIndex->lpClass);
    lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
  }

  /* We did it :-) */
  free(lpszLine);
  return TRUE;
}

/* Modify an index file to reflect data in one given entry (a multiple modify, like MMX ;-) */
BOOL ModifyEntriesToMatchOne(LPDIRLIST_INDEX lpIndex, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_ACCESSPLAN lpPlan, UCHAR cDelimiter, BOOL bEmitStatus)/*(LPCTSTR lpszIndexName, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_ACCESSPLAN lpPlan, BOOL bEmitStatus)*/
{
  /* Initialize variables */
  LPDIRLIST_SORTTABLE lpCurTable;
  LPDIRLIST_ENTRY lpCurrent = NULL, lpPrev = NULL;
  DIRLIST_ENTRY srcEntry;
  LPTSTR lpszLine = (LPTSTR) malloc(32);
  LPTSTR lpszCurrent;
  DWORD dwLineLen = 0, dwLineAlloc = 32;
  DWORD dwRemoveFlags = 0;
  BOOL bFormatError = FALSE;
  BOOL bFirst = TRUE;
  BOOL bIncrement;
  DWORD origSafeStringsLen = safeStringsLen;

  /* Tell the user that we're about to import the file */
  if (bEmitStatus) {
    infprintf("#Merging input into memory...\n");
    fflush(stdout);
  }

  /* If the username is being modified, ensure it doesn't exist */
  if ((lpPlan->dwFields & DISPLAY_USERNAME) && !bDelete)
  {
    /* Oh dear!  Continuing this operation would violate primary key rules! */
    if (bEmitStatus)
      infprintf("#Primary key violation; usernames must be unique.\n");
    free(lpszLine);
    return FALSE;
  }

  /* Clear the source entry */
  memset(&srcEntry, 0, sizeof(DIRLIST_ENTRY));

  /* Read the input line if we're not deleting */
  if (!bDelete)
  {
    if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead)) {
      if (bEmitStatus)
        infprintf("#Could not read the text line!\n");
      free(lpszLine);
      return FALSE;
    }
  }

  /* Read the model from stdin if we're modifying */
  if (!bDelete)
  {
    /* Tell the user that we're about to import the file */
    if (bEmitStatus) {
      infprintf("#Updating entries using given entry and search pattern...\n");
      fflush(stdout);
    }

    /* Read the username if it's in the column list (will only execute if we're deleting) */
    if (lpPlan->dwFields & DISPLAY_USERNAME) {
      lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
      if (bFirst)
        bFirst = FALSE;
      if (!lpszCurrent)
        bFormatError = TRUE;
      else
        srcEntry.lpszUserName = strdup(lpszCurrent);
    }

    /* Check to see if the names are separate, if any */
    if (!bFormatError && (lpPlan->dwFields & DISPLAY_SEPARATENAMES))
    {
      /* Check to see if which name is first */
      if ((lpPlan->dwFields & DISPLAY_LASTNAMEFIRST) == DISPLAY_LASTNAMEFIRST)
      {
        /* Get the last name, then first name */
        lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
        if (bFirst)
          bFirst = FALSE;
        if (!lpszCurrent)
          bFormatError = TRUE;
        else {
          srcEntry.lpszLastName = strdup(lpszCurrent);
          lpszCurrent = strtok2(NULL, cDelimiter);
          if (!lpszCurrent)
            bFormatError = TRUE;
          else
            srcEntry.lpszFirstName = strdup(lpszCurrent);
        }
      } else
      {
        /* Get the first name, then last name */
        lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
        if (bFirst)
          bFirst = FALSE;
        if (!lpszCurrent)
          bFormatError = TRUE;
        else {
          srcEntry.lpszFirstName = strdup(lpszCurrent);
          lpszCurrent = strtok2(NULL, cDelimiter);
          if (!lpszCurrent)
            bFormatError = TRUE;
          else
            srcEntry.lpszLastName = strdup(lpszCurrent);
        }
      }
    } else
    {
      /* See if both the first and last names are given together */
      if ((lpPlan->dwFields & DISPLAY_FIRSTNAME) && (lpPlan->dwFields & DISPLAY_LASTNAME))
      {
        /* Get the string, split it, and set it */
        lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
        if (bFirst)
          bFirst = FALSE;
        if (!lpszCurrent)
          bFormatError = TRUE;
        else
        {
          /* Lastname, Firstname import mode */
          if ((lpPlan->dwFields & DISPLAY_LASTNAMEFIRST) == DISPLAY_LASTNAMEFIRST)
          {
            /* Last name is first -- the comma position + 2 is the first name */
            LPTSTR lpszFirstName = strrchr(lpszCurrent, ',');
            if (!lpszFirstName)
            {
              /* Use whatever we have as the last name, and set the first name to "" */
              srcEntry.lpszLastName = strdup(lpszCurrent);
              srcEntry.lpszFirstName = malloc(1);
              srcEntry.lpszFirstName[0] = 0;
            } else
            {
              /* Set the first name (remove any leading spaces too) */
              lpszFirstName[0] = 0;
              do {
                lpszFirstName++;
              } while (lpszFirstName[0] == ' ');
              srcEntry.lpszFirstName = strdup(lpszFirstName);

              /* Set the last name */
              srcEntry.lpszLastName = strdup(lpszCurrent);
            }
          } else
          {
            /* Firstname Lastname import mode */
            LPTSTR lpszLastName = GetLastName(lpszCurrent);
            if (lpszLastName) {
              srcEntry.lpszLastName = strdup(lpszLastName);
              lpszCurrent[lpszLastName - lpszCurrent - 1] = 0;
              srcEntry.lpszFirstName = strdup(lpszCurrent);
            } else {
              srcEntry.lpszLastName = strdup(lpszCurrent);
              srcEntry.lpszFirstName = malloc(1);
              srcEntry.lpszFirstName[0] = 0;
            }
          }
        }
      }

      /* See if only the first name is to be set */
      if ((lpPlan->dwFields & DISPLAY_FIRSTNAME) && !(lpPlan->dwFields & DISPLAY_LASTNAME)) {
        lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
        if (bFirst)
          bFirst = FALSE;
        if (!lpszCurrent)
          bFormatError = TRUE;
        else
          srcEntry.lpszFirstName = strdup(lpszCurrent);
      }

      /* See if only the last name is to be set */
      if (!(lpPlan->dwFields & DISPLAY_FIRSTNAME) && (lpPlan->dwFields & DISPLAY_LASTNAME)) {
        lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
        if (bFirst)
          bFirst = FALSE;
        if (!lpszCurrent)
          bFormatError = TRUE;
        else
          srcEntry.lpszLastName = strdup(lpszCurrent);
      }
    }

    /* See if the flags need to be set */
    if (!bFormatError && (lpPlan->dwFields & DISPLAY_FLAGS))
    {
      /* Get each flag */
      DWORD i = (lpPlan->bAdminMode)? 2: 4;
      while ((i <= FLAGS_NOTELEPHONE) && !bFormatError) {
        lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
        if (bFirst)
          bFirst = FALSE;
        if (!lpszCurrent)
          bFormatError = TRUE;
        else {
          switch (lpPlan->wFlagModifyType)
          {
          case FLAG_SET :
            {
              if (strtoul(lpszCurrent, NULL, 10))
                srcEntry.dwFlags |= i;
              break;
            }

          case FLAG_REMOVE :
            {
              if (strtoul(lpszCurrent, NULL, 10))
                dwRemoveFlags |= i;
              break;
            }

          case FLAG_SETREMOVE :
            {
              DWORD dwFlag;
              if (!GetSetRemoveFlag(lpszCurrent, TRUE, &dwFlag)) {
                if (bEmitStatus)
                  infprintf("#Could not read the second part of the Flags field!\n");
                free(lpszLine);
                return FALSE;
              }
              if (dwFlag)
                srcEntry.dwFlags |= i;
              GetSetRemoveFlag(lpszCurrent, FALSE, &dwFlag);
              if (dwFlag)
                dwRemoveFlags |= i;
              break;
            }

          default :
            {
              if (strtoul(lpszCurrent, NULL, 10))
                srcEntry.dwFlags |= i;
              else
                dwRemoveFlags |= i;
              break;
            }
          }
        }

        /* Next flag */
        i *= 2;
      }
    }

    /* Read the website if it's in the column list */
    if (!bFormatError && (lpPlan->dwFields & DISPLAY_WEBSITE)) {
      lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
      if (bFirst)
        bFirst = FALSE;
      if (!lpszCurrent)
        bFormatError = TRUE;
      else
        srcEntry.lpszWebSite = strdup(lpszCurrent);
    }

    /* Read the email value if it's in the column list */
    if (!bFormatError && (lpPlan->dwFields & DISPLAY_EMAIL)) {
      lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
      if (bFirst)
        bFirst = FALSE;
      if (!lpszCurrent)
        bFormatError = TRUE;
      else
        srcEntry.lpszEMail = strdup(lpszCurrent);
    }

    /* Read the telephone number if it's in the column list */
    if (!bFormatError && (lpPlan->dwFields & DISPLAY_TELEPHONE)) {
      lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
      if (bFirst)
        bFirst = FALSE;
      if (!lpszCurrent)
        bFormatError = TRUE;
      else
        srcEntry.lpszTelephone = strdup(lpszCurrent);
    }

    /* Read the class entries if requested */
    /* DWORD dwExtColumns, DWORD dwExtValues, DWORD dwExtBoth, DWORD dwExtFlags; */
    if (!bFormatError && (lpPlan->dwExtFields))
    {
      DWORD i = 1;
      while ((i <= MAX_CLASS) && !bFormatError)
      {
        /* If this attribute is requested for import, then get the attribute */
        if (lpPlan->dwExtFields & i) {
          LPDIRLIST_CLASSDATA lpClassData = NULL;

          /* If the value isn't already defined, then create one */
          if (srcEntry.lpClassData)
            srcEntry.lpClassData = (LPDIRLIST_CLASSDATA) realloc(srcEntry.lpClassData, (++srcEntry.nClassDataValues) * sizeof(DIRLIST_CLASSDATA));
          else
            srcEntry.lpClassData = (LPDIRLIST_CLASSDATA) malloc((++srcEntry.nClassDataValues) * sizeof(DIRLIST_CLASSDATA));
          lpClassData = &(srcEntry.lpClassData[srcEntry.nClassDataValues - 1]);
          memset(lpClassData, 0, sizeof(DIRLIST_CLASSDATA));
          lpClassData->idClass = i;

          /* Get the flags if requested */
          if (lpPlan->dwExtFlags) {
            lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
            if (bFirst)
              bFirst = FALSE;
            if (!lpszCurrent)
              bFormatError = TRUE;
            else {
              if (!strtoul(lpszCurrent, NULL, 10))
                lpClassData->idClass = 0;
              lpszCurrent = strtok2(NULL, cDelimiter);
              if (!lpszCurrent)
                bFormatError = TRUE;
              else {
                if (strtoul(lpszCurrent, NULL, 10))
                  lpClassData->dwFlags |= CLASSDATA_OTHER;
                else
                  lpClassData->dwFlags &= ~CLASSDATA_OTHER;
              }
            }
          }

          /* If both values are given, then set them */
          if ((lpPlan->dwExtBoth & i) && !bFormatError) {
            lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
            if (bFirst)
              bFirst = FALSE;
            if (!lpszCurrent)
              bFormatError = TRUE;
            else {
              lpClassData->dwValue = strtoul(lpszCurrent, NULL, 10);
              lpszCurrent = strtok2(NULL, cDelimiter);
              if (!lpszCurrent)
                bFormatError = TRUE;
              else
              {
                /* If this value has the "other" flag set, then set the string; */
                /* otherwise, ignore any string that exists in that column */
                if (lpClassData->dwFlags & CLASSDATA_OTHER)
                  lpClassData->lpszValue = strdup(lpszCurrent);
                else
                  lpClassData->lpszValue = NULL;
              }
            }
          }

          /* If the value numbers are to be imported, then do so */
          if ((lpPlan->dwExtValues & i) && !(lpPlan->dwExtBoth & i) && !bFormatError) {
            lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
            if (bFirst)
              bFirst = FALSE;
            if (!lpszCurrent)
              bFormatError = TRUE;
            else
              lpClassData->dwValue = strtoul(lpszCurrent, NULL, 10);
          }

          /* If the string representation is imported instead, then do so */
          if (!(lpPlan->dwExtValues & i) && !(lpPlan->dwExtBoth & i) && !bFormatError) {
            lpszCurrent = strtok2((bFirst)? lpszLine: NULL, cDelimiter);
            if (bFirst)
              bFirst = FALSE;
            if (!lpszCurrent)
              bFormatError = TRUE;
            else
            {
              /* If this value has the "other" flag set, then set the string; */
              /* otherwise, ignore any string that exists in that column */
              if (lpClassData->dwFlags & CLASSDATA_OTHER)
                lpClassData->lpszValue = strdup(lpszCurrent);
              else
                lpClassData->lpszValue = NULL;
            }
          }
        }

        /* Increment safely */
        if (i == MAX_CLASS)
          i++;
        else
          i *= 2;
      }
    }

    /* If more columns exist, then assert a formatting error */
    if (!bFormatError && (strtok2((bFirst)? lpszLine: NULL, cDelimiter) != NULL)) {
      if (bEmitStatus) {
        infprintf("#Formatting error in input!  Too many columns exist.\n");
        fflush(stdout);
      }
      free(lpszLine);
      return FALSE;
    }

    /* If we had a formatting error, then tell the user */
    if (bFormatError) {
      if (bEmitStatus) {
        infprintf("#Formatting error in input!  Not enough columns exist.\n");
        fflush(stdout);
      }
      free(lpszLine);
      return FALSE;
    }
  }

  /* Walk the entries and modify those that match the search criteria */
  lpCurrent = lpIndex->lpDirEntry;
  lpPrev = lpCurrent;
  while (lpCurrent)
  {
    /* If this entry has FLAGS_EXCLUDE set and we are not in admin mode, then skip this one */
    if ((lpCurrent->dwFlags & FLAGS_EXCLUDE) && (!lpPlan->bAdminMode)) {
      lpPrev = lpCurrent;
      lpCurrent = (LPDIRLIST_ENTRY) lpCurrent->next;
      continue;
    }

    /* Check the relevance of this entry (make sure */
    bIncrement = TRUE;
    if (CheckRelevance(lpCurrent, lpIndex->lpClass, lpPlan))
    {
      /* Check to see if we're creative or destructive */
      if (bDelete)
      {
        /* Tell the user what we about to do */
        infprintf("*Deleted user \"%s %s\" (%s)\n",
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszFirstName, STR_NUL_TERM),
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszLastName, STR_NUL_TERM),
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszUserName, STR_NUL_TERM));
        safeStringsLen = origSafeStringsLen;

        /* If the current entry is the root entry, then just pop off the top */
        if (lpCurrent == lpIndex->lpDirEntry) {
          lpCurrent = (LPDIRLIST_ENTRY) lpCurrent->next;
          lpIndex->lpDirEntry->next = NULL;
          FreeEntries(lpIndex->lpDirEntry);
          lpIndex->lpDirEntry = lpCurrent;
          bIncrement = FALSE;
        } else
        {
          /* Remove the current entry */
          lpPrev->next = lpCurrent->next;
          lpCurrent->next = NULL;
          FreeEntries(lpCurrent);
          lpCurrent = (LPDIRLIST_ENTRY) lpPrev->next;
          bIncrement = FALSE;
        }
      } else
      {
        /* See if the first name is to be set */
        if (lpPlan->dwFields & DISPLAY_FIRSTNAME) {
          if (lpCurrent->lpszFirstName)
            free(lpCurrent->lpszFirstName);
          lpCurrent->lpszFirstName = strdup(srcEntry.lpszFirstName);
        }

        /* See if the last name is to be set */
        if (lpPlan->dwFields & DISPLAY_LASTNAME) {
          if (lpCurrent->lpszLastName)
            free(lpCurrent->lpszLastName);
          lpCurrent->lpszLastName = strdup(srcEntry.lpszLastName);
        }

        /* See if the flags need to be set */
        if (lpPlan->dwFields & DISPLAY_FLAGS)
        {
          /* Set those that need to be set and remove those that should be removed */
          lpCurrent->dwFlags |= srcEntry.dwFlags;
          lpCurrent->dwFlags &= ~dwRemoveFlags;
        }

        /* Read the website if it's in the column list */
        if (lpPlan->dwFields & DISPLAY_WEBSITE) {
          if (lpCurrent->lpszWebSite)
            free(lpCurrent->lpszWebSite);
          lpCurrent->lpszWebSite = strdup(srcEntry.lpszWebSite);
        }

        /* Read the email value if it's in the column list */
        if (lpPlan->dwFields & DISPLAY_EMAIL) {
          if (lpCurrent->lpszEMail)
            free(lpCurrent->lpszEMail);
          lpCurrent->lpszEMail = strdup(srcEntry.lpszEMail);
        }

        /* Read the telephone number if it's in the column list */
        if (lpPlan->dwFields & DISPLAY_TELEPHONE) {
          if (lpCurrent->lpszTelephone)
            free(lpCurrent->lpszTelephone);
          lpCurrent->lpszTelephone = strdup(srcEntry.lpszTelephone);
        }

        /* Set the class entries if requested */
        if (lpPlan->dwExtFields)
        {
          DWORD i = 1;
          while ((i <= MAX_CLASS) && !bFormatError)
          {
            /* If this attribute is requested for import, then get the attribute */
            if (lpPlan->dwExtFields & i) {
              LPDIRLIST_CLASSDATA lpClassData = GetClassData(lpCurrent, i);
              LPDIRLIST_CLASSDATA lpSrcClassData = GetClassData(&srcEntry, i);
              DWORD dwOrigFlags = (DWORD) NULL;

              /* If the value isn't already defined, then create one */
              if (!lpClassData) {
                if (lpCurrent->lpClassData)
                  lpCurrent->lpClassData = (LPDIRLIST_CLASSDATA) realloc(lpCurrent->lpClassData, (++lpCurrent->nClassDataValues) * sizeof(DIRLIST_CLASSDATA));
                else
                  lpCurrent->lpClassData = (LPDIRLIST_CLASSDATA) malloc((++lpCurrent->nClassDataValues) * sizeof(DIRLIST_CLASSDATA));
                lpClassData = &(lpCurrent->lpClassData[lpCurrent->nClassDataValues - 1]);
                memset(lpClassData, 0, sizeof(DIRLIST_CLASSDATA));
                lpClassData->idClass = i;
              } else
                dwOrigFlags = lpClassData->dwFlags;

              /* Get the flags if requested */
              if (lpPlan->dwExtFlags) {
                switch (lpPlan->wExtFlagModifyType)
                {

                case FLAG_SET :
                  {
                    /* Add the flags that are set to 1 */
                    lpClassData->dwFlags |= lpSrcClassData->dwFlags;
                    break;
                  }

                case FLAG_REMOVE :
                  {
                    /* Remove the flags that are set to 1 */
                    lpClassData->dwFlags &= ~lpSrcClassData->dwFlags;
                    break;
                  }

                default :
                  {
                    /* Replace the entire flag set with the given flags */
                    lpClassData->dwFlags = lpSrcClassData->dwFlags;
                    break;
                  }
                }
              }

              /* If both values are given, then set them */
              if (lpPlan->dwExtBoth & i) {
                lpClassData->dwValue = lpSrcClassData->dwValue;

                /* Free any previously allocated memory in the class data */
                if (lpClassData->lpszValue && (dwOrigFlags & CLASSDATA_OTHER))
                  free(lpClassData->lpszValue);

                /* If this value has the "other" flag set, then set the string; */
                /* otherwise, ignore any string that exists in that column */
                if (lpClassData->dwFlags & CLASSDATA_OTHER)
                  lpClassData->lpszValue = strdup(lpSrcClassData->lpszValue);
                else
                  lpClassData->lpszValue = NULL;
              }

              /* If the value numbers are to be imported, then do so */
              if ((lpPlan->dwExtValues & i) && !(lpPlan->dwExtBoth & i)) {
                lpClassData->dwValue = lpSrcClassData->dwValue;
              }

              /* If the string representation is imported instead, then do so */
              if (!(lpPlan->dwExtValues & i) && !(lpPlan->dwExtBoth & i))
              {
                /* Free any previously allocated memory in the class data */
                if (lpClassData->lpszValue && (dwOrigFlags & CLASSDATA_OTHER))
                  free(lpClassData->lpszValue);

                /* If this value has the "other" flag set, then set the string; */
                /* otherwise, ignore any string that exists in that column */
                if (lpClassData->dwFlags & CLASSDATA_OTHER)
                  lpClassData->lpszValue = strdup(lpSrcClassData->lpszValue);
                else
                  lpClassData->lpszValue = NULL;
              }
            }

            /* Increment safely */
            if (i == MAX_CLASS)
              i++;
            else
              i *= 2;
          }
        }

        /* Tell the user what we did */
        infprintf("*Modified user \"%s %s\" (%s)\n",
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszFirstName, STR_NUL_TERM),
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszLastName, STR_NUL_TERM),
                  SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpCurrent->lpszUserName, STR_NUL_TERM));
        safeStringsLen = origSafeStringsLen;
      }
    }

    /* Next entry */
    if (bIncrement && lpCurrent) {
      lpPrev = lpCurrent;
      lpCurrent = (LPDIRLIST_ENTRY) lpCurrent->next;
    }
  }

  /* Everything was modified/added/deleted correctly :-)!  Rebuild the sort tables */
  if (bEmitStatus) {
    infprintf("#Refreshing sort tables...\n");
    fflush(stdout);
  }
  lpCurTable = lpIndex->lpSortTable;
  while (lpCurTable) {
    CreateSortTable(lpCurTable, lpIndex->lpDirEntry, lpIndex->lpClass);
    lpCurTable = (LPDIRLIST_SORTTABLE) lpCurTable->next;
  }

  /* We did it :-) */
  free(lpszLine);
  return TRUE;
}

BOOL LockIndexFile(LPDIRLIST_INDEX lpIndex, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_USER lpUser, BOOL bAdminMode, BOOL bEmitStatus)
{
  /* If no support for locking exists, then return an error */
  DWORD origSafeStringsLen = safeStringsLen;
  if (!(lpIndex->lpHeader->dwFlags & HEADER_ENFORCELOCKS)) {
    if (bEmitStatus)
      infprintf("#Lock enforcement has been disabled; enable it and this function will succeed\n");
    return FALSE;
  }

  /* If a locking record exists, then we cannot lock in this version :-( */
  if (lpIndex->lpLock) {
    if (bEmitStatus)
      infprintf("#File already locked (if the file is locked by one user, you will always get this message in this version)\n");
    return FALSE;
  } else
  {
    /* Read the input and get our parameters from that */
    LPTSTR lpszLine = (LPTSTR) malloc(32);
    DWORD dwLineLen = 0, dwLineAlloc = 32;

    /* Create a new locking record in the index */
    lpIndex->lpLock = (LPDIRLIST_LOCK) malloc(sizeof(DIRLIST_LOCK));
    memset(lpIndex->lpLock, 0, sizeof(DIRLIST_LOCK));

    /* Get the operations to lock */
    if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead)) {
      if (bEmitStatus)
        infprintf("#Could not read the operations field!\n");
      free(lpIndex->lpLock);
      lpIndex->lpLock = NULL;
      free(lpszLine);
      return FALSE;
    }
    lpIndex->lpLock->dwOperations = strtoul(lpszLine, NULL, 10);

    /* Get the lock duration */
    if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead)) {
      if (bEmitStatus)
        infprintf("#Could not read the lock duration field!\n");
      free(lpIndex->lpLock);
      lpIndex->lpLock = NULL;
      free(lpszLine);
      return FALSE;
    }
    lpIndex->lpLock->dwLockDuration = strtoul(lpszLine, NULL, 10);

    /* Set the other variables */
    lpIndex->lpLock->lpUser = lpUser;
    lpIndex->lpLock->bAdminMode = bAdminMode;
    lpIndex->lpLock->tLockTime = time(NULL);

    /* Give our status if it was requested */
    if (bEmitStatus) {
      infprintf("#User %s locked %u in the database\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpUser->lpszName, STR_NUL_TERM),
                lpIndex->lpLock->dwOperations);
      safeStringsLen = origSafeStringsLen;
    }

    /* We are done :-) */
    free(lpszLine);
    return TRUE;
  }
}

BOOL UnlockIndexFile(LPDIRLIST_INDEX lpIndex, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_USER lpUser, BOOL bEmitStatus)
{
  /* If a locking record for this user exists, then remove it */
  DWORD origSafeStringsLen = safeStringsLen;
  if (lpIndex->lpLock->lpUser->idUser == lpUser->idUser)
  {
    /* Get the operations to unlock */
    LPTSTR lpszLine = (LPTSTR) malloc(32);
    DWORD dwLineLen = 0, dwLineAlloc = 32;
    DWORD dwOperationsToUnlock;
    if (!ReadTextLine(pInput, &lpszLine, &dwLineLen, &dwLineAlloc, lpdwMaxRead)) {
      if (bEmitStatus)
        infprintf("#Could not read the operations field!\n");
      free(lpIndex->lpLock);
      lpIndex->lpLock = NULL;
      free(lpszLine);
      return FALSE;
    }

    /* Remove the locked operations */
    dwOperationsToUnlock = strtoul(lpszLine, NULL, 10);
    lpIndex->lpLock->dwOperations &= ~dwOperationsToUnlock;

    /* Write our status */
    if (bEmitStatus) {
      infprintf("#Unlocking %u from %s\n", dwOperationsToUnlock,
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpIndex->lpLock->lpUser->lpszName, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
    }

    /* If no locked operations exist, then totally remove the lock */
    if (!lpIndex->lpLock->dwOperations) {
      free(lpIndex->lpLock);
      lpIndex->lpLock = NULL;
    }
    return TRUE;
  } else
  {
    /* User tried to unlock when it was not locked by him/her */
    if (bEmitStatus) {
      infprintf("#User %s did not have index locked; unlock failed\n",
                SafeResponseString(&safeStrings, &safeStringsLen, &safeStringsAlloc, lpIndex->lpLock->lpUser->lpszName, STR_NUL_TERM));
      safeStringsLen = origSafeStringsLen;
    }
    return FALSE;
  }
}
