/*
Directory listing CGI program
By Steven Lawrance            
(c) Steven Lawrance, 2000, 2002, 2004
(c) Bryant College, 1998-2000 

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __HEADER__DIRLIST__DATABASE_H
#define __HEADER__DIRLIST__DATABASE_H

#include "dirlist.h"

/* Class flags */
#define CLASS_DEFAULTVALUE     1

/* Maximum class ID, which is simply left-most bit 32 */
#define MAX_CLASS              0x80000000

/* Maximum lock duration (10 minutes) */
#define LOCK_MAXDURATION       600

/* Search levels */
#define SEARCHLEVEL_NONE       0
#define SEARCHLEVEL_ALL        1
#define SEARCHLEVEL_SOME       2

/* Header flags */
#define HEADER_SEQUENCENUMBERS 1
#define HEADER_ENFORCELOCKS    2
#define HEADER_AUDITEMAIL      4
#define HEADER_ENCRYPTFILE     8
#define HEADER_CACHEPASSWD     16
#define HEADER_EMAILDOMAIN     32
#define HEADER_DEFAULTURL      64

/* Display flags/fields */
#define DISPLAY_USERNAME       1
#define DISPLAY_FIRSTNAME      2
#define DISPLAY_LASTNAME       4
#define DISPLAY_LASTNAMEFIRST 14
#define DISPLAY_SEPARATENAMES 16
#define DISPLAY_MIN           32
#define DISPLAY_FLAGS         32
#define DISPLAY_WEBSITE       64
#define DISPLAY_EMAIL        128
#define DISPLAY_TELEPHONE    256
#define DISPLAY_MAX          256

/* Default username */
#define DEFAULT_USERNAME "Person with a web site"

/* Flags for the index files */
#define FLAGS_EXCLUDE        2
#define FLAGS_NOHOMEDIR      4
#define FLAGS_NONAMECHECK    8
#define FLAGS_NOWEBSITE     16
#define FLAGS_NOWEBCHECK    32
#define FLAGS_NOEMAIL       64
#define FLAGS_NOTELEPHONE  128
#define FLAGS_CLASSDATA    256

/* Flag modify operations */
#define FLAG_REPLACE         0
#define FLAG_SET             1
#define FLAG_REMOVE          2
#define FLAG_IGNORE          3
#define FLAG_SETREMOVE       4
#define FLAG_UNKNOWN         0xFFFFFFFF

/* Class data flags */
#define CLASSDATA_DWORD        1
#define CLASSDATA_STRING       2
#define CLASSDATA_NOASCERTAIN  4
#define CLASSDATA_OTHER        8  /* set if lpszValue belongs to the structure instead of CLASS_VALUE */

/* Database operations */
#define OPERATION_WEBQUERY               0
#define OPERATION_GETENTRIES             1
#define OPERATION_MODIFYENTRIES          2
#define OPERATION_DELETEENTRIES          4
#define OPERATION_GETCLASSES             8
#define OPERATION_MODIFYCLASSES         16
#define OPERATION_DELETECLASSES         32
#define OPERATION_GETSORTTABLES         64
#define OPERATION_MODIFYSORTTABLES     128
#define OPERATION_DELETESORTTABLES     256
#define OPERATION_GETINDEXINFO         512
#define OPERATION_MODIFYINDEXINFO     1024
#define OPERATION_CREATEINDEX         2048
#define OPERATION_DELETEINDEX         4096
#define OPERATION_LOCK                8192
#define OPERATION_UNLOCK             16384
#define OPERATION_GETUSERS           32768
#define OPERATION_MODIFYUSERS        65536
#define OPERATION_DELETEUSERS       131072
#define OPERATION_NOOP              262144

/* Header structure */
typedef struct _DIRLIST_HEADER
{
  /* Minor version information (major version is returned by CheckIndexHeader) */
  DWORD    dwVersion;        /* Minor version that saved the file */
  DWORD    dwLowestVersion;  /* Lowest minor version that can read this file */

  /* General information */
  DWORD    dwFlags;
  LPTSTR   lpszWebDocsDir;
  LPTSTR   lpszBaseDir;
  LPTSTR   lpszEMailDomain;
  LPTSTR   lpszDefaultURL;

  /* Statistical header data */
  DWORD    nEntryTypes;   /* Number of entry types, excluding @0 (4 in this version) */
  DWORD    nType1Entries;
  DWORD    nType2Entries;
  DWORD    nType3Entries;
  DWORD    nType4Entries;
  DWORD    nType5Entries;
  DWORD    nType6Entries;
  DWORD    dwSequenceNumber;

  /* Locking information */
  DWORD    dwFilePosLock;

  /* Destination e-mail for auditing */
  LPTSTR   lpszAuditEMail;
} DIRLIST_HEADER, *LPDIRLIST_HEADER;

/* User information structure */
typedef struct DIRLIST_USER
{
  /* Standard information */
  DWORD           idUser;
  LPTSTR          lpszName;
  LPTSTR          lpszPassword;
  LPTSTR          lpszDescription;
  DWORD           dwFilePos;

  /* User level (lower user levels have change access to higher users) */
  DWORD           dwUserLevel;

  /* Permitted operations */
  DWORD           dwOperations;
  BOOL            bAdminMode;
  DWORD           dwSearchLevel;

  /* Auditing */
  BOOL            bAudit;
  DWORD           dwAuditOperations;
  LPBYTE          next;
} DIRLIST_USER, *LPDIRLIST_USER;

/* Database lock structure */
typedef struct _DIRLIST_LOCK
{
  /* User information */
  DWORD           dwFilePosUser;
  LPDIRLIST_USER  lpUser;

  /* Operations used by the lock */
  DWORD           dwOperations;

  /* Whether or not this lock is administrative */
  BOOL            bAdminMode;

  /* The lock time and duration */
  time_t          tLockTime;
  DWORD           dwLockDuration;
} DIRLIST_LOCK, *LPDIRLIST_LOCK;

/* Structures for classification definitions and values */
typedef struct _DIRLIST_CLASSVALUE {
  DWORD    idValue;
  LPTSTR   lpszValue;
} DIRLIST_CLASSVALUE, *LPDIRLIST_CLASSVALUE;

typedef struct _DIRLIST_CLASS {
  DWORD    idClass;
  LPTSTR   lpszName;
  DWORD    nValues;
  DWORD    dwFlags;
  DWORD    idDefaultValue;
  LPDIRLIST_CLASSVALUE lpValues;
  LPBYTE   next;
} DIRLIST_CLASS, *LPDIRLIST_CLASS;

/* Class data structure */
typedef struct _DIRLIST_CLASSDATA {
  DWORD    idClass;
  DWORD    dwFlags;
  DWORD    dwValue;
  LPTSTR   lpszValue;
} DIRLIST_CLASSDATA, *LPDIRLIST_CLASSDATA;

/* Structure for directory list entries */
typedef struct _DIRLIST_ENTRY
{
  /* Standard information */
  DWORD    dwFlags;
  LPTSTR   lpszUserName;
  LPTSTR   lpszFirstName;
  LPTSTR   lpszLastName;
  LPTSTR   lpszWebSite;
  LPTSTR   lpszEMail;
  LPTSTR   lpszTelephone;
  LPBYTE   next;
  DWORD    dwFilePos;       /* Used by sort tables, but not stored in file */
  
  /* Class data, if any */
  DWORD nClassDataValues;
  LPDIRLIST_CLASSDATA lpClassData;
} DIRLIST_ENTRY, *LPDIRLIST_ENTRY;

/* Sort table entry structure */
typedef struct _DIRLIST_SORTENTRY {
  LPDIRLIST_ENTRY lpDirEntry;
  DWORD    dwFilePos;
  LPBYTE   next;
} DIRLIST_SORTENTRY, *LPDIRLIST_SORTENTRY;

/* Sort table structure */
typedef struct _DIRLIST_SORTTABLE {
  DWORD    idSort;          /* Sort table's ID */
  DWORD    nEntries;
  LPTSTR   lpszOrder;
  DWORD    dwFilePos;       /* ReadSortEntries and ReadSortEntriesArray use this */
  LPDIRLIST_SORTENTRY lpSortEntries;
  LPBYTE   next;
} DIRLIST_SORTTABLE, *LPDIRLIST_SORTTABLE;

/* Access plan structure */
typedef struct _DIRLIST_ACCESSPLAN
{
  /* Value filters for display */
  DWORD    dwFields;
  DWORD    dwXFields;
  DWORD    dwExtFields;
  DWORD    dwXExtFields;
  DWORD    dwExtValues;
  DWORD    dwExtBoth;
  DWORD    dwExtFlags;

  /* Sort order */
  DWORD    idSort;

  /* User */
  LPDIRLIST_USER lpUser;

  /* Data access limits */
  BOOL     bAdminMode;
  
  /* Cell background color */
  BOOL     bUseBackgroundColor;
  DWORD    dwBackgroundColor;

  /* Search/Range parameters */
  DWORD    nSearches;
  DWORD    *dwSearchFields;
  DWORD    *dwExtSearchFields;
  DWORD    *dwExtSearchOrdinals;
  BOOL     *bSearchNot;
  BOOL     *bSearchAnd;
  DWORD    *dwSearchLevel;
  DWORD    *dwSearchLevelEnd;
  DWORD    *fSearchExact;
  LPTSTR   *lpszSearch;   /* lpszSearch and lpszRangeLow must always equal */
  LPTSTR   *lpszRangeLow;
  LPTSTR   *lpszRangeHigh;
  BOOL     *bRangeLowInclusive;
  BOOL     *bRangeHighInclusive;

  /* Binary search values */
  DWORD    dwBinarySearchField;
  DWORD    dwBinarySearchExtField;
  LPTSTR   lpszBinarySearch;

  /* Counter values */
  DWORD    dwCounterBegin;
  DWORD    dwCounterIncrement;
  DWORD    dwCounterPrev;
  DWORD    dwCounterNext;
  DWORD    bCounterHints;
  DWORD    dwCounterDisplayed;
  DWORD    dwCounterTotal;

  /* Sequence number (version) of last data */
  DWORD    dwSequenceNumber;

  /* Flag modification types */
  WORD     wFlagModifyType;
  WORD     wExtFlagModifyType;

  /* Auditing */
  BOOL     bAudit;
  LPTSTR   lpszAuditMsg;
  DWORD    dwAuditMsgLen;
} DIRLIST_ACCESSPLAN, *LPDIRLIST_ACCESSPLAN;

typedef struct DIRLIST_INDEX
{
  /* Store everything that relates to an index file */
  LPDIRLIST_HEADER    lpHeader;
  LPDIRLIST_SORTTABLE lpSortTable;
  LPDIRLIST_CLASS     lpClass;
  LPDIRLIST_ENTRY     lpDirEntry;
  LPDIRLIST_USER      lpUser;
  LPDIRLIST_LOCK      lpLock;

  /* Counter variables (optional; default behavior points to values in lpHeader) */
  LPDWORD lpnDirEntries;
  LPDWORD lpnClasses;
  LPDWORD lpnTables;
  LPDWORD lpnUsers;
  LPDWORD lpnLocks;
} DIRLIST_INDEX, *LPDIRLIST_INDEX;

typedef struct DIRLIST_PASSWD_ENTRY
{
  /* Binary tree representation of /etc/passwd */
  LPTSTR lpszUser;
  LPTSTR lpszName;
  struct DIRLIST_PASSWD_ENTRY *left;
  struct DIRLIST_PASSWD_ENTRY *right;
} DIRLIST_PASSWD_ENTRY, *LPDIRLIST_PASSWD_ENTRY;

/* Forward References */
LPDIRLIST_CLASSVALUE GetClassValue(LPDIRLIST_CLASS lpClass, DWORD idValue);
LPDIRLIST_CLASS GetClass(LPDIRLIST_CLASS lpClassList, DWORD idClass);
LPDIRLIST_CLASSDATA GetClassData(LPDIRLIST_ENTRY lpDirEntry, DWORD idClass);
int CompareEntries(LPDIRLIST_ENTRY lpEntry1, LPDIRLIST_ENTRY lpEntry2, LPDIRLIST_CLASS lpClassRoot, LPDWORD dwOrder, DWORD nLayers, BOOL *bExt, BOOL *bDesc);
BOOL AddSortEntry(LPDIRLIST_SORTENTRY *lpRoot, LPDIRLIST_SORTENTRY *lpPrev, LPDIRLIST_SORTENTRY *lpCurrent, LPDIRLIST_ENTRY lpEntry, LPDIRLIST_CLASS lpClass, LPDWORD dwOrder, DWORD nLayers, BOOL *bExt, BOOL *bDesc);
BOOL CreateSortTable(LPDIRLIST_SORTTABLE lpTable, LPDIRLIST_ENTRY lpRoot, LPDIRLIST_CLASS lpClass);
BOOL SearchString(LPCTSTR lpszValue, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwCurrent);
BOOL SearchOrdinal(DWORD dwValue, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwCurrent);
BOOL SearchEntry(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_CLASS lpClass, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwCurrent);
BOOL SearchFields(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_CLASS lpClass, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwLevel, LPDWORD lpdwCurrent);
BOOL SearchFieldsOld(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_CLASS lpClass, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwLevel, LPDWORD lpdwCurrent);
BOOL CheckRelevance(LPDIRLIST_ENTRY lpDirEntry, LPDIRLIST_CLASS lpClass, LPDIRLIST_ACCESSPLAN lpPlan);
BOOL FetchEntry(LPDIRLIST_ENTRY *lpRoot, LPDIRLIST_ENTRY *lpPrev, LPDIRLIST_ENTRY *lpCurrent, LPTSTR lpszUserName, BOOL bCreate);
DWORD LoadDirectory(LPCTSTR dir, LPDIRLIST_ENTRY *lpDirEntry, LPCTSTR lpszEMailDomain);
void FreeSortTables(LPDIRLIST_SORTTABLE lpTable);
void FreeEntry(LPDIRLIST_ENTRY lpDirEntry);
void FreeHeader(LPDIRLIST_HEADER lpHeader);
void FreeUsers(LPDIRLIST_USER lpUser);
void FreeAccessPlan(LPDIRLIST_ACCESSPLAN lpPlan);
void FreeEntries(LPDIRLIST_ENTRY lpDirEntry);
void FreeClasses(LPDIRLIST_CLASS lpClass);
void FreeIndex(LPDIRLIST_INDEX lpIndex);
LPDIRLIST_USER LoginUser(LPDIRLIST_USER *lpUser, LPDIRLIST_ACCESSPLAN lpPlan, DWORD dwOperations, BOOL bEmitStatus);
DWORD GetUpdateMode(LPTSTR lpszClient, LPTSTR *lpszValue, BOOL bEmitStatus);
BOOL GetSetRemoveFlag(LPTSTR lpszStr, BOOL bFirst, DWORD *dwFlag);
BOOL ModifyIndexInfo(LPDIRLIST_INDEX lpIndex, FILE *pInput, LPDWORD lpdwMaxRead, UCHAR cDelimiter, BOOL bEmitStatus);
LPDIRLIST_USER GetUserByName(LPDIRLIST_INDEX lpIndex, LPCTSTR lpszName);
BOOL ModifyUsers(LPDIRLIST_INDEX lpIndex, LPDIRLIST_USER lpActiveUser, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, UCHAR cDelimiter, BOOL bEmitStatus);
BOOL ModifySortTables(LPDIRLIST_INDEX lpIndex, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, UCHAR cDelimiter, BOOL bEmitStatus);
BOOL ModifyClasses(LPDIRLIST_INDEX lpIndex, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, DWORD idClass, UCHAR cDelimiter, BOOL bEmitStatus);
BOOL ModifyEntries(LPDIRLIST_INDEX lpIndex, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_ACCESSPLAN lpPlan, UCHAR cDelimiter, BOOL bEmitStatus);
BOOL ModifyEntriesToMatchOne(LPDIRLIST_INDEX lpIndex, BOOL bDelete, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_ACCESSPLAN lpPlan, UCHAR cDelimiter, BOOL bEmitStatus);
BOOL LockIndexFile(LPDIRLIST_INDEX lpIndex, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_USER lpUser, BOOL bAdminMode, BOOL bEmitStatus);
BOOL UnlockIndexFile(LPDIRLIST_INDEX lpIndex, FILE *pInput, LPDWORD lpdwMaxRead, LPDIRLIST_USER lpUser, BOOL bEmitStatus);

#endif /* DATABASE_H */
