/*
Directory listing CGI program
By Steven Lawrance            
(c) Steven Lawrance, 2000, 2002, 2004
(c) Bryant College, 1998-2000 

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __HEADER__DIRLIST__DIRLIST_H
#define __HEADER__DIRLIST__DIRLIST_H

#include "config.h"
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/stat.h>
#include <errno.h>
#include <pwd.h>

/* Basic constants */
#define TRUE      1
#define FALSE     0

/* Check for data type definitions and define if they don't exist */
#ifndef u_int16_t
#define u_int16_t unsigned short
#endif
#ifndef u_int32_t
#define u_int32_t unsigned int
#endif
#ifndef u_int8_t
#define u_int8_t unsigned char
#endif
#ifndef int8_t
#define int8_t char
#endif

/* Some data types */
#define WORD    u_int16_t
#define DWORD   u_int32_t
#define LPDWORD DWORD *
#define LPWORD  WORD *
#define BOOL    int8_t
#define UCHAR   unsigned char
#define BYTE    u_int8_t
#define LPBYTE  BYTE *
#define LPTSTR  char *
#define LPCTSTR const char *

/* Set this value to the directory delimiter of your OS */
#define cDirDelimiter '/'

/* The E-Mail command for sending mail (used for auditing) */
#define EMAIL_COMMAND "%s -fnobody@%s -F\"DirList2 Audit\" %s"

/* The default /home directory */
#define DEFAULT_HOME "/home"

/* Program return error codes */
#define ERROR_SUCCESS          0
#define ERROR_NOSUPPORT        1
#define ERROR_PROCESS          2
#define ERROR_INVALID          3

/* Define our native pointer integer type (is there a standardized way to do this?) */
#if SIZEOF_CHAR_P == 4
  /* 32-bit CPU */
  #define P_TYPE DWORD
#else
#if SIZEOF_CHAR_P == 8
  /* 64-bit CPU */
  #define P_TYPE u_int64_t
#else
  /* Unsupported CPU type */
  #error DirList2 will only work on 32-bit and 64-bit CPUs at this time
#endif
#endif

/* Replacement for strcasecmp if using Windows */
#ifndef HAVE_STRCASECMP
#define strcasecmp(a, b) strcmpi(a, b)
#endif

/* Replacement for vprintf if _doprnt exists (I'm unsure what platform(s) this applies to) */
#ifndef HAVE_VPRINTF
#ifdef HAVE_DOPRNT
#define vfprintf(stream,fmt,args) _doprnt(fmt,args,stream)
#endif
#endif

#endif /* DIRLIST_H */
