/*
Directory listing CGI program
By Steven Lawrance            
(c) Steven Lawrance, 2000, 2002, 2004
(c) Bryant College, 1998-2000 

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __HEADER__DIRLIST__IO_H
#define __HEADER__DIRLIST__IO_H

#include "dirlist.h"
#include "database.h"

/* Encryption methods (used as thumbprints on the packets) */
#define ENCRYPT_NONE            0
#ifdef DIRLIST_USE_PACKETMANGLER
#define ENCRYPT_PACKETMANGLE_A  0xCAFE /* 8-byte mangled packet (expanded size is 12 bytes) */
#define ENCRYPT_PACKETMANGLE_S  0xCAF2 /* short, variable-length mangled packet */
#endif
#define ENCRYPT_BLOWFISH        0xB103 /* not implemented yet */

/* Forward References */
size_t WriteFile(const void *buffer, size_t size, size_t count, FILE *stream);
size_t ReadFile(void *buffer, size_t size, size_t count, FILE *stream);
int SeekFile(FILE *stream, long offset, int origin);
BOOL ReadWORDNoEncrypt(FILE *index, WORD *value);
BOOL ReadWORD(FILE *index, WORD *value);
BOOL ReadDWORDNoEncrypt(FILE *index, DWORD *value);
BOOL ReadDWORD(FILE *index, DWORD *value);
LPDWORD ReadSortEntriesArray(FILE *index, LPDIRLIST_SORTTABLE lpTable);
BOOL CheckBadTemplate(FILE *pFile, int badTemplateNumber);
void WriteString(FILE *index, LPCTSTR str, DWORD len);
BOOL WriteIndexEntry(FILE *index, void *lpData, UCHAR dwType, BOOL bEmitStatus);
int ReadString(FILE *index, LPTSTR *str);
int ReadIndexEntry(FILE *index, LPBYTE *lpData, LPDIRLIST_CLASS lpClassList);
void WriteIndexHeader(FILE *index);
DWORD CheckIndexHeader(FILE *index);
BOOL CheckUserDirectory(LPTSTR lpszCheckFile, BOOL bVerifyHTML, LPCTSTR lpszFormatStr, ...);
BOOL ReserveSpace(FILE *index, DWORD nBytes);
BOOL LoadIndexFile(LPCTSTR lpszIndexName, LPDIRLIST_INDEX lpIndex, BOOL bManualCount, BOOL bEmitStatus);
BOOL SaveIndexFile(LPCTSTR lpszIndexName, LPDIRLIST_INDEX lpIndex, BOOL bEmitStatus, BOOL bVeryVerbose);
BOOL ProcessDirectory(LPCTSTR lpszDir, LPDIRLIST_INDEX lpIndex, BOOL bEmitStatus);
BOOL ReadTextLine(FILE *pInput, LPTSTR *lpszLine, LPDWORD lpdwLineLen, LPDWORD lpdwLineAlloc, LPDWORD lpdwMaxRead);

#endif /* IO_H */
