/*
Directory listing CGI program
By Steven Lawrance            
(c) Steven Lawrance, 2000, 2002, 2004
(c) Bryant College, 1998-2000 

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "dirlist.h"
#include "globals.h"
#include "util.h"
#include "database.h"
#include "io.h"
#include "web.h"

/* Main program entry point */
int main(int argc, char **argv)
{
  /* Get the CGI variables, if any */
  LPTSTR *lpCGIVars;
  lpCGIVars = getcgivars();

  /* Initialize the global variables */
  bAuditEnable = FALSE;
  lpszAuditBuffer = NULL;
  dwAuditBufferSize = 0;
  bEncryptEnable = FALSE;
  wEncryptMethod = ENCRYPT_NONE;
  lpPasswdCache = NULL;
  lpszPasswdCache = NULL;
  bConvertEndian = FALSE;
  safeStrings = NULL;
  safeStringsLen = 0;
  safeStringsAlloc = 0;

  /* Behave differently depending on the mode we're being run in */
  if (!lpCGIVars)
  {
    /* We are running from a command line -- use admin mode */
    DIRLIST_INDEX index;

    /* If we're creating a new index file, then do it :-) */
    if (IsFlagSet(argv, "-c") || IsFlagSet(argv, "--create") || IsFlagSet(argv, "-create")) {
      LPTSTR lpszIndexFile = GetCmdLineParam(argv, 0, "");
      BOOL bVerbose = IsFlagSet(argv, "-v");
      BOOL bVeryVerbose = IsFlagSet(argv, "-vv");
      BOOL bPwCache = IsFlagSet(argv, "--pw-cache");
      BOOL bEncrypt = IsFlagSet(argv, "--encrypt");
      BOOL bNoAdmin = IsFlagSet(argv, "--no-admin");
      BOOL bNoGuest = IsFlagSet(argv, "--no-guest");
      BOOL bNoDomain = IsFlagSet(argv, "--no-domain");
      BOOL bNoURL = IsFlagSet(argv, "--no-url");
      DWORD dwLineAlloc, dwLineLen;
      LPDIRLIST_USER lpGuest;
      
      /* Very-verbose implies verbose */
      if (bVeryVerbose)
        bVerbose = TRUE;

      /* Display our status if we're allowed to */
      if (bVerbose) {
        printf("Create Index:\nIndex:      \"%s\"\nEncryption: %s\nPwCache:    %s\n\n", lpszIndexFile, (bEncrypt)? "yes": "no", (bPwCache)? "yes": "no");
        fflush(stdout);
      }

      /* Create a header */
      dwLineAlloc = dwLineLen = 20;
      memset(&index, (char) NULL, sizeof(DIRLIST_INDEX));
      index.lpHeader = (LPDIRLIST_HEADER) malloc(sizeof(DIRLIST_HEADER));
      index.lpHeader->lpszWebDocsDir = (LPTSTR) strdup(DIRLIST_WEBDOCS);
      index.lpHeader->lpszBaseDir = (LPTSTR) malloc(1);
      index.lpHeader->lpszAuditEMail = (LPTSTR) malloc(1);
      index.lpHeader->lpszEMailDomain = (bNoDomain)? strdup(DIRLIST_EMAILDOMAIN): malloc(dwLineAlloc+1);
      index.lpHeader->lpszBaseDir[0] = 0;
      index.lpHeader->lpszAuditEMail[0] = 0;
      if (bNoURL) {
        index.lpHeader->lpszDefaultURL = (LPTSTR) malloc(1);
        index.lpHeader->lpszDefaultURL[0] = 0;
      } else
        index.lpHeader->lpszDefaultURL = malloc(dwLineAlloc+1);

      /* Get the e-mail domain */
      if (!bNoDomain)
      {
        fprintf(stderr, "Enter the e-mail domain to use (default=%s): ", DIRLIST_EMAILDOMAIN);
        if (!ReadTextLine(stdin, &(index.lpHeader->lpszEMailDomain), &dwLineLen, &dwLineAlloc, (LPDWORD) NULL)) {
          printf("\nCould not read e-mail domain!  Action aborted.\n");
          FreeIndex(&index);
          return ERROR_PROCESS;
        }
        if (!index.lpHeader->lpszEMailDomain[0]) {
          free(index.lpHeader->lpszEMailDomain);
          index.lpHeader->lpszEMailDomain = strdup(DIRLIST_EMAILDOMAIN);
        }
      }

      /* Get the default URL */
      if (!bNoURL)
      {
        dwLineAlloc = dwLineLen = 20;
        fprintf(stderr, "Enter the default URL to use (default=/search.html): ");
        if (!ReadTextLine(stdin, &(index.lpHeader->lpszDefaultURL), &dwLineLen, &dwLineAlloc, (LPDWORD) NULL)) {
          printf("\nCould not read default URL!  Action aborted.\n");
          FreeIndex(&index);
          return ERROR_PROCESS;
        }
        if (!index.lpHeader->lpszDefaultURL[0]) {
          free(index.lpHeader->lpszDefaultURL);
          index.lpHeader->lpszDefaultURL = strdup("/search.html");
        }
      }

      /* Set any command-line flags */      
      index.lpHeader->dwFlags = ((bEncrypt)? HEADER_ENCRYPTFILE: 0) | ((bPwCache)? HEADER_CACHEPASSWD: 0) | HEADER_EMAILDOMAIN | HEADER_DEFAULTURL;
      
      /* Create the Administrator */
      if (!bNoAdmin) {
        dwLineAlloc = dwLineLen = 10;
        index.lpUser = (LPDIRLIST_USER) malloc(sizeof(DIRLIST_USER));
        memset(index.lpUser, 0, sizeof(DIRLIST_USER));
        index.lpUser->idUser = 0;
        index.lpUser->lpszName = strdup("Administrator");
        index.lpUser->lpszPassword = malloc(dwLineAlloc+1);
        index.lpUser->lpszDescription = strdup("Administrator of the DirList2 database");
        index.lpUser->dwOperations = 0xFFFFFFFF;
        index.lpUser->bAdminMode = TRUE;
        index.lpUser->dwSearchLevel = SEARCHLEVEL_ALL;
        index.lpUser->next = (LPBYTE) NULL;
        
        /* Get the initial Administrator password */
        fprintf(stderr, "Enter initial Administrator password: ");
        if (!ReadTextLine(stdin, &(index.lpUser->lpszPassword), &dwLineLen, &dwLineAlloc, (LPDWORD) NULL)) {
          printf("\nCould not read password!  Action aborted.\n");
          FreeIndex(&index);
          return ERROR_PROCESS;
        }

        /* Allocate guest's memory */
        lpGuest = (LPDIRLIST_USER) (index.lpUser->next = (LPBYTE) malloc(sizeof(DIRLIST_USER)));
      } else
      {
        /* Allocate guest's memory */
        lpGuest = index.lpUser = (LPDIRLIST_USER) malloc(sizeof(DIRLIST_USER));
      }

      /* Create the Guest */
      memset(lpGuest, 0, sizeof(DIRLIST_USER));
      lpGuest->idUser = 1;
      lpGuest->lpszName = strdup("Guest");
      lpGuest->lpszPassword = strdup("");
      lpGuest->lpszDescription = strdup("Guest/Web User");
      lpGuest->dwOperations = OPERATION_WEBQUERY;
      lpGuest->bAdminMode = FALSE;
      lpGuest->next = (LPBYTE) NULL;

      /* Get the initial Guest search level, if we're supposed to */
      if (!bNoGuest) {
        do {
          fprintf(stderr, "Select the initial Guest search level number:\n0  None - No web search rights\n1  All  - Full web search rights\n2  Some - Can search if a text field has at least two characters specified\n:");
          scanf("%u", &(lpGuest->dwSearchLevel));
        } while (lpGuest->dwSearchLevel > SEARCHLEVEL_SOME); /* currently the highest */
      } else
        lpGuest->dwSearchLevel = SEARCHLEVEL_NONE;

      /* Create the index file */
      if (!SaveIndexFile(lpszIndexFile, &index, bVerbose, bVeryVerbose)) {
        printf("Error creating index file\n");
        FreeIndex(&index);
        return ERROR_PROCESS;
      } else {
        printf("Index file created.\n");
        FreeIndex(&index);
        return ERROR_SUCCESS;
      }
    }

    /* If this is an update, then perform the directory update :-) */
    if (IsFlagSet(argv, "-u") || IsFlagSet(argv, "--update") || IsFlagSet(argv, "-update")) {
      LPTSTR lpszIndexFile = GetCmdLineParam(argv, 0, "");
      LPTSTR lpszIndexDirectory = GetCmdLineParam(argv, 1, NULL);
      BOOL bVerbose = IsFlagSet(argv, "-v");
      BOOL bVeryVerbose = IsFlagSet(argv, "-vv");

      /* Very-verbose implies verbose */
      if (bVeryVerbose)
        bVerbose = TRUE;

      /* Load the index file */
      memset(&index, (char) NULL, sizeof(DIRLIST_INDEX));
      if (!LoadIndexFile(lpszIndexFile, &index, FALSE, bVerbose)) {
        printf("Could not load index file!\n");
        return ERROR_PROCESS;
      }

      /* Determine what the source directory should be */
      if (!lpszIndexDirectory) {
        if (index.lpHeader->lpszBaseDir && index.lpHeader->lpszBaseDir[0])
          lpszIndexDirectory = index.lpHeader->lpszBaseDir;
        else
          lpszIndexDirectory = DEFAULT_HOME;
      }

      /* Display our status if we're allowed to */
      if (bVerbose) {
        printf("Update Index:\nIndex:     \"%s\"\nDirectory: \"%s\"\n\n", lpszIndexFile, lpszIndexDirectory);
        fflush(stdout);
      }

      /* Create a /etc/passwd cache if we're supposed to */
      if (bVerbose)
        printf("Initializing /etc/passwd cache...");
      if (index.lpHeader->dwFlags & HEADER_CACHEPASSWD) {
        CreatePasswdCache();
        if (bVerbose) {
          if (lpPasswdCache)
            printf(" (done)\n\n");
          else
            printf(" (failed; using slower getpwnam() instead)\n\n");
        }
      } else {
        if (bVerbose)
          printf(" (disabled; using slower getpwnam() instead)\n\n");
      }

      /* Update the directories and return the success/failure */
      if (ProcessDirectory(lpszIndexDirectory, &index, bVerbose) >= 0)
      {
        /* Save the index file */
        if (!SaveIndexFile(lpszIndexFile, &index, bVerbose, bVeryVerbose)) {
          printf("Error saving index file.%s\n", (bVerbose)? "": "  Try running again with -v or -vv for more details.");
          FreeIndex(&index);
          FreePasswdCache();
          return ERROR_PROCESS;
        } else {
          printf("Index is now up to date.\n");
          FreeIndex(&index);
          FreePasswdCache();
          return ERROR_SUCCESS;
        }
      } else {
        printf("Error!  %s\n", (bVerbose)? "Check your parameters and try again.": "Try running again with -v or -vv for more details.");
        FreeIndex(&index);
        FreePasswdCache();
        return ERROR_PROCESS;
      }
    }

    /* Command line not handled!  Display parameters */
    printf("Directory List v%i.%i Command Line Interface\n\nUsage:\n", DIRLIST_VERSION_MAJOR, DIRLIST_VERSION_MINOR);
    printf("%s {-u | --update | -update} [-v|-vv] IndexFile [HomeBaseDirectory]\n", argv[0]);
    printf("%s {-c | --create | -create} [-v|-vv] [other_options] IndexFile\n", argv[0]);
    printf("-v          Verbose mode -- display progress and other information\n");
    printf("-vv         Very-verbose -- display very detailed progress information\n\n");
    printf("Special options available during Create Mode:\n");
    printf("--encrypt    Enable encryption on the index file initially (default is no)\n");
    printf("--pw-cache   Enable caching of /etc/passwd (default is no cache)\n");
    printf("--no-admin   Don't prompt for an Administrator password (default is to ask)\n");
    printf("--no-guest   Force guest's search level to \"none\" (default is to ask level)\n");
    printf("--no-domain  Use %s as the e-mail domain (default is to ask)\n", DIRLIST_EMAILDOMAIN);
    printf("--no-url     Don't set a default URL (default is to ask)\n\n");
    printf("For web access, place your .idx2 index files in the dirlist_indexes\n");
    printf("subdirectory of the directory that %s is in.\n\n", argv[0]);
    printf("Similarly, bad templates are stored in the dirlist_html subdirectory.\n\n");
	printf("Please feel free to send any suggestions, comments, questions,\n");
	printf("or bug reports to %s.\n\n", PACKAGE_BUGREPORT);
    return ERROR_NOSUPPORT;
  } else
  {
    /* We're being accessed via CGI -- process the request */
    DIRLIST_INDEX index;
    LPTSTR lpszIndexName = GetQuery(lpCGIVars, "i", "");
    DWORD dwIndexNameLen;
    LPTSTR lpszIndexFile;
    DWORD dwOperation = strtoul(GetQuery(lpCGIVars, "op", "0"), NULL, 10);
    UCHAR cDelimiter = (dwOperation)? GetQuery(lpCGIVars, "d", "\t")[0]: (UCHAR) NULL;
    UCHAR cStringChar = (dwOperation)? GetQuery(lpCGIVars, "sc", "")[0]: (UCHAR) NULL;
    LPTSTR lpszHTMLName = GetQuery(lpCGIVars, "h", "list");
    BOOL bJavaScript = (strtoul(GetQuery(lpCGIVars, "js", "0"), NULL, 10))? TRUE: FALSE;
    DIRLIST_ACCESSPLAN plan;
    LPDIRLIST_USER lpUser = NULL;
    DWORD dwMaxRead = strtoul((getenv("CONTENT_LENGTH"))? getenv("CONTENT_LENGTH"): "0", NULL, 10);
    BOOL bGoDefault = FALSE;
    BOOL bContentTypeWritten = FALSE;
    BOOL bErrorCondition = FALSE;
    LPTSTR lpszError = NULL;

    /* If we don't have an index name, then redirect to the default request */
    if (strcmp(lpszIndexName, "") == 0) {
      lpszIndexName = "users";
      bGoDefault = TRUE;
    }

    /* Ensure that all characters in the index name are alphanumeric */
    dwIndexNameLen = SafeFileName(lpszIndexName);

    /* Allocate the index file name string and create it */
    lpszIndexFile = malloc(dwIndexNameLen + 22);
    sprintf(lpszIndexFile, "dirlist_indexes/%s.idx2", lpszIndexName);

    /* Ensure that all characters in the HTML output template name are alphanumeric */
    SafeFileName(lpszHTMLName);

    /* Write the content-type if this isn't a web query */
    if (dwOperation & ~OPERATION_WEBQUERY) {
      printf("Content-type: ");
      bContentTypeWritten = TRUE;
      switch (dwOperation)
      {
      case OPERATION_GETENTRIES :
        printf("dirlist2/getentries"); break;

      case OPERATION_MODIFYENTRIES :
        printf("dirlist2/modifyentries"); break;

      case OPERATION_DELETEENTRIES :
        printf("dirlist2/deleteentries"); break;

      case OPERATION_GETCLASSES :
        printf("dirlist2/getclasses"); break;

      case OPERATION_MODIFYCLASSES :
        printf("dirlist2/modifyclasses"); break;

      case OPERATION_DELETECLASSES :
        printf("dirlist2/deleteclasses"); break;

      case OPERATION_GETSORTTABLES :
        printf("dirlist2/getsorttables"); break;

      case OPERATION_MODIFYSORTTABLES :
        printf("dirlist2/modifysorttables"); break;

      case OPERATION_DELETESORTTABLES :
        printf("dirlist2/deletesorttables"); break;

      case OPERATION_GETINDEXINFO :
        printf("dirlist2/getindexinfo"); break;

      case OPERATION_MODIFYINDEXINFO :
        printf("dirlist2/modifyindexinfo"); break;

      case OPERATION_CREATEINDEX :
        printf("dirlist2/createindex"); break;

      case OPERATION_DELETEINDEX :
        printf("dirlist2/deleteindex"); break;

      case OPERATION_LOCK :
        printf("dirlist2/lock"); break;

      case OPERATION_UNLOCK :
        printf("dirlist2/unlock"); break;

      case OPERATION_GETUSERS :
        printf("dirlist2/getusers"); break;

      case OPERATION_MODIFYUSERS :
        printf("dirlist2/modifyusers"); break;

      case OPERATION_DELETEUSERS :
        printf("dirlist2/deleteusers"); break;

      case OPERATION_NOOP :
        printf("dirlist2/noop"); break;

      default :
        printf("dirlist2/error\n\n");
        printf("-Unknown operation\n");
        FreeCGIVars(lpCGIVars);
        return 0;
      }
      printf("\n\n");
      fflush(stdout);
    }

    /* Get the access plan */
    if (!GetAccessPlan(lpCGIVars, stdin, &dwMaxRead, cDelimiter, &plan, TRUE, TRUE, TRUE, &lpszError)) {
      if (!bContentTypeWritten)
        printf("Content-type: text/html\n\n");
      if (lpszError != NULL) {
        printf(lpszError);
        free(lpszError);
        lpszError = NULL;
      }
      printf("-Error, could not get access plan\n");
      FreeCGIVars(lpCGIVars);
      return 0;
    }

    /* Load the index file into memory unless if we are doing OPERATION_WEBQUERY */
    memset(&index, (char) NULL, sizeof(DIRLIST_INDEX));
    if ((dwOperation & ~OPERATION_WEBQUERY) || bGoDefault) {
      if (!LoadIndexFile(lpszIndexFile, &index, TRUE, (!bGoDefault)? TRUE: FALSE))
      {
        /* We cannot continue without the file :-( */
        if (!bContentTypeWritten)
          printf("Content-type: text/html\n\n");
        if (!bGoDefault)
          printf("-Error\n");
        else
          printf("Index \"users\" cannot be loaded to determine default URL; make sure it exists and is readable.");
        free(lpszIndexFile);
        FreeCGIVars(lpCGIVars);
        return 0;
      }

      /* If we are going to the default location, then go now */
      if (bGoDefault) {
        if (index.lpHeader && (index.lpHeader->dwFlags & HEADER_DEFAULTURL) && index.lpHeader->lpszDefaultURL && index.lpHeader->lpszDefaultURL[0])
          printf("Location: %s\n\n", index.lpHeader->lpszDefaultURL);
        else {
          if (!bContentTypeWritten)
            printf("Content-type: text/html\n\n");
          printf("Default URL not defined in \"users\" index; please correct.");
        }
        if (lpszAuditBuffer)
          free(lpszAuditBuffer);
        FreeAccessPlan(&plan);
        FreeIndex(&index);
        free(lpszIndexFile);
        FreeCGIVars(lpCGIVars);
        return 0;
      }

      /* Print this program's version number */
      printf("VDirList%u.%u\n", DIRLIST_VERSION_MAJOR, DIRLIST_VERSION_MINOR);

      /* Print the file sequence number */
      printf("Q%u\n", index.lpHeader->dwSequenceNumber);

      /* Turn logging on to catch authentication errors if global logging is on */
      if (index.lpHeader->dwFlags & HEADER_AUDITEMAIL)
        bAuditEnable = TRUE;

      /* Authenticate the user */
      if (!(lpUser = LoginUser(&(index.lpUser), &plan, dwOperation, TRUE))) {
        infprintf("-Error\n");
        bErrorCondition = TRUE;
        lpUser = plan.lpUser;
        if (lpUser->lpszDescription)
          free(lpUser->lpszDescription);
        lpUser->lpszDescription = strdup("UNAUTHENTICATED USER - LOGIN FAILURE");

        /* Generate audit messages for everything except NO-OP when no password exists */
        if (dwOperation != OPERATION_NOOP || plan.lpUser->lpszPassword[0])
          lpUser->dwAuditOperations = 0xFFFFFFFF;
      }

      /* If this operation requires auditing, set the appropriate flag */
      if (dwOperation & lpUser->dwAuditOperations)
        bAuditEnable = TRUE;
      else
        bAuditEnable = FALSE;
    } else
      lpUser = NULL;

    /* Switch between the types of operations that we support */
    if (!bErrorCondition) {
      switch (dwOperation)
      {

      case OPERATION_WEBQUERY :
        {
          /* Display the list, sandwitched between the header and footer HTML files */
          printf("Content-type: text/html\n\n");
          if (!cDelimiter)
            WriteHTMLHeader(lpszIndexName, lpszHTMLName);
          if (!ListDirectory(lpszIndexFile, &plan, cDelimiter, cStringChar, bJavaScript, &(index.lpHeader)) && cDelimiter)
            infprintf("-Error while listing the directory due to the above error message(s)\n");
          lpUser = plan.lpUser;
          if (!cDelimiter)
            WriteHTMLFooter(lpszIndexName, lpszHTMLName);
          break;
        }

      case OPERATION_GETENTRIES :
        {
          /* Query Mode; Ensure this is a POST */
          if ((CompareStrings(getenv("REQUEST_METHOD"), "POST") == 0) && (CompareStrings(getenv("CONTENT_TYPE"), "dirlist2/operation") == 0))
          {
            LPTSTR lpszLine = (LPTSTR) malloc(32);
            DWORD dwLineLen = 0, dwLineAlloc = 32;
            infprintf("#Processing:\n");
            fflush(stdout);
            if (!ReadTextLine(stdin, &lpszLine, &dwLineLen, &dwLineAlloc, &dwMaxRead)) {
              infprintf("#Could not read sort order string\n");
              infprintf("-Error, %u bytes unread\n", dwMaxRead);
            } else
            {
              /* Build a sort table based on the input sorting type */
              LPDIRLIST_SORTTABLE lpQuerySort = (LPDIRLIST_SORTTABLE) malloc(sizeof(DIRLIST_SORTTABLE));
              memset(lpQuerySort, 0, sizeof(DIRLIST_SORTTABLE));
              lpQuerySort->lpszOrder = lpszLine;
              if (!CreateSortTable(lpQuerySort, index.lpDirEntry, index.lpClass))
                infprintf("-Error creating sort table\n");
              else
              {
                /* Go through each entry in the sort table and display those that match */
                LPDIRLIST_SORTENTRY lpCurrent = lpQuerySort->lpSortEntries;
                while (lpCurrent)
                {
                  /* If this one is relevant, then display it :-) */
                  if (CheckRelevance(lpCurrent->lpDirEntry, index.lpClass, &plan)) {
                    printf(">");
                    DisplayPerson(lpCurrent->lpDirEntry, &plan, index.lpClass, cDelimiter, cStringChar);
                  }

                  /* Next entry */
                  lpCurrent = (LPDIRLIST_SORTENTRY) lpCurrent->next;
                }           
              }

              /* Free the sort table */
              lpQuerySort->lpszOrder = NULL;
              FreeSortTables(lpQuerySort);

              /* We're done :-) */
              infprintf("+Done\n");
            }

            /* Free the line */
            free(lpszLine);
          } else
            infprintf("-Request method must be POST and Content-type must be dirlist2/operation\n");
          break;
        }

      case OPERATION_MODIFYENTRIES :
      case OPERATION_DELETEENTRIES :
        {
          /* Entry Modification Mode; Ensure this is a POST */
          if ((CompareStrings(getenv("REQUEST_METHOD"), "POST") == 0) && (CompareStrings(getenv("CONTENT_TYPE"), "dirlist2/operation") == 0))
          {
            infprintf("#Processing:\n");
            fflush(stdout);
            if ((plan.nSearches)? !ModifyEntriesToMatchOne(&index, (dwOperation == OPERATION_MODIFYENTRIES)? FALSE: TRUE, stdin, &dwMaxRead, &plan, cDelimiter, TRUE): !ModifyEntries(&index, (dwOperation == OPERATION_MODIFYENTRIES)? FALSE: TRUE, stdin, &dwMaxRead, &plan, cDelimiter, TRUE))
              infprintf("-Error, %u bytes unread\n", dwMaxRead);
            else
            {
              /* Save the index file */
              if (!SaveIndexFile(lpszIndexFile, &index, TRUE, FALSE))
                infprintf("-Error saving index file\n");
              else
                infprintf("+Done\n");
            }
          } else
            infprintf("-Request method must be POST and Content-type must be dirlist2/operation\n");
          break;
        }

      case OPERATION_GETCLASSES :
        {
          /* Get the class information */
          LPDIRLIST_CLASS lpClass = index.lpClass;
          LPDIRLIST_CLASS lpCurrent;
          DWORD i;

          /* Walk the linked list, writing each value */
          infprintf("#Processing:\n");
          lpCurrent = lpClass;
          while (lpCurrent)
          {
            /* Write the class information */
            printf(">%u\n|", lpCurrent->idClass);
            if (cStringChar)
              printf("%c", cStringChar);
            printf("%s", lpCurrent->lpszName);
            if (cStringChar)
              printf("%c", cStringChar);
            printf("\n|%u\n", lpCurrent->dwFlags);

            /* If the class flags include CLASS_DEFAULTVALUE, then write the default value */
            if (lpCurrent->dwFlags & CLASS_DEFAULTVALUE)
              printf("|%u\n", lpCurrent->idDefaultValue);

            /* Write the number of values */
            printf("|%u\n", lpCurrent->nValues);

            /* Write the values */
            for (i = 0; i < lpCurrent->nValues; i++) {
              printf("|%u", lpCurrent->lpValues[i].idValue);
              if (cDelimiter)
                printf("%c", cDelimiter);
              if (cStringChar)
                printf("%c", cStringChar);
              printf("%s", lpCurrent->lpValues[i].lpszValue);
              if (cStringChar)
                printf("%c", cStringChar);
              printf("\n");
            }

            /* Next class */
            lpClass = lpCurrent;
            lpCurrent = (LPDIRLIST_CLASS) lpCurrent->next;
            lpClass->next = NULL;
            FreeClasses(lpClass);
          }
          index.lpClass = NULL;

          /* We're done :-) */
          infprintf("+Done\n");
          break;
        }

      case OPERATION_MODIFYCLASSES :
      case OPERATION_DELETECLASSES :
        {
          /* Class Modification Mode; Ensure this is a POST */
          if ((CompareStrings(getenv("REQUEST_METHOD"), "POST") == 0) && (CompareStrings(getenv("CONTENT_TYPE"), "dirlist2/operation") == 0))
          {
            infprintf("#Processing:\n");
            fflush(stdout);
            if (!ModifyClasses(&index, (dwOperation == OPERATION_MODIFYCLASSES)? FALSE: TRUE, stdin, &dwMaxRead, plan.dwExtFields, cDelimiter, TRUE))
              infprintf("-Error, %u bytes unread\n", dwMaxRead);
            else
            {
              /* Save the index file */
              if (!SaveIndexFile(lpszIndexFile, &index, TRUE, FALSE))
                infprintf("-Error saving index file\n");
              else
                infprintf("+Done\n");
            }
          } else
            infprintf("-Request method must be POST and Content-type must be dirlist2/operation\n");
          break;
        }

      case OPERATION_GETSORTTABLES :
        {
          /* Get the sort table information */
          LPDIRLIST_SORTTABLE lpTable = index.lpSortTable;
          LPDIRLIST_SORTTABLE lpCurrent;

          /* Walk the linked list, writing each value */
          infprintf("#Processing:\n");
          lpCurrent = lpTable;
          while (lpCurrent)
          {
            /* Write the sort table information */
            printf(">%u\n|", lpCurrent->idSort);
            if (cStringChar)
              printf("%c", cStringChar);
            printf("%s", lpCurrent->lpszOrder);
            if (cStringChar)
              printf("%c", cStringChar);
            printf("\n|%u\n", lpCurrent->nEntries);

            /* Next sort table */
            lpTable = lpCurrent;
            lpCurrent = (LPDIRLIST_SORTTABLE) lpCurrent->next;
            lpTable->next = NULL;
            FreeSortTables(lpTable);
          }
          index.lpSortTable = NULL;

          /* We're done :-)! */
          infprintf("+Done\n");
          break;
        }

      case OPERATION_MODIFYSORTTABLES :
      case OPERATION_DELETESORTTABLES :
        {
          /* Sort Table Modification Mode; Ensure this is a POST */
          if ((CompareStrings(getenv("REQUEST_METHOD"), "POST") == 0) && (CompareStrings(getenv("CONTENT_TYPE"), "dirlist2/operation") == 0))
          {
            infprintf("#Processing:\n");
            fflush(stdout);
            if (!ModifySortTables(&index, (dwOperation == OPERATION_MODIFYSORTTABLES)? FALSE: TRUE, stdin, &dwMaxRead, cDelimiter, TRUE))
              infprintf("-Error, %u bytes unread\n", dwMaxRead);
            else
            {
              /* Save the index file */
              if (!SaveIndexFile(lpszIndexFile, &index, TRUE, FALSE))
                infprintf("-Error saving index file\n");
              else
                infprintf("+Done\n");
            }
          } else
            infprintf("-Request method must be POST and Content-type must be dirlist2/operation\n");
          break;
        }

      case OPERATION_GETINDEXINFO :
        {
          /* Get the index information from the file and display it */
          LPDIRLIST_HEADER lpHeader = index.lpHeader;
          infprintf("#Processing\n");
          
          /* Write the information in the header */
          printf(">%s\n", lpszIndexFile);
          printf("|%u\t%u\n", lpHeader->dwVersion, lpHeader->dwLowestVersion);
          printf("|%u\n|%s\n|%s\n", lpHeader->dwFlags, (lpHeader->lpszWebDocsDir)? lpHeader->lpszWebDocsDir: "", (lpHeader->lpszBaseDir)? lpHeader->lpszBaseDir: "");
          printf("|%u\n", lpHeader->nEntryTypes);
          
          /* Write the entry type statistics */
          printf("|%u\n|%u\n|%u\n|%u\n", lpHeader->nType1Entries, lpHeader->nType2Entries, lpHeader->nType3Entries, lpHeader->nType4Entries);
          lpHeader->nEntryTypes -= 4;
          while (lpHeader->nEntryTypes > 0) {
            printf("|0\n");
            (lpHeader->nEntryTypes)--;
          }
          
          /* Write the sequence number if the flags permit */
          if (lpHeader->dwFlags & HEADER_SEQUENCENUMBERS)
            printf("|%u\n", lpHeader->dwSequenceNumber);
          
          /* Write the file position of the lock information (though this cannot help the client) */
          if (lpHeader->dwFlags & HEADER_ENFORCELOCKS)
            printf("|%u\n", lpHeader->dwFilePosLock);

          /* Write the audit e-mail address and the audit operations if set */
          if (lpHeader->dwFlags & HEADER_AUDITEMAIL)
            printf("|%s\n", lpHeader->lpszAuditEMail);

          /* Write the e-mail domain if set */
          if (lpHeader->dwFlags & HEADER_EMAILDOMAIN)
            printf("|%s\n", lpHeader->lpszEMailDomain);

          /* Write the default URL if set */
          if (lpHeader->dwFlags & HEADER_DEFAULTURL)
            printf("|%s\n", lpHeader->lpszDefaultURL);
          
          /* We're done :-) */
          infprintf("+Done\n");
          break;
        }

      case OPERATION_MODIFYINDEXINFO:
        {
          /* Index Information Modification Mode; Ensure this is a POST */
          if ((CompareStrings(getenv("REQUEST_METHOD"), "POST") == 0) && (CompareStrings(getenv("CONTENT_TYPE"), "dirlist2/operation") == 0))
          {
            infprintf("#Processing:\n");
            fflush(stdout);
            if (!ModifyIndexInfo(&index, stdin, &dwMaxRead, cDelimiter, TRUE))
              infprintf("-Error, %u bytes unread\n", dwMaxRead);
            else
            {
              /* Save the index file */
              if (!SaveIndexFile(lpszIndexFile, &index, TRUE, FALSE))
                infprintf("-Error saving index file\n");
              else
                infprintf("+Done\n");
            }
          } else
            infprintf("-Request method must be POST and Content-type must be dirlist2/operation\n");
          break;
        }

      case OPERATION_CREATEINDEX :
      case OPERATION_DELETEINDEX :
        {
          infprintf("-Error, not implemented\n");
          break;
        }

      case OPERATION_LOCK :
        {
          /* Lock the index file */
          if ((CompareStrings(getenv("REQUEST_METHOD"), "POST") == 0) && (CompareStrings(getenv("CONTENT_TYPE"), "dirlist2/operation") == 0))
          {
            infprintf("#Processing:\n");
            if (!LockIndexFile(&index, stdin, &dwMaxRead, lpUser, plan.bAdminMode, TRUE))
              infprintf("-Error\n");
            else
            {
              /* Save the index file */
              if (!SaveIndexFile(lpszIndexFile, &index, TRUE, FALSE))
                infprintf("-Error saving index file\n");
              else
                infprintf("+Done\n");
            }
          } else
            infprintf("-Request method must be POST and Content-type must be dirlist2/operation\n");
          break;
        }

      case OPERATION_UNLOCK :
        {
          /* Unlock the index file */
          if ((CompareStrings(getenv("REQUEST_METHOD"), "POST") == 0) && (CompareStrings(getenv("CONTENT_TYPE"), "dirlist2/operation") == 0))
          {
            infprintf("#Processing:\n");
            if (!UnlockIndexFile(&index, stdin, &dwMaxRead, lpUser, TRUE))
              infprintf("-Error\n");
            else
            {
              /* Save the index file */
              if (!SaveIndexFile(lpszIndexFile, &index, TRUE, FALSE))
                infprintf("-Error saving index file\n");
              else
                infprintf("+Done\n");
            }
          } else
            infprintf("-Request method must be POST and Content-type must be dirlist2/operation\n");
          break;
        }

      case OPERATION_GETUSERS :
        {
          /* Get the user information */
          LPDIRLIST_USER lpCurrent = index.lpUser;

          /* Walk the linked list, writing each value */
          infprintf("#Processing:\n");
          while (lpCurrent)
          {
            /* Write the user information */
            printf(">%u\n|", lpCurrent->idUser);
            if (cStringChar)
              printf("%c", cStringChar);
            printf("%s", lpCurrent->lpszName);
            if (cStringChar)
              printf("%c", cStringChar);
            printf("\n|");
            if (cStringChar)
              printf("%c", cStringChar);
            printf("%s", lpCurrent->lpszDescription);
            if (cStringChar)
              printf("%c", cStringChar);
            printf("\n|%u\n", lpCurrent->dwUserLevel);
            printf("|%u\n", lpCurrent->dwOperations);
            printf("|%u\n", lpCurrent->bAdminMode);
            printf("|%u\n", lpCurrent->dwSearchLevel);
            printf("|%u\n", lpCurrent->bAudit);
            printf("|%u\n", lpCurrent->dwAuditOperations);

            /* Next user */
            lpCurrent = (LPDIRLIST_USER) lpCurrent->next;
          }

          /* We're done :-) */
          infprintf("+Done\n");
          break;
        }

      case OPERATION_MODIFYUSERS :
      case OPERATION_DELETEUSERS :
        {
          /* User Modification Mode; Ensure this is a POST */
          if ((CompareStrings(getenv("REQUEST_METHOD"), "POST") == 0) && (CompareStrings(getenv("CONTENT_TYPE"), "dirlist2/operation") == 0))
          {
            infprintf("#Processing:\n");
            fflush(stdout);
            if (!ModifyUsers(&index, lpUser, (dwOperation == OPERATION_MODIFYUSERS)? FALSE: TRUE, stdin, &dwMaxRead, cDelimiter, TRUE))
              infprintf("-Error, %u bytes unread\n", dwMaxRead);
            else
            {
              /* Save the index file */
              if (!SaveIndexFile(lpszIndexFile, &index, TRUE, FALSE))
                infprintf("-Error saving index file\n");
              else
                infprintf("+Done\n");
            }
          } else
            infprintf("-Request method must be POST and Content-type must be dirlist2/operation\n");
          break;
        }

      case OPERATION_NOOP :
        {
          /* No operation */
          infprintf("+Done\n");
          break;
        }
      }
    }

    /* If auditing was enabled, then send the e-mail message now */
#ifdef HAVE_SENDMAIL
    if (bAuditEnable && (index.lpHeader->dwFlags & HEADER_AUDITEMAIL))
    {
      /* Open a pipe to the e-mail program */
      FILE *hEMail;
#ifdef HAVE_SNPRINTF
      LPTSTR lpszCommand = malloc(8192);
      int commandLength = snprintf(lpszCommand, 8192, EMAIL_COMMAND, HAVE_SENDMAIL, index.lpHeader->lpszEMailDomain, index.lpHeader->lpszAuditEMail);
      if (commandLength < 0 || commandLength > 8191) {
        printf("-Could not create command line for auditing\n");
        lpszCommand[0] = 0;
      }
#else
#warning Using sprintf() instead of snprintf() in main(); sprintf() is less secure than snprintf(), so ensure that your system has snprintf() and re-run configure
      LPTSTR lpszCommand = malloc(65536); /* This was the value in previous versions */
      lpszCommand[0] = 0;
      sprintf(lpszCommand, EMAIL_COMMAND, HAVE_SENDMAIL, index.lpHeader->lpszEMailDomain, index.lpHeader->lpszAuditEMail);
#endif
      if (lpszCommand[0] && (hEMail = popen(lpszCommand, "w")) != NULL)
      {
        /* Write the e-mail header */
        LPTSTR lpszOperation = "performed an undefined operation";
        LPTSTR lpszRemoteHost = getenv("REMOTE_HOST");
        LPTSTR lpszRemoteAddr = getenv("REMOTE_ADDR");
        fprintf(hEMail, "Received: from %s ([%s])\n", (lpszRemoteHost)? lpszRemoteHost: lpszRemoteAddr, lpszRemoteAddr);
        fprintf(hEMail, "From: DirList2 Audit <nobody@%s>\n", index.lpHeader->lpszEMailDomain);
        fprintf(hEMail, "To: %s\n", index.lpHeader->lpszAuditEMail);
        fprintf(hEMail, "Subject: Audit Message\n");
        fprintf(hEMail, "X-UserAgent: %s\n\n", getenv("HTTP_USER_AGENT"));

        /* Determine the operation name */
        switch (dwOperation)
        {
        case OPERATION_WEBQUERY :         lpszOperation = "performed a web query"; break;
        case OPERATION_GETENTRIES :       lpszOperation = "retrieved directory entries"; break;
        case OPERATION_MODIFYENTRIES :    lpszOperation = "modified directory entries"; break;
        case OPERATION_DELETEENTRIES :    lpszOperation = "deleted directory entries"; break;
        case OPERATION_GETCLASSES :       lpszOperation = "retrieved categorical classes"; break;
        case OPERATION_MODIFYCLASSES :    lpszOperation = "modified categorical classes"; break;
        case OPERATION_DELETECLASSES :    lpszOperation = "deleted categorical classes"; break;
        case OPERATION_GETSORTTABLES :    lpszOperation = "retrieved sort tables"; break;
        case OPERATION_MODIFYSORTTABLES : lpszOperation = "modified sort tables"; break;
        case OPERATION_DELETESORTTABLES : lpszOperation = "deleted sort tables"; break;
        case OPERATION_GETINDEXINFO :     lpszOperation = "retrieved index information"; break;
        case OPERATION_MODIFYINDEXINFO :  lpszOperation = "modified index information"; break;
        case OPERATION_CREATEINDEX :      lpszOperation = "created a new index file"; break;
        case OPERATION_DELETEINDEX :      lpszOperation = "deleted an index file"; break;
        case OPERATION_LOCK :             lpszOperation = "locked the index"; break;
        case OPERATION_UNLOCK :           lpszOperation = "unlocked the index"; break;
        case OPERATION_GETUSERS :         lpszOperation = "retrieved a list of users"; break;
        case OPERATION_MODIFYUSERS :      lpszOperation = "modified users"; break;
        case OPERATION_DELETEUSERS :      lpszOperation = "deleted users"; break;
        case OPERATION_NOOP :             lpszOperation = "did nothing (no-op)"; break;
        }

        /* Write the actual e-mail message */
        fprintf(hEMail, "User \"%s\" (%s) at %s logged in and %s%s.\n\n", (lpUser)? lpUser->lpszName: "Anonymous", (lpUser)? lpUser->lpszDescription: "nobody", lpszRemoteAddr, lpszOperation, (bErrorCondition)? " but did not succeed (see transcript for details)": "");

        /* Write the access plan */
        WriteAccessPlanLog(hEMail, &plan);

        /* Write the transcript if we have one */
        if (lpszAuditBuffer) {
          fprintf(hEMail, "A transcript of the communication is given below:\n");
          fprintf(hEMail, "%s\n", lpszAuditBuffer);
        } else {
          fprintf(hEMail, "The operation completed with no transcript.\n");
        }

        /* We're done */
        pclose(hEMail);
      }

      /* Free the command line */
      free(lpszCommand);
    }
#else
    if (bAuditEnable && (index.lpHeader->dwFlags & HEADER_AUDITEMAIL))
      printf("#Auditing enabled but cannot send e-mail alerts because sendmail was not found at compile-time");
#endif

    /* Unallocate the memory */
    if (lpszAuditBuffer)
      free(lpszAuditBuffer);
    FreeAccessPlan(&plan);
    FreeIndex(&index);
    free(lpszIndexFile);
    FreeCGIVars(lpCGIVars);
    if (safeStrings) {
      free(safeStrings);
    }
    return 0;
  }
}
