/*
Directory listing CGI program
By Steven Lawrance
(c) Steven Lawrance, 2000
(c) Bryant College, 1998-2000

Permission granted for release under the GPL in Q2 2000
by Bryant College.  Steven Lawrance can be contacted at
slawrance@yahoo.com.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __HEADER__DIRLIST__UTIL_H
#define __HEADER__DIRLIST__UTIL_H

#include "dirlist.h"
#include "database.h"

#define FA_DIREC  4
#define FA_FILE   16

typedef struct {
  unsigned long  ff_attrib;      /* attribute found */
  LPTSTR         ff_name;       /* found file name */
  LPTSTR         ff_dir;
  LPTSTR         ff_find;
  DIR            *openDir;
  int            searchExtensionLength;
  int            searchLastPeriod;
  BOOL           caseSensitive;
} FILEFIND_DATA, *LPFILEFIND_DATA;

/* Different types of searches */
#define SEARCH_SUBSTRING        0
#define SEARCH_EXACT            1
#define SEARCH_STARTSWITH       2
#define SEARCH_STRICTSTARTSWITH 3
#define SEARCH_ENDSWITH         4
#define SEARCH_LIKE             5
#define SEARCH_ESCAPEMASK       0x0000FF00  /* these bits are reserved for the escape character for SEARCH_LIKE */

/* Unknown length for SafeWebString text input */
#define STR_NUL_TERM 0xFFFFFFFF

/* SafeWebString escape modes */
#define SAFEWEBSTRING_ESCAPE_JAVASCRIPT 0
#define SAFEWEBSTRING_ESCAPE_HTML 1
#define SAFEWEBSTRING_ESCAPE_RESPONSE 2

/* Forward References */
LPTSTR TrimString(LPTSTR str);
DWORD infprintf(LPCTSTR args, ...);
DWORD SafeEMailAddress(LPTSTR str);
DWORD SafeDomainName(LPTSTR str);
DWORD SafeFileName(LPTSTR str);
LPCTSTR SafeResponseString(LPTSTR *safeStrings, LPDWORD safeStringsLen, LPDWORD safeStringsAlloc, LPCTSTR text, DWORD textLen);
LPCTSTR SafeHTMLString(LPTSTR *safeStrings, LPDWORD safeStringsLen, LPDWORD safeStringsAlloc, LPCTSTR text, DWORD textLen);
LPCTSTR SafeJavaScriptString(LPTSTR *safeStrings, LPDWORD safeStringsLen, LPDWORD safeStringsAlloc, LPCTSTR text, DWORD textLen);
LPCTSTR SafeWebString(LPTSTR *safeStrings, LPDWORD safeStringsLen, LPDWORD safeStringsAlloc, LPCTSTR text, DWORD textLen, DWORD escapeMode, va_list *vaList);
BOOL SearchLIKEStrings(LPCTSTR str, LPCTSTR startstr, BYTE cEscape);
BOOL SearchEndingStrings(LPCTSTR str, LPCTSTR startstr);
BOOL SearchStartingStrings(LPCTSTR str, LPCTSTR startstr);
BOOL SearchStrictStartingStrings(LPCTSTR str, LPCTSTR startstr);
int CompareStrings(LPCTSTR str1, LPCTSTR str2);
BOOL FindSubstring(LPCTSTR str, LPCTSTR substr);
int CompareNames(LPCTSTR lpszFirst1, LPCTSTR lpszLast1, LPCTSTR lpszFirst2, LPCTSTR lpszLast2, BOOL bLastFirst);
BOOL ProcessWildcards(LPCTSTR file, LPCTSTR search, int searchLastPeriod, int searchExtensionLength, BOOL caseSensitive);
BOOL FindFirst(LPCTSTR lpszDir, LPCTSTR lpszName, LPFILEFIND_DATA ffData, BOOL caseSensitive);
BOOL FindNext(LPFILEFIND_DATA ffData);
void CloseFind(LPFILEFIND_DATA ffData);
BOOL VerifyHTML(FILE *pFile);
void CreatePasswdCache();
BOOL GetPasswdEntry(LPCTSTR lpszUser, LPTSTR *lpszName);
void FreePasswdCacheNode(LPDIRLIST_PASSWD_ENTRY lpNode);
void FreePasswdCache();
LPTSTR strtok2(LPTSTR lpszStr, UCHAR cDelimiter);
BOOL IsFlagSet(LPTSTR *lpszCharList, LPTSTR lpszFlagName);
LPTSTR GetCmdLineParam(LPTSTR *lpszCharList, DWORD dwPos, LPTSTR lpszDefault);
LPTSTR GetLastName(LPCTSTR fullName);

#endif /* UTIL_H */
